#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#  Łukasz 'sil2100' Zemczak
#  Robert Bruce Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Build Job

Prepare, build and check packages.

Environment variables:

PACKAGES_TO_REBUILD
    Only rebuild the specified source package names (space separated).

WATCH_ONLY
    Don't trigger a build, only watch what's already building in the PPA.

FORCE_REBUILD
    Force rebuilding an MP even if there is no diff with dest.

IGNORE_STEP
    Perform build even if one was already done.

IGNORE_MISSING_TWINS
    Don't enforce building a -gles twin for packages that usually have it.

DONT_REORDER_PREREQS
    Don't forcefully reorder branches if MP prerequisites are used.

TAKE_WHOLE_COMMIT_MESSAGE
    Use whole commit message in changelog, not just the brief part.

REBUILD_SOURCES_FOR_SYNC
    Only useful for sync requests. Instead of doing a binary copy of the
    packages from the selected archive, copy the sources and rebuild.

DEBUG
    Enable debug infos
"""

import glob
import logging
import os
import re
import shutil
import subprocess
import sys

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from citrain import watch_ppa
from citrain.prepare_silo import clean_source
from cupstream2distro import (
    branchhandling,
    packagemanager,
)
from cupstream2distro.settings import (
    BOT_DEBFULLNAME,
    SERIES_VERSION,
)
from cupstream2distro.launchpadmanager import lp
from cupstream2distro.silomanager import SiloState
from cupstream2distro.project import DotProject
from cupstream2distro.utils import (
    SILO_DIR,
    env,
    find_all,
    log_value_of,
    os_path_join_safe,
    reorder_branches,
    run_script,
)


silo_state = SiloState.safe_load(env.SILONAME)


def correct_changelog(package_dir, version, series):
    """Insert the correct series & version into the changelog."""
    changelog_path = os_path_join_safe(package_dir, 'debian', 'changelog')
    log_value_of.changelog_path('Mangling')
    with open(changelog_path) as f:
        changelog = f.readlines()
    parts = changelog[0].split(' ')
    parts[1] = '({})'.format(version)
    parts[2] = '{};'.format(series)
    changelog[0] = ' '.join(parts)
    log_value_of.parts()
    with open(changelog_path, 'w') as f:
        f.writelines(changelog)


def append_tag_to_version(version, tag):
    """Append a tag to the version in the correct place."""
    parts = version.split('-')
    parts[-2 if len(parts) > 1 else 0] += '~' + tag
    new_version = '-'.join(parts)
    log_value_of.new_version()
    return new_version


def prepare_source_sync(pkg, current_version, source_archive, source_series,
                        src_ppa, series, target_archive_name,
                        target_current_version):
    """Do a source copy from one silo to another.

    This function tries to prepare a sync of the selected source package
    from the source_archive, repackaging it and preparing for upload to the
    src_ppa silo PPA. If this fails, we return an empty tuple.
    """
    new_version = current_version
    direction = (source_archive.distribution.name, target_archive_name)
    if direction == ('ubuntu', 'ubuntu-rtm'):
        new_version = append_tag_to_version(current_version, 'rtm')
    elif direction == ('ubuntu-rtm', 'ubuntu'):
        new_version.replace('~rtm', '', 1)

    if packagemanager.is_upstream_version_for_series_in_dest(
            pkg, new_version, series, src_ppa):
        logging.info(
            "Package {} seems to be up-to-date in the silo PPA "
            "(version {})".format(pkg, new_version))
        return (None, None)

    dest_source = packagemanager.get_source_package_from_dest(
        pkg, source_archive, current_version, source_series)
    log_value_of.dest_source()

    # here's a tricky hack... in some cases we might need to
    # re-upload existing packages that got deleted since LP does
    # not allow same-version re-uploads, we need to actually bump
    # the ubuntu version number for it
    version_in_ppa = packagemanager.get_current_version_for_series(
        pkg, series.name, dest=src_ppa, all_packages=True)
    if version_in_ppa == new_version:
        new_version = packagemanager.increment_ubuntu_package_version(
            new_version)

    correct_changelog(dest_source, new_version, series.name)

    new_upstream_version = packagemanager.get_upstream_version(new_version)
    upstream_version = packagemanager.get_upstream_version(current_version)
    branch_parent = os.path.dirname(dest_source)

    log_value_of.current_version()
    log_value_of.new_version()
    # rename the tarball, if needed
    if new_version != current_version:
        tarball_pattern = os_path_join_safe(branch_parent, pkg + '*.tar.*')
        for orig_tarball in glob.glob(tarball_pattern):
            log_value_of.orig_tarball()
            new_orig_tarball = orig_tarball.replace(
                upstream_version, new_upstream_version)
            os.rename(orig_tarball, new_orig_tarball)
            log_value_of.new_orig_tarball()

    # to use our building tools, we actually need to hack around
    # and make a bzr repository (for bzr bd -S)
    logging.debug("Creating the bzr repository")
    subprocess.call(['bzr', 'init', '.'], cwd=dest_source)
    subprocess.call(['bzr', 'add', '.'], cwd=dest_source)

    # build the source package
    logging.info(
        "Preparing source package for the sync of {} version "
        "{}".format(pkg, new_version))
    packagemanager.build_source_package(dest_source, series.name, new_version)
    DotProject.soft_save(pkg, dest=current_version, ours=new_version)

    return (dest_source, new_version)


def prepare_binary_sync(pkg, current_version, source_archive, source_series,
                        src_ppa, series, target_archive_name,
                        target_current_version):
    """Perform a binary copy from one silo to another.

    This function tries to perform a binary sync of the selected source
    package from the source_archive, copying all the binaries (only building if
    it's missing some archs).
    """

    if packagemanager.is_upstream_version_for_series_in_dest(
            pkg, current_version, series, src_ppa):
        logging.info(
            "Package {} seems to be up-to-date in the silo PPA "
            "(version {})".format(pkg, current_version))
        return (None, None)

    src_ppa.copyPackage(
        from_archive=source_archive,
        from_pocket="Release",
        from_series=source_series,
        include_binaries=True,
        to_pocket="Release",
        to_series=series.name,
        source_name=pkg,
        version=current_version)

    dest_source_package = packagemanager.get_source_package_from_dest(
        pkg, source_archive, current_version, source_series)

    return (dest_source_package, current_version)


def also_build_for(branch_dir, series, version):
    """Copy an existing branch into a second branch for a different series."""
    series_dir = os_path_join_safe(
        SILO_DIR(), series, os.path.basename(branch_dir))
    series_version = append_tag_to_version(version, series)
    shutil.rmtree(series_dir, ignore_errors=True)
    shutil.copytree(branch_dir, series_dir)
    correct_changelog(series_dir, series_version, series)
    return series_dir


def main():
    silo_state.set_preparing()
    silo_state.save_config()

    src_ppa = lp.load(silo_state.ppa)
    dest = lp.load(silo_state.dest)
    series = lp.load(silo_state.series)

    sync_request = False

    if silo_state.source_archive:
        logging.debug(
            'Is this landing sync from a silo? '
            '{}'.format(silo_state.source_sync_from_silo))
        sync_request = True

    if not env.PACKAGES_TO_REBUILD and env.WATCH_ONLY != 'true':
        if not silo_state.is_empty:
            if silo_state.source_sync_from_silo:
                logging.info('Doing a whole rebuild for this silo sync.')
            elif env.IGNORE_STEP == 'true':
                logging.info('Honoring your request to rebuild everything.')
            else:
                logging.error(
                    "You already tried to build everything. You need to "
                    "either:\n - use the ignore step option (to rebuild all "
                    "MPs)\n - use the \"only rebuild packages\" option (to "
                    "rebuild only some components associated to a MP) \n - "
                    "use the watch only option (to only check ppa status "
                    "without any rebuild).")
                silo_state.set_build_failed(
                    "Can't build: Already built, check available options.")
                silo_state.save_config()
                return 1

    # reset the status
    silo_state.set_empty()
    silo_state.save_config()

    # detect if we have a prepare only source not in configuration
    components_not_in_config = []
    for source in env.PACKAGES_TO_REBUILD.split():
        if source not in silo_state.mps and (
                not sync_request or source not in silo_state.sources):
            components_not_in_config.append(source)
    if components_not_in_config:
        message = (
            "One or more components provided in rebuildpackages ({}) are "
            "not in the MP silo configuration. Ask for a reconfiguration of "
            "the silo with the new MPs to add please.".format(
                ", ".join(components_not_in_config)))
        logging.error(message)
        silo_state.set_build_failed("Can't build: " + message)
        silo_state.save_config()
        return 1

    # cycle through all branches and prepare them
    components_to_prepare = {}
    mps = silo_state.mps
    for source in mps:
        if ((not env.PACKAGES_TO_REBUILD and env.WATCH_ONLY != 'true') or
                source in env.PACKAGES_TO_REBUILD.split()):
            logging.info("Adding {} MP(s) to prepare".format(source))
            for mp_link in mps[source]:
                components_to_prepare.setdefault(source, []).append(
                    lp.load(mp_link))
            # cleaning (in case of partial rebuild)
            clean_source(source)

    all_sources = list(components_to_prepare) + silo_state.sources
    missing = packagemanager.check_if_packages_require_twin_upload(all_sources)
    if missing:
        message = (
            "Some projects are missing their 'twin package' uploads (e.g. "
            "their -gles counter-parts): {}.".format(", ".join(missing)))
        if env.IGNORE_MISSING_TWINS == 'true':
            logging.warning(message + " But the ignore flag was set.")
        else:
            logging.error(message)
            silo_state.set_build_failed("Can't build: " + message)
            silo_state.save_config()
            return 1

    # prepare source and dput by one
    src_ppa_name = "{}/{}/{}".format(
        src_ppa.owner.name, src_ppa.distribution.name, src_ppa.name)
    dest_ppa_name = "{}/{}/{}".format(
        dest.owner.name, src_ppa.distribution.name, dest.name)
    log_value_of.src_ppa_name()
    log_value_of.dest_ppa_name()

    for source_package in components_to_prepare:
        logging.info(
            "=== Trying to prepare {} ===".format(source_package))

        # get dest branch first
        dest_branch = components_to_prepare[
            source_package][0].target_branch.web_link
        logging.info("Branching {} to {}".format(dest_branch, source_package))
        pending_branch = os_path_join_safe(SILO_DIR(), source_package)
        changelog_path = os_path_join_safe(
            pending_branch, 'debian', 'changelog')
        branchhandling.get_branch(dest_branch, pending_branch)
        dest_tip_bzr_rev = branchhandling.get_tip_bzr_revision(pending_branch)
        dest_version = packagemanager.get_packaging_version(pending_branch)

        # first, try looking into the merge list and re-order it if there are
        # any prereqs present if no prereqs are present, the components list
        # stays the same
        if env.DONT_REORDER_PREREQS != 'true':
            components = reorder_branches(
                components_to_prepare[source_package])
        else:
            components = components_to_prepare[source_package]

        # get all branches one after another and merge them in pending_branch
        for mp in components:
            branch = mp.source_branch.web_link
            logging.info("Trying to merge {} (at rev {}) into {}".format(
                branch,
                mp.source_branch.revision_count,
                mp.target_branch.web_link))

            bugs_fixes = ""
            for bug in mp.getRelatedBugTasks():
                if not bugs_fixes:
                    bugs_fixes = "Fixes: "
                else:
                    bugs_fixes += ", "
                bugs_fixes += "#" + bug.self_link.split('/')[-1]

            # use a set to fetch the reviewer list
            approved_by_set = set()
            for review in mp.all_comments:
                if review.vote == "Approve":
                    approved_by_set.add(review.author.display_name)
                elif review.vote in ("Disapprove", "Needs Fixing"):
                    if review.author.display_name in approved_by_set:
                        approved_by_set.remove(review.author.display_name)

            approved_by = ""
            for reviewer in approved_by_set:
                if not approved_by:
                    approved_by = "\nApproved by: "
                else:
                    approved_by += ", "
                approved_by += reviewer

            commit_message = mp.commit_message
            if commit_message:
                commit_message = "{} {}{}".format(
                    commit_message.encode('utf-8'),
                    bugs_fixes,
                    approved_by.encode('utf-8'))

            # grab committers
            committers = branchhandling.grab_committers_compared_to(
                pending_branch, branch)

            try:
                if not branchhandling.merge_branch(
                        pending_branch,
                        branch,
                        commit_message,
                        authors=committers):
                    logging.error(
                        "Conflicts when attempting to merge {} into previous "
                        "merged + {}. You will need to reconcile manually "
                        "branches that you provided as a parameter for "
                        "merging into the destination branch and rerun this "
                        "job".format(branch, mp.target_branch.web_link))
                    silo_state.set_build_failed("Can't build: Merge conflict.")
                    silo_state.save_config()
                    return 1
            # TODO: we should loop to show if there is more issues and then
            # exit
            except branchhandling.NoCommitFoundException:
                message = (
                    "There is no commit message in {} and we can't get it "
                    "from debian/changelog. Please check that you set a "
                    "commit message on all your MPs if you didn't touch "
                    "debian/changelog.".format(branch))
                logging.error(message)
                silo_state.set_build_failed("Can't build: " + message)
                silo_state.save_config()
                return 1

        #  Get changelog informations
        logging.info("Get changelog info")
        prep_version = packagemanager.get_packaging_version(pending_branch)

        # Check version in ppa
        logging.info("Get previous version in silo")
        version_in_ppa = packagemanager.get_current_version_for_series(
            source_package, series.name, src_ppa_name)
        if packagemanager.is_version1_higher_than_version2(
                version_in_ppa,
                prep_version):
            logging.info(
                "A version in the ppa ({}) is higher than the proposed "
                "version in bzr ({}) (previous tests/builds failing?). "
                "Basing on that one.".format(
                    version_in_ppa,
                    dest_version))
            previous_packaging_version = version_in_ppa
        else:
            previous_packaging_version = prep_version

        # Check version in distro
        logging.info("Get version available in distro")
        distro_version = packagemanager.get_current_version_for_series(
            source_package,
            series.name,
            dest=dest.distribution.main_archive)

        # if the target is the distro, we ensure we have all releases in
        # proposed branch
        if lp.is_link_distro_archive(silo_state.dest):
            prev_version = distro_version
        # target is a ppa
        else:
            prev_version = packagemanager.get_current_version_for_series(
                source_package, series.name, dest=dest)

        # ensure that the previous release (in distro or in dest ppa) ended up
        # in the vcs
        force_rebuild = False
        if env.FORCE_REBUILD == 'true':
            logging.info(
                "Force rebuild option set. Will update even if destination "
                "don't contain latest version requested.")
            force_rebuild = True
        else:
            with open(changelog_path) as changelog:
                if not packagemanager.is_version_in_changelog(
                        prev_version,
                        changelog):
                    message = (
                        "A version ({}) is available at the destination "
                        "archive for that component but is not in the "
                        "destination branch which is still at {}. You "
                        "need to ensure that your version contains the "
                        "fix in the destination or you can force "
                        "rebuild to bypass the check.".format(
                            prev_version,
                            prep_version))
                    logging.warning(message)
                    silo_state.set_build_failed(
                        "Can't build: Some projects are missing "
                        "revisions from distribution.")
                    silo_state.save_config()
                    return 1

        if not lp.is_link_distro_archive(silo_state.dest):
            # if the version in the distro is higher than the previous
            # packaging version (either the version in intermediate ppa or
            # packaging bzr version We are going to force releasing, but basing
            # on this version
            if packagemanager.is_version1_higher_than_version2(
                    distro_version,
                    previous_packaging_version):
                logging.info(
                    "The version in distro ({}) is higher than what is "
                    "proposed in the bzr version for that branch ({}). "
                    "Basing on the distro version.".format(
                        distro_version, previous_packaging_version))
                logging.info(
                    "Forcing release mode as the version in distro ({}) is "
                    "higher than current destination ppa ({})".format(
                        distro_version, previous_packaging_version))
                previous_packaging_version = distro_version
                force_rebuild = True

        # Check if we need to republish
        dest_source_package = None
        logging.info("Check if we need to release a new package")
        dest_source_package = packagemanager.get_source_package_from_dest(
            source_package,
            dest,
            prev_version,
            series.name)
        # nothing to release
        if not packagemanager.is_new_content_relevant(pending_branch,
                                                      dest_source_package):
            if not force_rebuild:
                logging.info(
                    "No new useful revision published compared to "
                    "dest, no need to upload this component")
                continue
            logging.info(
                "No new useful revision published but force rebuilding is set")

        if lp.is_link_distro_archive(silo_state.dest):
            # only version if we don't see the flag to not version it.
            if not packagemanager.has_dont_change_version_flag(pending_branch):
                version = packagemanager.create_new_packaging_version(
                    previous_packaging_version,
                    series_version=SERIES_VERSION[series.name])
            else:
                version = packagemanager.get_packaging_version(pending_branch)
        else:
            version = packagemanager.create_new_packaging_version(
                previous_packaging_version,
                destppa=dest_ppa_name,
                series_version=SERIES_VERSION[series.name])

        # Refresh symbol files
        logging.info("Update symbol files version if needed")
        packagemanager.refresh_symbol_files(pending_branch, version)

        logging.info("Prepare changelog content")
        authors_commits = {}

        try:
            # first try: in case someone release and release a version without
            # merging back (we don't want to duplicate the commits). This is
            # the main case anyway.
            log_content = branchhandling.log_diff_from_tag(
                pending_branch, dest_version)
        except:
            try:
                # second hammer: we then recollect from the last version in the
                # destination (some lines can be duplicated)
                log_content = branchhandling.log_diff_from_tag(
                    pending_branch, prev_version)
            except:
                # last hammer with FIXME: -100 is a hack as we don't have
                # anymore any information on when the last release happen.
                # People are transitionned to a tag system.
                log_content = branchhandling.log_diff_since_last_release(
                    branchhandling.log_diff(pending_branch, '-100'))
        alreadyfixed_bugs = set()
        with open(changelog_path) as f:
            alreadyfixed_bugs = packagemanager.collect_bugs(
                f, source_package)
        (authors_commits, bugs_set) = branchhandling.collect_author_commits(
            log_content,
            alreadyfixed_bugs,
            take_whole_commit=env.TAKE_WHOLE_COMMIT_MESSAGE == 'true')
        if force_rebuild:
            authors_commits.setdefault(
                BOT_DEBFULLNAME, []).append("New rebuild forced")
        packagemanager.update_changelog(
            pending_branch, version, series.name, None, authors_commits)
        branchhandling.commit_release(pending_branch, version)

        logging.info("Sync launchpad with current proposed work")
        lp.open_bugs_for_source(bugs_set, source_package, series, dest)

        logging.info("Build source package")

        branch_dir = os_path_join_safe(SILO_DIR(), source_package)

        packagemanager.build_source_package(
            branch_dir, series.name, prev_version)

        # TODO: Enable dual builds once publication support lands.
        # if series.name == 'vivid' and env.WATCH_ONLY != 'true':
        #     series_dir = also_build_for(branch_dir, 'utopic', version)
        #     packagemanager.build_source_package(
        #         series_dir, 'utopic', prev_version)

        newdsc_path = os_path_join_safe(SILO_DIR(), '{}_{}.dsc'.format(
            source_package, version.split(':')[-1]))

        # second diff step: in case the source package was created from nothing
        # (like a diff between trunk and the source package, but none after
        # newly .dsc created extraction)
        if not force_rebuild and dest_source_package:
            logging.info(
                "Check that the newly created source "
                "package has relevant diff")
            if not packagemanager.is_relevant(
                    newdsc_path, dest_source_package):
                logging.warning(
                    "There was a diff between trunk and the source "
                    "package in ubuntu triggering a source package build. "
                    "However, after that one was created, we diffed again "
                    "between that source in ubuntu and the newly created "
                    "source extracted and found nothing relevant. Not upload "
                    "that one then. Please check that the relevant diff in "
                    "the package creation is justified by diffing manually "
                    "trunk and the source in ubuntu.")

        # check the .changes file: that we don't have an empty changelog
        logging.info("Checking finale changelog")
        with open(os_path_join_safe(SILO_DIR(), '{}_{}_source.changes'.format(
                source_package,
                version.split(':')[-1]))) as f:
            for line in f:
                if re.match("   \* *$", line):
                    message = (
                        "You generated an emtpy changelog content for {}. "
                        "This generally means that one or many merge "
                        "proposals have a change in debian/changelog with "
                        "no UNRELEASED or no content in the changelog. "
                        "If you modify debian/changelog in your commit, "
                        "the commit message for the merge proposal isn't "
                        "used, so you need to put yourself some changelog "
                        "content into that file. Please change those MP "
                        "and rerun the build job.".format(source_package))
                    logging.error(message)
                    silo_state.set_build_failed(
                        "Build failed: Some packages generate an empty "
                        "changelog. Check the job for more info.")
                    silo_state.save_config()
                    return 1

        # Prepare if a packaging change is part of the diff
        logging.info("Generate a diff if meaningfull packaging changes")
        diff_filepath = DotProject(source_package).get_diff_path(version)
        olddsc_path = None
        if dest_source_package:
            dest_parent = os.path.dirname(dest_source_package)
            for path in glob.glob(os_path_join_safe(dest_parent, '*.dsc')):
                olddsc_path = path
        packagemanager.generate_diff_between_dsc(
            diff_filepath, olddsc_path, newdsc_path)
        DotProject.soft_save(
            source_package, branch, dest_tip_bzr_rev, prev_version, version)

    # this landing has a sync request in it
    if sync_request and env.WATCH_ONLY != 'true':
        source_archive = lp.load(silo_state.source_archive)
        source_series = silo_state.source_series

        if silo_state.source_sync_from_silo:
            logging.debug("Syncing whole silo contents from the selected silo")

            request_id = silo_state.source_sync_from_silo
            log_value_of.request_id()
            if request_id:
                # check if the silo still exists
                source_distro = source_archive.distribution.name
                siloname = silo_state.source_archive.split("/")[-1]
                log_value_of.siloname()
                try:
                    source_state = SiloState(
                        os_path_join_safe(source_distro, siloname))
                except Exception as e:
                    logging.warning('Failed to load source silo: {}'.format(e))
                    source_state = None
                log_value_of.source_state()
                if (not source_state or
                        source_state.requestid != request_id or
                        source_state.is_cleaning):
                    logging.warning(
                        'Source silo gone, using {} distro'.format(
                            source_distro))
                    source_archive = lp.distributions[
                        source_distro].main_archive
                    silo_state.source_archive = source_archive.self_link
                    silo_state.source_sync_from_silo = False

        rebuild = env.PACKAGES_TO_REBUILD.split()
        if not rebuild:
            sources_to_sync = silo_state.sources
        else:
            sources_to_sync = set(silo_state.sources).intersection(rebuild)

        for pkg in sources_to_sync:
            logging.info("Looking at {} as per sync request".format(pkg))
            current_version = packagemanager.get_current_version_for_series(
                pkg,
                source_series,
                dest=source_archive)

            long_func = packagemanager.get_current_version_for_series
            dest_current_version = long_func(
                pkg, series.name, dest=dest)

            new_version = current_version

            if env.REBUILD_SOURCES_FOR_SYNC == "true":
                sync_func = prepare_source_sync
            else:
                sync_func = prepare_binary_sync

            (dest_source_package, new_version) = sync_func(
                pkg,
                current_version,
                source_archive,
                source_series,
                src_ppa,
                series,
                dest.distribution.name,
                dest_current_version)

            if not new_version or not dest_source_package:
                continue

            log_value_of.dest_source_package()
            log_value_of.new_version()

    if env.WATCH_ONLY != 'true':
        logging.info('Uploading packages to the PPA.')
        for changes in find_all('_source.changes'):
            packagemanager.simple_upload(changes, src_ppa_name)

    # watch ppa
    silo_state.set_building()
    silo_state.save_config()
    logging.info("Check ppa status in {}".format(src_ppa.web_link))
    if watch_ppa.main() != 0:
        message = 'Build failed: Some packages failed to build.'
        logging.error(message)
        silo_state.set_build_failed(message)
        silo_state.save_config()
        return 1
    logging.info('Build job complete, exiting.')
    return 0


run_script(__name__, __doc__, main)
