#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Check Publication Migration Job

Environment Variables:

DEBUG
    Enable debug infos
"""

import logging
import os
import sys

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from citrain import merge_clean

from cupstream2distro.utils import env, log_value_of, run_script
from cupstream2distro import packagemanager
from cupstream2distro.silomanager import SiloState
from cupstream2distro.project import DotProject
from cupstream2distro.settings import (
    ALL_SILO_NAMES,
    SILOS_DIR,
)


def main():
    if not os.path.isdir(SILOS_DIR):
        logging.warning("No silo available/setup")
        return 0

    # loop on all known silos
    for silo_name in ALL_SILO_NAMES:
        logging.info("Checking {}".format(silo_name))
        try:
            env.SILONAME = silo_name
        except OSError:
            logging.debug("{} isn't created".format(silo_name))
            continue

        silo_state = SiloState.safe_load(silo_name)
        if not silo_state:
            continue

        config_timestamp = silo_state.get_mtime()

        if not silo_state.is_published:
            log_value_of.silo_name('Skipping unpublished')
            continue

        # get latest uploaded version for all packages
        logging.info(
            "Getting all sources that were uploaded with their latest version")
        packages_in_dest = {}
        packages = DotProject.get_all_packages_uploaded()
        for (source, version, rev, branch) in packages:
            previous_scanned_uploaded_version = packages_in_dest.get(
                source, "0")
            if packagemanager.is_version1_higher_than_version2(
                    version,
                    previous_scanned_uploaded_version):
                packages_in_dest[source] = version

        # check if all packages have reached their destination
        result = packagemanager.check_package_reached_destination(
            silo_state, packages_in_dest, False, False)
        (all_pkg_in_dest, message) = result

        # check that the config file status hasn't change
        try:
            if config_timestamp != silo_state.get_mtime():
                logging.warning(
                    "Another process has updated the config file for {}. "
                    "Do not update it.".format(silo_name))
                continue
        except OSError:
            logging.warning(
                "{} isn't attributed anymore. Probably cleaned up meanwhile. "
                "Do not update it".format(silo_name))
            continue

        # we can have a transient race here (but unlikely to be permanent: as
        # we change the message everytime we start a new process, so finale
        # result will override)
        silo_state.set_migrating('Migration: ' + message)
        silo_state.save_config()

        if all_pkg_in_dest:
            # TODO: If utopic+vivid, push utopic packages to utopic PPA here.
            merge_clean.main()
    return 0


run_script(__name__, __doc__, main)
