#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#  Łukasz 'sil2100' Zemczak
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Merge & Clean Script

Merge all branches to merge, clean the ppa and free the silo slot.

Environment variables:

ONLY_FREE_SILO
    Only free the ppa and don't merge/check anything.

IGNORE_MISSINGPROJECTS
    Ignore if some projects that are going to be merged back were not
    published. No merge will then be processed for those.

IGNORE_PACKAGES_NOTINDEST
    Ignore if some packages are not published in the destination.

IGNORE_STEP
    Ignore if previous step didn't successfully end.

SILONAME
    Determines which silo to merge & clean.

DEBUG
    Enable debug infos
"""

import logging
import os
import shutil
import sys

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from cupstream2distro import (
    branchhandling,
    packagemanager,
)
from cupstream2distro.settings import (
    PROPOSED_URL,
    BRANCH_URL,
)
from cupstream2distro.utils import (
    SILO_DIR,
    ignored,
    env,
    log_value_of,
    os_path_join_safe,
    run_script,
)
from cupstream2distro.launchpadmanager import lp
from cupstream2distro.silomanager import SiloState
from cupstream2distro.project import DotProject


def merge_and_reconciliate(silo_state, packages_in_dest):
    """Merge and eventually reconcialiate all branches."""
    logging.info(
        "Check the revision from all merged branch in trunk didn't "
        "change. If changed, try to reconciliate them back.")
    reconciliate_all_branches = True
    series = lp.load(silo_state.series).name
    dest = lp.load(silo_state.dest)

    mps = silo_state.mps
    for source in mps:
        source_branch_uri = os_path_join_safe(SILO_DIR(), source)
        target_branch = lp.load(mps[source][0]).target_branch

        stored_tip_rev = DotProject(source).rev
        tip_rev = str(target_branch.revision_count)

        # prepare and remerge
        if tip_rev != stored_tip_rev:
            logging.info(
                "New revision ({}) in {} when prepared is based on {}. Trying "
                "to merge it for you.".format(tip_rev, source, stored_tip_rev))
            if not branchhandling.merge_branch(
                    source_branch_uri,
                    target_branch.web_link,
                    "Resync trunk"):
                proposed_branch = PROPOSED_URL.format(
                    target_branch.project.name,
                    dest.distribution.name, series)
                logging.error(
                    "We can't reconciliate your proposed branch (pushed to "
                    "{}) with the new commits in target (merge "
                    "conflicts).".format(proposed_branch))
                reconciliate_all_branches = False
    if not reconciliate_all_branches:
        message = (
            "Some branches couldn't be reconciliated with the changes "
            "happening meaning in destination branch. Please revert your "
            "changes in trunk and relaunch this job.")
        logging.error(message)
        silo_state.set_merge_failed('Can\'t merge: ' + message)
        silo_state.save_config()
        return 1

    logging.info("Pushing the branches to the parent ones")
    push_all_branches = True
    for source in mps:
        branch_uri = os_path_join_safe(SILO_DIR(), source)
        target_branch_link = lp.load(mps[source][0]).target_branch.web_link
        logging.info("Trying to push to {}".format(target_branch_link))
        if not branchhandling.push_to_branch(branch_uri, target_branch_link):
            url = BRANCH_URL.format(
                source,
                "{}-recup".format(dest.distribution.name))
            logging.error(
                "Can't push this branch to {}, pushing temporarly "
                "to {}.".format(target_branch_link, url))
            branchhandling.push_to_branch(branch_uri, url, overwrite=True)
            push_all_branches = False
    if not push_all_branches:
        message = (
            "Some branches couldn't be pushed to the destination. Please "
            "pull them from the url link above and relaunch this job with "
            "only free slot flag.")
        logging.error(message)
        silo_state.set_merge_failed('Can\'t merge: ' + message)
        silo_state.save_config()
        return 1


def main():
    silo_state = SiloState(env.SILONAME)
    if silo_state.is_merging:
        logging.error('The merge&clean job is already running. Skipping.')
        return 1
    elif silo_state.is_merge_failed:
        logging.warning('The last merge & clean job has failed.')
        if env.ONLY_FREE_SILO == 'true':
            logging.info('Ignoring that and attempting once again.')
        else:
            logging.error(
                "Please re-run the merge & clean job manually with "
                "ONLY_FREE_SILO after making sure that the branches are "
                "merged in by hand.")
            return 1

    silo_state.set_merging()
    silo_state.save_config()

    if env.ONLY_FREE_SILO != 'true':
        if not silo_state.is_published:
            if env.IGNORE_STEP == 'true':
                logging.info('Last step didn\'t succeed, continuing anyway.')
            else:
                logging.error(
                    "Last step didn't finish successfully. You need to "
                    "either ignore that the previous step didn't finished "
                    "successfully or ensuring that prepare, build and "
                    "checked passed.")
                silo_state.set_merge_failed(
                    "Can't merge: wrong status or parameters for job.")
                silo_state.save_config()
                return 1
            # reset it
            silo_state.set_published()
            silo_state.save_config()

    # get latest uploaded version for all packages
    logging.info('Getting latest version of all uploaded sources')
    packages_in_dest = {}
    packages = DotProject.get_all_packages_uploaded()
    log_value_of.packages()
    for (source, version, rev, branch) in packages:
        previous_scanned_uploaded_version = packages_in_dest.get(source, "0")
        log_value_of.previous_scanned_uploaded_version()
        if packagemanager.is_version1_higher_than_version2(
                version,
                previous_scanned_uploaded_version):
            packages_in_dest[source] = version
    log_value_of.packages_in_dest()

    # check that all packages have reached their destination
    result = packagemanager.check_package_reached_destination(
        silo_state,
        packages_in_dest,
        env.IGNORE_MISSINGPROJECTS == 'true',
        env.IGNORE_PACKAGES_NOTINDEST == 'true',
        for_merge_and_clean=True)
    (all_pkg_in_dest, message) = result
    if not all_pkg_in_dest and env.ONLY_FREE_SILO != 'true':
        silo_state.set_migrating(message)
        silo_state.save_config()
        return 1

    if env.ONLY_FREE_SILO != 'true':
        if merge_and_reconciliate(silo_state, packages_in_dest):
            logging.error("Merge and reconciliation failed, aborting.")
            return 1
    else:
        logging.warning('Not merging potential branches.')

    silo_state.set_done()

    if env.ONLY_FREE_SILO != 'true':
        silo_state.set_cleaning('Landed. Cleaning silo.')
    # we gave up
    else:
        if not all_pkg_in_dest:
            silo_state.set_cleaning('Gave up this landing. Cleaning silo.')
        else:
            silo_state.set_cleaning(
                'Landed without merging branch. Cleaning silo.')

    silo_state.save_config()

    ppa = lp.load(silo_state.ppa)
    series = lp.load(silo_state.series)
    logging.info(
        "Clean now the source ppa {} in {}".format(ppa.web_link, series.name))

    for pkg in ppa.getPublishedSources(distro_series=series):
        if pkg.status != "Deleted":
            logging.info(
                "Requesting deletion for {}".format(pkg.source_package_name))
            pkg.requestDeletion()

    # Finally, nuke the silo dir so we don't waste space on the server
    with ignored(OSError):
        shutil.rmtree(SILO_DIR())
        os.mkdir(SILO_DIR())
    silo_state.remove_rsync_file()
    return 0


run_script(__name__, __doc__, main)
