#!/usr/bin/python -u
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Didier Roche
#  Łukasz 'sil2100' Zemczak
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Prepare Silo Job

Prepare and assign a free silo from a ppa.
You must specify either MPs or sources.

Environment variables:

DEBUG
    Enable debug infos

DEST
    Use this destination PPA instead of {series}-proposed

DISTRIBUTION
    Which distribution to target (default: ubuntu)

IGNORECONFLICTS
    Allow same source package in multiple silos. You must coordinate both silos
    to ensure that when one publishes, the other one rebuilds, otherwise the
    second publication will revert the work done in the initial publication.

LANDERS
    Landers who are owners of that request (should be IRC nicks for pinging)

MERGE_PROPOSALS
    Merge proposal urls (space separated)

RECONFIGURE_SILO
    Reconfigure an existing silo

REQUEST_ID
    Update/Configure this request id

RESTRICTED
    Only add MPs for already-locked source packages (won't allow new sources)

SERIES
    Series used to build the package

SOURCES
    Source packages for manual upload (space separated)

SYNC_REQUEST
    Copy packages from a different silo/archive rather than building MPs.
"""

import lazr
import logging
import os
import sys
import glob
import shutil

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from cupstream2distro import branchhandling
from cupstream2distro.settings import (
    SILO_BUILDPPA_SCHEME,
    SILOS_DIR,
    SILO_NAME_LIST,
)
from cupstream2distro.utils import (
    SILO_DIR,
    env,
    ignored,
    log_value_of,
    os_path_join_safe,
    run_script,
)
from cupstream2distro.silomanager import SiloState
from cupstream2distro.launchpadmanager import lp
from cupstream2distro.project import DotProject


NEW_COMPONENT_FAILURE = """\
{} was not in the initial list of components for that silo. Please ask the \
trainguards to reconfigure this silo for you."""


def clean_source(source):
    """Clean all related source content from current silos."""
    silo_dir = SILO_DIR()
    config_file = DotProject(source).filename
    source_dir = os_path_join_safe(silo_dir, source)
    ubuntu_dir = os_path_join_safe(silo_dir, 'ubuntu', source)
    with ignored(OSError):
        log_value_of.config_file('Deleting!')
        os.remove(config_file)
    with ignored(OSError):
        log_value_of.source_dir('Deleting!')
        shutil.rmtree(source_dir)
    with ignored(OSError):
        log_value_of.ubuntu_dir('Deleting!')
        shutil.rmtree(ubuntu_dir)
    fileglob = '{}_*'.format(source)
    for filename in [(fileglob,), ('build-area', fileglob)]:
        for path in glob.glob(os_path_join_safe(silo_dir, *filename)):
            log_value_of.path('Deleting!')
            os.remove(path)


def clean_removed_sources(all_silo_projects, prev_components):
    logging.info("Check and clean all removed sources")
    log_value_of.all_silo_projects()
    prev_components = set(prev_components)

    result = DotProject.get_all_packages_uploaded()
    for (source, version, rev, branch) in result:
        log_value_of.source()
        prev_components.add(source)
    for source in sorted(prev_components):
        if source not in all_silo_projects:
            logging.info('Removing {} from silo.'.format(source))
            clean_source(source)


def main():
    if not env.REQUEST_ID:
        logging.error('REQUEST_ID is mandatory.')
        return 1
    if not (env.MERGE_PROPOSALS or env.SOURCES or env.SYNC_REQUEST):
        logging.error('Need either MERGE_PROPOSALS, SOURCES, or SYNC_REQUEST')
        return 1
    if env.RESTRICTED == 'true' and env.RECONFIGURE_SILO != 'true':
        logging.error('RESTRICTED mode requires RECONFIGURE_SILO')
        return 1

    project_urls = env.MERGE_PROPOSALS.split()
    project_urls.extend(env.SOURCES.split())

    distribution = env.DISTRIBUTION or 'ubuntu'
    # make sure the distribution parameter is one of the two supported right
    # now
    if distribution not in SILO_NAME_LIST:
        logging.error(
            "Distribution {} is currently not supported in the CI "
            "Train. Not assigning".format(distribution))
        return 1

    # create all silo names
    with ignored(OSError):
        os.mkdir(SILOS_DIR)
    for siloname in SILO_NAME_LIST[distribution]:
        with ignored(OSError):
            os.mkdir(os_path_join_safe(SILOS_DIR, siloname))

    # first, check that a silo is available
    dest = None
    series = None

    old_state = SiloState.find_by_requestid(env.REQUEST_ID)

    # check provided input matches
    if env.RECONFIGURE_SILO != 'true' and old_state:
        logging.error(
            'REQUEST_ID {} found in {} but you did not ask to '
            'reconfigure.'.format(env.REQUEST_ID, old_state.siloname))
        return 1
    if env.RECONFIGURE_SILO == 'true' and not old_state:
        logging.error('Could not find REQUEST_ID {} in any silo.'.format(
            env.REQUEST_ID))
        return 1

    if env.RECONFIGURE_SILO == 'true' and old_state:
        siloname = old_state.siloname
        logging.info('Reconfiguring {} for REQUEST_ID {}'.format(
            siloname, env.REQUEST_ID))
        # set the same destination, series
        dest = lp.load(old_state.dest)
        series = lp.load(old_state.series)
        # ensure that the series matches the old series field or yell
        if env.SERIES and series.name != env.SERIES:
            logging.error(
                'Can not change series from {} to {} by reconfiguring. '
                'Free this silo and reassign.'.format(series.name, env.SERIES))
            return 1
        prev_components = old_state.all_projects
        log_value_of.prev_components()
    else:
        logging.info("Checking that a silo is available")
        siloname = SiloState.find_first_available(
            distribution=env.DISTRIBUTION)
        if not siloname:
            logging.error("No silo is available")
            return 1

    # silo attribution (made automatically or explicitely)
    if not series:
        if not env.SERIES:
            logging.error(
                "You need to provide a series to attribute a new silo")
            return 1
        series = lp.get_series(env.SERIES, distro_name=env.DISTRIBUTION)
        if env.DEST:
            dest = lp.get_ppa(env.DEST)
        else:
            dest = lp.distributions[env.DISTRIBUTION].main_archive

    logging.info("Configuring for {} ({}) in {}".format(
        ", ".join(project_urls), series.name, dest.name))

    silo_state = SiloState.new_blank(
        siloname=siloname,
        requestid=env.REQUEST_ID,
        dest=dest.self_link,
        series=series.self_link,
        landers=env.LANDERS.split())

    if env.SYNC_REQUEST:
        split = env.SYNC_REQUEST.split(",")
        # check if it's a sync from a PPA or a distro
        if split[0].startswith("ppa:"):
            logging.info("Sync from a PPA ({}) requested".format(split[0]))

            ppa_name = split[0].split(":")[1]
            if ppa_name[0] == "~":
                ppa_name = ppa_name[1:]
            source_archive = lp.get_ppa(ppa_name)
            if len(split) > 1:
                source_series = split[1]
        elif split[0].isdigit():
            # for silo synces we only handle ubuntu and ubuntu-rtm
            source_distro = (
                "ubuntu" if env.DISTRIBUTION == "ubuntu-rtm" else
                "ubuntu-rtm")
            source_landing = "landing-{:03d}".format(int(split[0]))
            source_archive = lp.get_ppa(
                "ci-train-ppa-service/{}/{}".format(
                    source_distro,
                    source_landing))
            source_series = (
                "vivid" if env.DISTRIBUTION == "ubuntu-rtm" else
                "14.09")

            logging.info(
                "Sync from a silo ({}/{}) requested".format(
                    source_distro,
                    source_landing))

            # read the configuration of the source silo
            try:
                source_state = SiloState(
                    os_path_join_safe(source_distro, source_landing))
            except Exception as e:
                logging.error('Could not load source silo sync ({})'.format(e))
                return 1

            silo_state.source_sync_from_silo = source_state.requestid
            if not env.SOURCES:
                logging.info('Syncing whole silo.')
                silo_state.source_sync_all_contents = True
                silo_state.sources = source_state.all_projects
        else:
            logging.info('Syncing from distro {}.'.format(split[0]))

            source_archive = lp.distributions[split[0]].main_archive
            if len(split) > 1:
                source_series = split[1]

        if not source_archive:
            logging.error('Source archive {} invalid.'.format(split[0]))
            return 1

        if not source_series:
            logging.error('Sync request {} missing series'.format(
                env.SYNC_REQUEST))
            return 1

        if not env.SOURCES and not silo_state.source_sync_all_contents:
            logging.error('You need to specify source packages to sync.')
            return 1

        silo_state.source_archive = source_archive.self_link
        silo_state.source_series = source_series

    silo_path = os_path_join_safe(SILOS_DIR, siloname)
    log_value_of.silo_path()
    log_value_of.siloname()

    if env.SOURCES:
        logging.info(
            "Additional packages will be directly uploaded to the ppa: "
            "{}".format(", ".join(env.SOURCES.split())))
        silo_state.sources = env.SOURCES.split()

    one_project_already_in_use = False
    source_packages_seen = set()
    # grab the project name of all MP
    for mp_url in env.MERGE_PROPOSALS.split():
        log_value_of.mp_url()
        mp = lp.load_url(mp_url)
        try:
            long_func = branchhandling.get_source_package_name_from_branch
            source_package_name = long_func(mp.target_branch.web_link)
        except AttributeError:
            logging.error(
                "{} is not a valid merge proposal".format(mp_url))
            return 1
        project_name = mp.target_branch.project.name
        log_value_of.project_name()

        if (source_package_name != project_name):
            message = ("Project name ({}) doesn't align with the source "
                       "package name ({})".format(project_name,
                                                  source_package_name))
            logging.warning(message)

        silo_state.append_mp(source_package_name, mp.self_link)

        log_value_of.source_packages_seen()
        if (source_package_name not in source_packages_seen):
            source_packages_seen.add(source_package_name)
            # check that any of those components are part of any current
            # transaction
            proj = mp.target_branch.project.name
            if not SiloState.is_project_unique(dest, series, proj, siloname):
                one_project_already_in_use = True

    # check that additional sources are not as well already in use
    for source in silo_state.sources:
        log_value_of.source()
        log_value_of.source_packages_seen()
        if (source not in source_packages_seen):
            source_packages_seen.add(source)
            if not SiloState.is_project_unique(dest, series, source, siloname):
                one_project_already_in_use = True

    if env.RECONFIGURE_SILO == env.RESTRICTED == 'true':
        for source in source_packages_seen:
            if source not in prev_components:
                message = NEW_COMPONENT_FAILURE.format(source)
                logging.error(message)
                old_state.set_reconfigure_failed(message)
                return 1

    if one_project_already_in_use:
        message = ("One or more projects are already in use for the same "
                   "destination and series in another silo (see above).")
        if env.IGNORECONFLICTS == 'true':
            logging.warning(
                message + " However, the ignore flag was set. Please inform "
                "both silos owners directly.")
        else:
            logging.error(message)
            return 1

    # check that all MPs for a source are targetting the same branch
    logging.info(
        "Checking that all target branch for the same source in mps are "
        "targetting the same branches")
    all_same_dest = True
    mps = silo_state.mps
    for source in mps:
        first_mp = None
        for mp_link in mps[source]:
            mp = lp.load(mp_link)
            if not first_mp:
                first_mp = mp
            else:
                current_branch_link = mp.target_branch.web_link
                if current_branch_link != first_mp.target_branch.web_link:
                    logging.error(
                        "{} is targetting {} while {} is targetting "
                        "{}".format(
                            mp.web_link,
                            current_branch_link,
                            first_mp.web_link,
                            first_mp.target_branch.web_link))
                    all_same_dest = False

    if not all_same_dest:
        logging.error(
            "Not all MP targetting the same source package are targetting "
            "the same branch. Please recheck your configuration.")
        return 1

    ppa_uri = SILO_BUILDPPA_SCHEME.format(
        env.DISTRIBUTION, siloname.split('/')[-1])

    try:
        ppa = lp.get_ppa(ppa_uri)
    except lazr.restfulclient.errors.NotFound:
        logging.error("PPA {} not found".format(ppa_uri))
        return 1
    silo_state.ppa = ppa.self_link

    logging.info('Assigned to {}, PPA is {}'.format(siloname, ppa.web_link))

    # if reconfigure, ensure we remove all projects file and packages from the
    # ppa that are not to take into account anymore
    if env.RECONFIGURE_SILO == 'true':
        os.chdir(silo_path)
        clean_removed_sources(silo_state.all_projects, prev_components)

    silo_state.set_ready()
    silo_state.save_config()
    return 0


run_script(__name__, __doc__, main)
