#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2014 Canonical
#
# Authors:
#  Robert Bruce Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,b
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Publish Job

Environment variables:

ACK_PACKAGING
    Publish packaging changes. Only done after manual reviews of the diff.

PUBLISHER
    Launchpad nickname of the person publishing that package.

BUILD_USER_ID
    This provides a default value if PUBLISHER is unset (Jenkins sets this one)

IGNORE_VERSIONDESTINATION
    Ignore if the latest version in archive doesn't match what was prepared.

ALLOW_UNAPPROVED
    Allow publishing even if there are branches that are not approved yet.

DEBUG
    Enable debug infos
"""


import logging
import glob
import sys
import os

from xml.sax.saxutils import escape
from collections import defaultdict

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from citrain.watch_ppa import find_ppa_sources
from cupstream2distro import branchhandling, packagemanager
from cupstream2distro.errors import CITrainError, PublishError
from cupstream2distro.silomanager import SiloState
from cupstream2distro.launchpadmanager import lp
from cupstream2distro.project import DotProject
from cupstream2distro.settings import (
    PUBLISHER_ARTEFACTS_FILENAME,
    PROPOSED_URL,
)
from cupstream2distro.utils import (
    SILO_DIR,
    env,
    ignored,
    log_value_of,
    os_path_join_safe,
    run_script,
)


# FIXME Set this to the correct utopic-devel PPA when that feature is enabled.
# UTOPIC_DEVEL_PPA = """\
# https://api.launchpad.net/devel/~ci-train-ppa-service/\
# +archive/ubuntu/staging-000"""
UNSTABLE_XML_STRING = """\
<testsuite errors="0" failures="{}" name="" tests="1" time="0.1">
  <testcase classname="MarkUnstable" name="Publisher" time="0.0">{}
  </testcase>
</testsuite>
"""
FAILURE_TEMPLATE = """
    <failure type="exception">{}</failure>"""


class PackageList:
    FIELDS = (
        'src_ppa',
        'src_pocket',
        'src_series',
        'target_pocket',
        'target_series',
        'source_name',
        'version',
        'version_in_dest',
        'publisher_nick',
        'target_distribution',
    )
    # '{src_ppa}\t{src_pocket}\t...'
    TEMPLATE = '\t'.join('{{{}}}'.format(f) for f in FIELDS) + '\n'
    PREFIX = 'packagelist_rsync_'
    FILENAME_FORMAT = PREFIX + '{}-{}'
    RSYNC_PATTERN = 'rsync://{{RSYNCSVR}}/cu2d_out/{}*'.format(PREFIX)

    def __init__(self):
        self.lines = []
        self.silo = env.SILONAME.split('/')[-1]

    def publish_source(self, source):
        """Publish one source package."""
        name = source.source_package_name
        self.lines.append(self.TEMPLATE.format(
            src_ppa='ci-train-ppa-service/' + env.SILONAME,
            src_pocket=source.pocket,
            src_series=env.SERIES,
            target_pocket='Proposed',
            target_series=env.SERIES,
            source_name=name,
            version=source.source_package_version,
            version_in_dest=DotProject(name).dest_current_version,
            publisher_nick=env.PUBLISHER,
            target_distribution=env.DISTRO,
        ))

    def write(self):
        filename = self.FILENAME_FORMAT.format(self.silo, env.SERIES)
        filepath = os_path_join_safe(SILO_DIR(), filename)
        logging.info('Writing packagelist for rsync.')
        log_value_of.filepath()
        logging.debug('\n' + ''.join(self.lines))
        with open(filepath, 'w') as f:
            f.writelines(self.lines)


def get_distro_and_series(series_link):
    """Parse the distro and series name from a series link."""
    env.DISTRO, env.SERIES = series_link.split('/')[-2:]
    logging.info('Found distro {} and series {} in link {}.'.format(
        env.DISTRO, env.SERIES, series_link))


def prevent_duplicate_publishing(is_publishing):
    """Prevent the publish job from being run simultaneously."""
    if is_publishing:
        raise PublishError('The publisher job is already running. Skipping.')
    with ignored(OSError):
        os.remove(os_path_join_safe(SILO_DIR(), PUBLISHER_ARTEFACTS_FILENAME))


def check_versions_at_destination(dest, source_names):
    """Prevent the publish job from clobbering manual archive uploads."""
    msg = 'Version of {} at dest changed from {} to {} since packages built.'
    get_dest_version = packagemanager.get_current_version_for_series
    if env.IGNORE_VERSIONDESTINATION != 'true':
        logging.info('Checking versions at destination are what we expect...')
        if lp.is_distro_archive(dest):
            for source in source_names:
                version = get_dest_version(source, env.SERIES, dest=dest)
                supposed = DotProject(source).dest_current_version
                if supposed and version != supposed:
                    raise PublishError(msg.format(source, supposed, version))
    else:
        logging.warning('Honoring your request to ignore dest versions.')


def check_for_unapproved(mps, lp_load):
    """Potentially raise an error if we have some unapproved merges."""
    if env.ALLOW_UNAPPROVED != 'true':
        logging.info('Checking that all merges are approved...')
        unapproved = []
        for source in mps:
            for mp_link in mps[source]:
                mp = lp_load(mp_link)
                if mp.queue_status not in ('Approved', 'Merged'):
                    unapproved.append(mp.web_link)
        if unapproved:
            logging.error('These merges must be approved before publishing:')
            for mp in unapproved:
                logging.error(mp)
            raise PublishError('Some merges are unapproved.')
    else:
        logging.warning('Honoring your request to publish unapproved MPs.')


def confirm_publisher_identity(lp_people):
    """Identify the person who is running this job."""
    nick = env.PUBLISHER or env.BUILD_USER_ID
    try:
        logging.debug('lp.people["{}"]: {}'.format(nick, lp_people[nick]))
    except KeyError:
        raise PublishError('{} is not a valid launchpad user id.'.format(nick))
    logging.info('Publishing the silo as {}.'.format(nick))
    env.PUBLISHER = nick


def ack_packaging_diffs():
    """Prevent publishing silos with packaging diffs without an ACK."""
    if env.ACK_PACKAGING != 'true':
        fileglob = DotProject('*').get_diff_path('*')
        diffs = glob.glob(fileglob)
        for diff in diffs:
            basename = os.path.basename(diff)
            logging.error('Packaging changes found: {}'.format(basename))
        if diffs:
            generate_xml_artifacts(diffs)
            raise PublishError('Packaging changes need manual ACKing.')
    else:
        logging.warning('Honoring your request to ACK packaging changes.')


# FIXME: Call this during the build job, so we have access to it sooner.
def push_branches(mps, lp_load):
    """Push all local branches to Launchpad."""
    for source in sorted(mps):
        project_name = lp_load(mps[source][0]).target_branch.project.name
        branch = os_path_join_safe(SILO_DIR(), source)
        proposed = PROPOSED_URL.format(project_name, env.DISTRO, env.SERIES)
        logging.info('Pushing {} to {}'.format(source, proposed))
        if not branchhandling.push_to_branch(branch, proposed, overwrite=True):
            raise PublishError('Failed pushing to {}'.format(proposed))


def initiate_publication(dest, ppa, silo_state_set_package_version_list):
    """Determine whether to use copyPackage or rsync packagelists."""
    sources = set()
    versions = defaultdict(set)
    find_ppa_sources(ppa, sources, versions)
    log_value_of.sources()
    log_value_of.versions()
    if lp.is_distro_archive(dest):
        write_rsync_file(sources, dest)
    else:
        copy_to_dest(sources, dest)
    silo_state_set_package_version_list(versions)


def write_rsync_file(sources, dest):
    """Write rsync packagelist file necessary for publishing to distro."""
    packagelist = PackageList()
    for source in sources:
        if source.distro_series_link.endswith(env.SERIES):
            published_version = packagemanager.get_current_version_for_series(
                source.source_package_name, env.SERIES, dest=dest)
            if published_version != source.source_package_version:
                packagelist.publish_source(source)
                DotProject(source.source_package_name).mark_as_published(
                    source.source_package_version)
            else:
                logging.info('{} {} found at destination, skipping.'.format(
                    source.source_package_name, source.source_package_version))
        # TODO: Enable this utopic+vivid hack if necessary.
        # elif env.SERIES == 'vivid':
        #     if source.distro_series_link.endswith('utopic'):
        #         copy_to_dest([source], dest=lp_load(UTOPIC_DEVEL_PPA))
    packagelist.write()


def copy_to_dest(sources, dest, target_pocket='Release'):
    """Binary-copy a package to the given dest archive."""
    logging.info('Copying sources directly to destination...')
    for source in sources:
        logging.info('Copying {} {} to {}.'.format(
            source.source_package_name, source.source_package_version, dest))
        series = source.distro_series_link.split('/')[-1]
        dest.copyPackage(
            from_archive=source.archive_link,
            from_pocket=source.pocket,
            from_series=series,
            include_binaries=True,
            to_pocket=target_pocket,
            to_series=series,
            source_name=source.source_package_name,
            version=source.source_package_version)


def generate_xml_artifacts(details=[]):
    """Generate a fake test name xml for marking the build as unstable."""
    failure = ''.join(FAILURE_TEMPLATE.format(escape(d)) for d in details)
    path = os_path_join_safe(SILO_DIR(), PUBLISHER_ARTEFACTS_FILENAME)
    with open(path, 'w') as f:
        f.write(UNSTABLE_XML_STRING.format(1 if details else 0, failure))


def main():
    """Conduct a beautiful orchestra of publication."""
    silo_state = SiloState.safe_load(env.SILONAME)
    dest = lp.load(silo_state.dest)
    ppa = lp.load(silo_state.ppa)
    try:
        prevent_duplicate_publishing(silo_state.is_publishing)
        get_distro_and_series(silo_state.series)
        silo_state.set_publishing()
        silo_state.save_config()
        check_versions_at_destination(dest, silo_state.all_projects)
        check_for_unapproved(silo_state.mps, lp.load)
        confirm_publisher_identity(lp.people)
        ack_packaging_diffs()
        push_branches(silo_state.mps, lp.load)
        initiate_publication(dest, ppa, silo_state.set_package_version_list)
        generate_xml_artifacts()
        silo_state.set_published()
        silo_state.set_migrating()
    except CITrainError as e:
        logging.error(e.message)
        silo_state.set_publish_failed(e.message)
        return 1
    finally:
        silo_state.save_config()
    return 0


run_script(__name__, __doc__, main)
