#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2012-2014 Canonical
#
# Authors:
#  Didier Roche
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Setup CI Train Job

Create/Update the configuration of the Jenkins jobs for CI Train

Environment variables:

DEBUG
    Enable debug infos
"""

import logging
import sys
import os

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from cupstream2distro.utils import (
    ignored,
    log_value_of,
    os_path_join_safe,
    run_script,
)
from cupstream2distro.settings import (
    ALL_SILO_NAMES,
    CITRAIN_BINDIR,
    SILOS_DIR,
    SILO_RSYNCDIR,
)


JOBS_DIR = os.path.expanduser(os_path_join_safe('~', 'jobs'))
TEMPLATE_DIR = os_path_join_safe(
    os.path.dirname(__file__), 'jenkins-templates')


def template(template_name):
    return template_name + '.xml.tmpl'


def load_template(template_name, context):
    with open(os_path_join_safe(TEMPLATE_DIR, template_name)) as templ:
        return templ.read().format(**context)


def setup_job(jobname, template_name=None, context={}):
    """Populate templates and then write jenkins jobs to disk."""
    logging.info('Configuring job {}.'.format(jobname))
    log_value_of.jobname()
    log_value_of.template_name()
    context['BINDIR'] = CITRAIN_BINDIR
    context['RSYNC_OUTDIR'] = SILO_RSYNCDIR
    jobdir = os_path_join_safe(JOBS_DIR, jobname)
    config_file = os_path_join_safe(jobdir, 'config.xml')
    with ignored(OSError):
        os.makedirs(jobdir)
    with open(config_file, 'w') as f:
        f.write(load_template(template_name or template(jobname), context))


def setup_silo(siloname):
    """Setup a silo jenkins job."""
    context = dict(SILOS_DIR=SILOS_DIR, SILO_NAME=siloname)
    siloname = siloname.replace('/', '-')
    steps = ('reconfigure', 'build', 'publish', 'merge-clean')
    for i, job in enumerate(steps, start=0):
        setup_job('{}-{}-{}'.format(siloname, i, job), template(job), context)


def main():
    for siloname in ALL_SILO_NAMES:
        setup_silo(siloname)
    setup_job('prepare-silo')
    setup_job('check-publication-migration')
    setup_job('upgrade-chroot')
    setup_job('apt-get-clean')
    setup_job('pbuilder-clean')
    return 0


run_script(__name__, __doc__, main)
