#!/usr/bin/python
# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Robert Bruce Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

"""CI Train Watch PPA Script

Watch for published package in a ppa

Environment variables:

DEBUG
    Enable debug infos
"""

import logging
import os
import sys
import time

sys.path.append(os.path.dirname(os.path.dirname(__file__)))

from collections import defaultdict

from cupstream2distro import packagemanager
from cupstream2distro.utils import env, find_all, log_value_of, run_script
from cupstream2distro.launchpadmanager import lp
from cupstream2distro.silomanager import SiloState
from cupstream2distro.project import DotProject
from cupstream2distro.settings import (
    PPA_TIMEOUT,
    TIME_BETWEEN_PPA_CHECKS,
)


COMMA = ', '
BUILD_FAILURES = (
    'Build for superseded Source',
    'Cancelled build',
    'Chroot problem',
    'Failed to build',
    'Failed to upload',
)


silo_state = SiloState.safe_load(env.SILONAME)


# FIXME: This class should be named something like ArchWatcher but inherited
# a lowercase name by replacing a function for the same purpose.
class watch_arch:
    def __init__(self, arches):
        """Define what arches we care about for this source package."""
        self.arches = arches

    def __call__(self, build):
        """Return true only if the given build is for an arch we care about.

        If self.arches is None, that represents an Arch:all package which means
        we should observe whatever arch we happen to find in the PPA.
        """
        return build.arch_tag in self.arches or not self.arches


def find_supported_arches(dest, source_name):
    """Determine what arches we should watch by checking the destination.

    Not all packages we build support all arches. If the latest version of a
    given package does not have binaries for a given arch at the destination
    archive, then we shouldn't care about build failures for that arch because
    it does not represent a regression. However if there is a binary at the
    destination but the build for that arch fails here, that represents a
    regression and must result in failing the build job. Thus we must determine
    what arches at the destination archive contain builds for this source
    package.
    """
    series = lp.load(silo_state.series)
    source_collection = packagemanager.sort_by_date_created(
        dest.getPublishedSources(
            exact_match=True,
            source_name=source_name,
            distro_series=series,
            status='Published',
        ))
    if not source_collection:
        return
    for binary in source_collection[0].getPublishedBinaries():
        # Must check architecture_specific field to filter out false positives
        # caused by Arch:all packages showing up even on arches that we don't
        # expect builds to succeed on.
        if binary.architecture_specific:
            yield binary.distro_arch_series.architecture_tag


def find_all_uploaded():
    """Return all source package names and versions uploaded to the ppa."""
    versions = defaultdict(set)
    for uploaded in find_all('_source.ppa.upload'):
        name = os.path.basename(uploaded).split('_')[0]
        # The version in the filename lacks the Epoch so we must use DotProject
        versions[name].add(DotProject(name).packaging_version)
    return dict(versions)


def find_ppa_sources(ppa, sources, versions):
    """Populate sources set with all Published and Pending sources in a PPA.

    Note this modifies the arguments in place.
    """
    logging.info('Checking PPA for source packages...')
    sources.clear() and versions.clear()
    for status in ('Published', 'Pending'):
        sources.update(list(ppa.getPublishedSources(status=status)))
    for source in sources:
        versions[source.source_package_name].add(source.source_package_version)


def compare_versions(ppa_versions, uploaded, manual_sources):
    """Make sure everything we uploaded is in the PPA."""
    log_value_of.ppa_versions()
    for pkg in uploaded:
        if pkg not in ppa_versions:
            return False
        if uploaded[pkg] - ppa_versions[pkg]:
            return False
    if set(manual_sources) - set(ppa_versions):
        return False
    # It's ok if there's elements in ppa_versions not in uploaded because
    # that will be the case for manual source uploads.
    return True


def pretty_mismatch(ppa_versions, uploaded, manual_sources):
    """Return a pretty error message in case of version mismatches."""
    message = 'Versions in PPA did not match our expectations.'
    template = '\nExpected: {pkg} {upload}\nFound:    {pkg} {found}'
    missing = (set(uploaded) | set(manual_sources)) - set(ppa_versions)
    if missing:
        message += '\nMissing any version of: {}'.format(
            COMMA.join(sorted(missing)))
    for pkg in sorted(uploaded):
        if pkg not in ppa_versions:
            continue  # Already covered above
        missing = uploaded[pkg] - ppa_versions[pkg]
        if missing:
            message += template.format(
                pkg=pkg,
                upload=COMMA.join(sorted(uploaded[pkg])),
                found=COMMA.join(sorted(ppa_versions[pkg])))
    return message


def wait_for_ppa(uploaded):
    """Poll the PPA until it contains all the packages we expect to find."""
    ppa = lp.load(silo_state.ppa)
    started = time.time()
    sources = set()
    ppa_versions = defaultdict(set)
    while not compare_versions(ppa_versions, uploaded, silo_state.sources):
        if (time.time() - started) > PPA_TIMEOUT:
            log_value_of.ppa_versions()
            logging.error(
                pretty_mismatch(ppa_versions, uploaded, silo_state.sources))
            return
        time.sleep(TIME_BETWEEN_PPA_CHECKS)
        find_ppa_sources(ppa, sources, ppa_versions)
    return sources


def main():
    dest = lp.load(silo_state.dest)

    uploaded = find_all_uploaded()
    log_value_of.uploaded()

    sources = wait_for_ppa(uploaded)
    if not sources:
        return 1

    failed = set()
    while sources:
        for source in sources.copy():
            name = source.source_package_name
            version = source.source_package_version
            series = source.distro_series_link.split('/')[-1]
            if name in uploaded and version not in uploaded[name]:
                # Launchpad often reports superceded builds, skip those.
                sources.discard(source)
                continue
            arches = set(find_supported_arches(dest, name))
            log_value_of.arches('Supported arches at dest')
            builds = list(filter(watch_arch(arches), source.getBuilds()))
            logging.info('Checking {}:'.format(source.display_name))
            success = 0
            for build in builds:
                logging.info('{}: {}'.format(build.arch_tag, build.buildstate))
                if build.buildstate in BUILD_FAILURES:
                    failed.add(source)
                    sources.discard(source)
                    logging.error(build.build_log_url)
                    continue
                if build.buildstate == 'Successfully built':
                    success += 1
            if success == len(builds):
                sources.discard(source)
                prev_version = packagemanager.get_current_version_for_series(
                    name, series, dest=dest)
                DotProject.soft_save(name, dest=prev_version, ours=version)
        if sources:
            time.sleep(TIME_BETWEEN_PPA_CHECKS)
    log_value_of.failed()
    return len(failed)


run_script(__name__, __doc__, main)
