# -*- coding: utf-8 -*-
# Copyright (C) 2012 Canonical
#
# Authors:
#  Didier Roche
#  Łukasz 'sil2100' Zemczak
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.cat
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from __future__ import unicode_literals

from launchpadlib.launchpad import Launchpad
from lazr.restfulclient.errors import BadRequest, ServerError
import logging
import os

from cupstream2distro.utils import env, os_path_join_safe, singleton
from cupstream2distro.settings import (
    COMMON_LAUNCHPAD_CACHE_DIR
)


CREDS_PATH = os.path.expanduser(
    os_path_join_safe('~', '.launchpad.credentials'))


@singleton
class lp:
    """Trigger connections to launchpadlib only when needed."""
    _instance = None

    @property
    def _launchpad(self):
        """Connect to launchpad if we haven't already."""
        if not self._instance:
            self._instance = get_launchpad()
        return self._instance

    def __getattr__(self, attr):
        """Wrap launchpadlib so tightly you can't tell the difference."""
        return getattr(self._launchpad, attr)

    def get_ppa(self, ppa_name):
        """Return a launchpad ppa."""
        parts = ppa_name.split('/')
        distro = self.distributions[parts[1] if len(parts) >= 3 else 'ubuntu']
        return self.people[parts[0]].getPPAByName(
            name=parts[-1], distribution=distro)

    def is_distro_archive(self, series):
        """Return if series is the given distribution's main archive."""
        return series.distribution.main_archive_link == series.self_link

    def is_link_distro_archive(self, series_link):
        """Return if series_link is the given distribution's main archive."""
        return self.is_distro_archive(self.load(series_link))

    def load_url(self, url):
        """Return a lp resource from a launchpad url."""
        return lp.load(self.resource_type_link.replace('/#service-root', '') +
                       url.split('launchpad.net')[1])

    def get_series(self, series_name, dest=None, distro_name='ubuntu'):
        """Return the launchpad object for the requested series name."""
        if dest is not None:
            distro = self.load(dest.distribution_link)
        else:
            distro = self.distributions[distro_name]
        return distro.getSeries(name_or_version=series_name)

    def open_bugs_for_source(self, bugs, source, series, dest):
        msg = 'Failed to sync upstream/downstream bugs for bug #{}.'
        distro = self.load(dest.distribution_link)
        dest = distro if distro.current_series.name == series.name else series
        package = dest.getSourcePackage(name=source)
        for bug_num in bugs:
            try:
                bug = self.bugs[bug_num]
                bug.addTask(target=package)
                bug.lp_save()
            except (KeyError, BadRequest, ServerError):
                # ignore non existing or available bugs
                logging.info(msg.format(bug_num))


def get_launchpad(use_staging=False):
    """Get THE Launchpad."""
    # as launchpadlib isn't multiproc, fiddling the cache dir if any
    launchpadlib_dir = None
    if env.JOB_NAME:
        launchpadlib_dir = os_path_join_safe(
            COMMON_LAUNCHPAD_CACHE_DIR, env.JOB_NAME)
    login_args = dict(
        application_name='cupstream2distro',
        service_root='staging' if use_staging else 'production',
        allow_access_levels=['WRITE_PRIVATE'],
        version='devel',  # Need devel for copyPackage
        launchpadlib_dir=launchpadlib_dir,
    )
    if os.path.exists(CREDS_PATH):
        login_args['credentials_file'] = CREDS_PATH
    return Launchpad.login_with(**login_args)
