# -*- coding: utf-8 -*-
# Copyright (C) 2012-2014 Canonical
#
# Authors:
#  Didier Roche
#  Rodney Dawes
#  Łukasz 'sil2100' Zemczak
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import datetime
import fileinput
import logging
import os
import re
import shutil
import sys
import glob

from six.moves.urllib.parse import unquote
from six.moves.urllib.request import urlretrieve
from subprocess import Popen, PIPE
from operator import attrgetter

from cupstream2distro.launchpadmanager import lp
from cupstream2distro import settings
from cupstream2distro.utils import (
    SILO_DIR,
    ignored,
    log_value_of,
    os_path_join_safe,
    std_log,
)


def _filter_out_invalidated(src):
    """Returns True if package is published or pending, otherwise False."""
    return src.status in ('Published', 'Pending')


def sort_by_date_created(sources, all_packages=False):
    # we also filter out any non-published or not-pending packages by default
    if not all_packages:
        filtered_sources = list(filter(_filter_out_invalidated, sources))
    else:
        filtered_sources = sources
    return sorted(
        filtered_sources, key=attrgetter("date_created"), reverse=True)


def get_current_version_for_series(source_package_name,
                                   series_name,
                                   ppa_name=None,
                                   dest=None,
                                   all_packages=False):
    """Get current version for a package name in that series."""
    if not dest:
        dest = lp.get_ppa(ppa_name) if ppa_name else lp.main_archive
    series = lp.get_series(series_name, dest=dest)
    source_collection = dest.getPublishedSources(
        exact_match=True,
        source_name=source_package_name,
        distro_series=series)
    try:
        return sort_by_date_created(
            source_collection,
            all_packages)[0].source_package_version
    # was never in the dest, set the lowest possible version
    except IndexError:
        return "0"


def is_upstream_version_for_series_in_dest(source_package_name,
                                           version,
                                           series,
                                           dest,
                                           pocket="Release",
                                           all_packages=False):
    """Look for upstream version in dest.

    Return if dest and series has the given package name with the selected
    upstream version part (not counting the -0ubuntu1 parts).
    """
    # XXX: this method might proove not to be very useful
    pkgs = dest.getPublishedSources(
        exact_match=True,
        source_name=source_package_name,
        distro_series=series,
        pocket=pocket)
    upstream_version = get_upstream_version(version)
    if not all_packages:
        pkgs = list(filter(_filter_out_invalidated, pkgs))
    if "-" in version:
        upstream_version += "-"
    interesting_pkgs = [
        x for x in pkgs if upstream_version in x.source_package_version]

    return len(interesting_pkgs) > 0


def is_version_for_series_in_dest(source_package_name,
                                  version,
                                  series,
                                  dest,
                                  pocket="Release"):
    """Return if version for a package name in that series is in dest."""
    return dest.getPublishedSources(
        exact_match=True,
        source_name=source_package_name,
        version=version,
        distro_series=series,
        pocket=pocket).total_size > 0


def is_version_in_queue(source_package_name, version, dest_serie, queue):
    """Return if version for a package name in that series is in dest."""
    return dest_serie.getPackageUploads(
        exact_match=True,
        name=source_package_name,
        version=version,
        status=queue).total_size > 0


def is_version1_higher_than_version2(version1, version2):
    """Return if version1 is higher than version2."""
    process = Popen(
        ["dpkg", "--compare-versions", version1, 'gt', version2],
        stdout=PIPE, stderr=PIPE, cwd=SILO_DIR())
    std_log(process.communicate())
    return process.returncode == 0


def is_version_in_changelog(version, f):
    """Return True if the version is in the changelog and not UNRELEASED."""
    if version == '0':
        return True
    version = ' ({}) '.format(version)
    for line in f:
        if version in line and 'UNRELEASED; urgency=' not in line:
            return True
    return False


def get_latest_upstream_bzr_rev(f, dest_ppa=None):
    """Report latest bzr rev in the file.

    If dest_ppa, first try to fetch the dest ppa tag.
    Otherwise, fallback to first distro version.
    """
    distro_regex = re.compile("{} (\d+)".format(settings.REV_STRING_FORMAT))
    destppa_regexp = re.compile(
        "{} (\d+) \(ppa:{}\)".format(settings.REV_STRING_FORMAT, dest_ppa))
    distro_rev = None
    candidate_destppa_rev = None
    candidate_distro_rev = None
    destppa_element_found = False

    # handle marker spread on two lines
    end_of_line_regexp = re.compile(" *(.*\))")
    previous_line = None

    for line in f:
        line = line[:-1]

        if previous_line:
            try:
                line = previous_line + end_of_line_regexp.findall(line)[0]
            except IndexError:
                None
            previous_line = None

        if dest_ppa:
            try:
                candidate_destppa_rev = int(destppa_regexp.findall(line)[0])
                destppa_element_found = True
            except IndexError:
                destppa_element_found = False
        if (not distro_rev and
                not destppa_element_found and
                "(ppa:" not in line):
            try:
                candidate_distro_rev = int(distro_regex.findall(line)[0])
                distro_element_found = True
            except IndexError:
                distro_element_found = False

        # try to catchup next line if we have a marker start without anything
        # found
        if settings.REV_STRING_FORMAT in line and (
                dest_ppa and not destppa_element_found
                ) and not distro_element_found:
            previous_line = line

        if line.startswith(" -- "):
            # first grab the dest ppa
            if candidate_destppa_rev:
                return candidate_destppa_rev
            if not distro_rev and candidate_distro_rev:
                distro_rev = candidate_distro_rev
            if not dest_ppa and distro_rev:
                return distro_rev

    # we didn't find any dest ppa result but there is a distro_rev one
    if dest_ppa and distro_rev:
        return distro_rev

    # we force a bootstrap commit for new components
    return 0


def list_packages_info_in_str(packages_set):
    """Return the packages info in a string."""
    results = []
    for package in packages_set:
        results.append("{} ({})".format(package.source_name, package.version))
    return " ".join(results)


def get_packaging_version(bzr_repo):
    """Get current packaging version."""
    instance = Popen(
        ['dpkg-parsechangelog'],
        stdout=PIPE, stderr=PIPE, cwd=bzr_repo)
    (stdout, stderr) = instance.communicate()
    if instance.returncode != 0:
        raise Exception(stderr.decode("utf-8").strip())
    expr = re.compile("Version: (.*)")
    for line in stdout.splitlines():
        packaging_version = expr.findall(line)
        if packaging_version:
            return packaging_version[0]

    raise Exception(
        "Didn't find any Version in the package: {}".format(stdout))


def get_upstream_version(version, remove_epoch=True):
    """Return upstream version."""
    if remove_epoch:
        # remove epoch is there is one
        version = version.split(':')[-1]
    splitted_version = version.split('-')
    if len(splitted_version) > 1:
        # we don't want the ubuntu or debian version (it's not in the source
        # package name)
        splitted_version = splitted_version[:-1]
    return '-'.join(splitted_version)


def get_source_package_from_dest(source_package_name,
                                 dest_archive,
                                 dest_current_version,
                                 series_name,
                                 download_dir='ubuntu'):
    """Download and return a path containing the current dest version.

    None if this package was never published to dest archive.
    """
    if dest_current_version == "0":
        logging.info(
            "This package was never released to the destination archive, "
            "don't return downloaded source")
        return None

    logging.info(
        "Grab code for {} ({}) from {} to the '{}' directory".format(
            source_package_name,
            dest_current_version,
            series_name,
            download_dir))
    source_package_download_dir = os_path_join_safe(
        SILO_DIR(), download_dir, source_package_name)
    series = lp.get_series(series_name, dest=dest_archive)
    with ignored(OSError):
        os.makedirs(source_package_download_dir)

    try:
        sourcepkg = sort_by_date_created(
            dest_archive.getPublishedSources(
                exact_match=True,
                source_name=source_package_name,
                distro_series=series,
                version=dest_current_version),
            True)[0]
    except IndexError:
        raise Exception("Couldn't get in the destination the expected version")
    logging.info(
        'Downloading %s version %s', source_package_name, dest_current_version)
    for url in sourcepkg.sourceFileUrls():
        filename = unquote(url.split('/')[-1])
        filepath = os_path_join_safe(source_package_download_dir, filename)
        urlretrieve(url, filepath)
    dscs = glob.glob(os_path_join_safe(
        source_package_download_dir, '*{}.dsc'.format(
            dest_current_version.split(':')[-1])))
    instance = Popen(
        ['dpkg-source', '-x'] + dscs,
        stdout=PIPE, stderr=PIPE, cwd=source_package_download_dir)
    (stdout, stderr) = instance.communicate()
    if instance.returncode != 0:
        raise Exception(stderr.decode("utf-8").strip())

    # check the dir exist
    # remove epoch is there is one
    splitted_version = dest_current_version.split(':')[-1].split('-')
    if len(splitted_version) > 1:
        # we don't want the ubuntu or debian version (it's not in the source
        # package name)
        splitted_version = splitted_version[:-1]
    version_for_source_file = '-'.join(splitted_version)
    source_dirname = os_path_join_safe(
        source_package_download_dir, '{}-{}'.format(
            source_package_name, version_for_source_file))
    if not os.path.isdir(source_dirname):
        raise Exception(source_dirname + ': Directory not found')
    return source_dirname


def is_new_content_relevant(bzr_repo, dest_version_source):
    """Return True if a new snapshot is needed.

    dest_version_source can be None if no released version was done before.
    """
    # Always release new source packages regardless of the other criteria
    if not dest_version_source:
        return True

    # now check the relevance of the committed changes compared to the version
    # in the repository (if any)
    diffinstance = Popen(
        ['diff', '-Nrup', '.', dest_version_source],
        stdout=PIPE, stderr=PIPE, cwd=bzr_repo)
    filterinstance = Popen(
        ['filterdiff', '--clean', '-x', '*po', '-x', '*pot', '-x',
            '*local-options'],
        stdin=diffinstance.stdout, stdout=PIPE, stderr=PIPE, cwd=bzr_repo)
    lsdiffinstance = Popen(
        ['lsdiff'],
        stdin=filterinstance.stdout, stdout=PIPE, stderr=PIPE, cwd=bzr_repo)
    (relevant_changes, err) = Popen(
        ['grep', '-Ev', '.bzr|.pc'],
        stdin=lsdiffinstance.stdout, stdout=PIPE, stderr=PIPE, cwd=bzr_repo
    ).communicate()

    # detect if the only change is a Vcs* target changes (with or without
    # changelog edit). We won't release in that case
    number_of_changed_files = relevant_changes.count("\n")
    if ((number_of_changed_files == 1 and
            "debian/control" in relevant_changes) or
       (number_of_changed_files == 2 and
           "debian/control" in relevant_changes and
           "debian/changelog" in relevant_changes)):
        control = os_path_join_safe(bzr_repo, 'debian', 'control')
        dest_control = os_path_join_safe(
            dest_version_source, "debian", "control")
        (results, err) = Popen(
            ['diff', control, dest_control],
            stdout=PIPE, stderr=PIPE, cwd=bzr_repo).communicate()
        for diff_line in results.split('\n'):
            if diff_line.startswith("< ") or diff_line.startswith("> "):
                if (not diff_line[2:].startswith("Vcs-") and
                        not diff_line[2:].startswith("#")):
                    return True
        return False

    log_value_of.relevant_changes()

    return (relevant_changes != '')


def is_relevant(newdsc_path, dest_version_source):
    """Check if the generated source diff is different from previous one."""
    with ignored(OSError):
        os.makedirs(os_path_join_safe(SILO_DIR(), 'generated'))
    extracted_generated_source = os_path_join_safe(
        SILO_DIR(), 'generated', os.path.basename(newdsc_path).split('_')[0])
    with ignored(OSError):
        log_value_of.extracted_generated_source('Deleting!')
        shutil.rmtree(extracted_generated_source)

    # remove epoch is there is one
    process = Popen(
        ['dpkg-source', '-x', newdsc_path, extracted_generated_source],
        stdout=PIPE, stderr=PIPE, cwd=SILO_DIR())
    std_log(process.communicate())
    if process.returncode != 0:
        raise Exception("dpkg-source command returned an error.")

    # now check the relevance of the committed changes compared to the version
    # in the repository (if any)
    diffinstance = Popen(
        ['diff', '-Nrup', extracted_generated_source, dest_version_source],
        stdout=PIPE, stderr=PIPE, cwd=SILO_DIR())
    (diff, err) = Popen(
        ['filterdiff', '--clean', '-x', '*po', '-x', '*pot', '-x',
            '*local-options'],
        stdin=diffinstance.stdout, stdout=PIPE, stderr=PIPE, cwd=SILO_DIR()
    ).communicate()

    # there is no important diff if the diff only contains 12 lines,
    # corresponding to "Automatic daily release" marker in debian/changelog
    if (diff.count('\n') <= 12):
        return False
    return True


def _packaging_changes_between_dsc(oldsource_dsc, newsource_dsc):
    """Return if there has been a packaging change between two dsc files.

    We ignore the changelog only changes.
    """
    if not oldsource_dsc:
        return True
    if not os.path.isfile(oldsource_dsc) or not os.path.isfile(newsource_dsc):
        raise Exception(
            "{} or {} doesn't not exist, can't create a diff".format(
                oldsource_dsc, newsource_dsc))
    diffinstance = Popen(
        ['debdiff', oldsource_dsc, newsource_dsc],
        stdout=PIPE, stderr=PIPE, cwd=SILO_DIR())
    filterinstance = Popen(
        ['filterdiff', '--clean', '-i', '*debian/*', '-x', '*changelog'],
        stdin=diffinstance.stdout, stdout=PIPE, stderr=PIPE, cwd=SILO_DIR())
    (change_in_debian, filter_err) = filterinstance.communicate()
    # we can't rely on diffinstance returncode as the signature key is maybe
    # not present and it will exit with 1
    if filterinstance.returncode != 0:
        raise Exception(
            "Error in diff: {}".format(filter_err.decode("utf-8").strip()))
    return(change_in_debian != "")


def generate_diff_between_dsc(diff_filepath, oldsource_dsc, newsource_dsc):
    """Generate a diff in diff_filepath if there is a packaging diff.

    The diff contains autotools files and cmakeries.
    """
    if _packaging_changes_between_dsc(oldsource_dsc, newsource_dsc):
        with open(diff_filepath, "a") as f:
            if not oldsource_dsc:
                f.writelines(
                    "This source is a new package, if the destination is "
                    "ubuntu, please ensure it has been preNEWed by an archive "
                    "admin before publishing that stack.")
                return
            f.write(
                "/!\ Remember that this diff only represents packaging "
                "changes and build tools diff, not the whole content "
                "diff!\n\n")
            diff = Popen(
                ['debdiff', oldsource_dsc, newsource_dsc],
                stdout=PIPE, stderr=PIPE, cwd=SILO_DIR())
            (changes_to_publish, err) = Popen(
                ['filterdiff', '--remove-timestamps', '--clean', '-i',
                    '*setup.py', '-i', '*Makefile.am', '-i', '*configure.*',
                    '-i', '*debian/*', '-i', '*CMakeLists.txt'],
                stdin=diff.stdout, stdout=PIPE, stderr=PIPE, cwd=SILO_DIR()
            ).communicate()

            # look for any modifications that might mean a new binary package
            # has appeared TODO: make better, considering only changes for
            # debian/control
            split_diff = changes_to_publish.split('\n')
            re_replace = re.compile('\+Package: *(.*)')
            new_binary_packages = []

            logging.debug("Looking for new binary packages in the diff")
            for line in split_diff:
                if line.startswith('+Package:'):
                    logging.debug("Found a new binary package")
                    reverse = '-' + line[1:]
                    if reverse in split_diff:
                        logging.debug("It seems to be a move, ignoring")
                        continue
                    new_binary_packages.append(re.sub(re_replace, '\\1', line))
            if new_binary_packages:
                f.write(
                    "(!) Warning! This package seems to add new binary "
                    "packages ({}). Please consult an archive admin before "
                    "proceeding!\n\n".format(', '.join(new_binary_packages)))

            f.write(changes_to_publish)


def check_if_packages_require_twin_upload(sources):
    """Ugly method to prevent us from uploading packages without their twins.

    The most well known ones are the -gles counterparts for some Qt/UITK
    related packages.
    """
    twin_packages = {
        "qtbase-opensource-src": "qtbase-opensource-src-gles",
        "qtdeclarative-opensource-src": "qtdeclarative-opensource-src-gles",
        "qt3d-opensource-src": "qt3d-opensource-src-gles",
        "qtlocation-opensource-src": "qtlocation-opensource-src-gles",
        "qtubuntu": "qtubuntu-gles",
        "ubuntu-ui-toolkit": "ubuntu-ui-toolkit-gles",
        "qtmultimedia-opensource-src": "qtmultimedia-opensource-src-gles",
        "qtmir": "qtmir-gles",
    }

    missing = []
    for source in sources:
        if source in twin_packages:
            # Check if the twin package has been also uploaded
            if twin_packages[source] not in sources:
                missing.append(twin_packages[source])

    return missing


def create_new_packaging_version(base_package_version,
                                 series_version,
                                 destppa=''):
    """Deliver a new packaging version, based on simple rules:

    Version would be <upstream_version>.<series>.<yyyymmdd(.minor)>-0ubuntu1
    if we already have something delivered today, it will be .minor, then,
    .minor+1…

    We append the destination ppa name if we target a dest ppa and
    not distro.
    """
    native_pkg = False

    # handle the special case of RTM
    special_tag = ""
    if series_version == "14.09":
        logging.debug('Appending the ~rtm tag')
        special_tag = "~rtm"
        # XXX: we agreed that to keep the versioning sane, we need to stick to
        # versioning as per the main archive
        series_version = "15.04"

    today_version = datetime.date.today().strftime('%Y%m%d')
    destppa = destppa.replace("-", '.').replace("_", ".").replace("/", ".")
    # bootstrapping mode or direct upload or UNRELEASED for bumping to a new
    # series TRANSITION
    if not ("daily" in base_package_version or "+" in base_package_version):
        # support both 42, 42-0ubuntu1
        upstream_version = base_package_version.split('-')[0]
        # if we have 42ubuntu1 like a wrong native version
        if "ubuntu" in upstream_version:
            upstream_version = upstream_version.split('ubuntu')[0]
    elif "-" not in base_package_version and "+" in base_package_version:
        # extract the day of previous daily upload and bump if already uploaded
        regexp = re.compile("(.*)\+([\d\.]{5})\.(\d{8})\.?([\d]*).*")
        try:
            previous_day = regexp.findall(base_package_version)[0]
            upstream_version = previous_day[0]
            native_pkg = True
            if (previous_day[1] == series_version and
                    previous_day[2] == today_version):
                minor = 1
                if previous_day[3]:  # second upload of the day
                    minor = int(previous_day[3]) + 1
                today_version = "{}.{}".format(today_version, minor)
        except IndexError:
            raise Exception(
                "Unable to get previous day from native version: %s"
                % base_package_version)
    else:
        # extract the day of previous daily upload and bump if already uploaded
        # today
        regexp = re.compile("(.*)\+([\d\.]{5})\.(\d{8})\.?([\d]*).*-.*")
        try:
            previous_day = regexp.findall(base_package_version)[0]
            log_value_of.previous_day()
        except IndexError:
            # TRANSITION FALLBACK
            try:
                regexp = re.compile("(.*)(daily)([\d\.]{8})\.?([\d]*).*-.*")
                previous_day = regexp.findall(base_package_version)[0]
                # make the version compatible with the new version
                previous_day = (
                    previous_day[0],
                    previous_day[1],
                    "20" + previous_day[2].replace(".", ""),
                    previous_day[3])
                log_value_of.previous_day()
            except IndexError:
                raise Exception(
                    "Didn't find a correct versioning in the current package: "
                    "{}".format(base_package_version))
        upstream_version = previous_day[0]
        log_value_of.upstream_version()
        log_value_of.series_version()
        log_value_of.today_version()
        if (previous_day[1] == series_version and
                previous_day[2] == today_version):
            minor = 1
            if previous_day[3]:  # second upload of the day
                minor = int(previous_day[3]) + 1
            today_version = "{}.{}".format(today_version, minor)
            log_value_of.today_version('After incrementing')

    new_upstream_version = (
        "{upstream}+{series}.{date}{destppa}{special}".format(
            upstream=upstream_version, series=series_version,
            date=today_version, destppa=destppa, special=special_tag))
    if native_pkg is not True:
        new_upstream_version = "{}-0ubuntu1".format(new_upstream_version)

    log_value_of.new_upstream_version()
    return new_upstream_version


def collect_bugs(f, source_package_name):
    """Collect all bugs in the changelog until latest snapshot."""
    bugs = set()
    # matching only bug format that launchpad accepts
    group_bugs_regexp = re.compile("lp: ?(.*\d{5,})", re.IGNORECASE)
    bug_decipher_regexp = re.compile("(#\d{5,})+")
    new_upload_changelog_regexp = re.compile(
        settings.NEW_CHANGELOG_PATTERN.format(source_package_name))
    drop_hash = lambda bug_with_hash: bug_with_hash.replace('#', '')
    for line in f:
        grouped_bugs_list = group_bugs_regexp.findall(line)
        for grouped_bugs in grouped_bugs_list:
            for bug in map(drop_hash,
                           bug_decipher_regexp.findall(grouped_bugs)):
                bugs.add(bug)
        #  a released upload to distro (automated or manual)n exit as bugs
        #  before were already covered
        if new_upload_changelog_regexp.match(line):
            return bugs

    return bugs


def update_changelog(bzr_repo,
                     new_package_version,
                     series,
                     tip_bzr_rev,
                     authors_commits,
                     dest_ppa=None):
    """Update the changelog for the incoming upload."""
    dch_env = os.environ.copy()
    for author in authors_commits:
        dch_env["DEBFULLNAME"] = author
        for bug_desc in authors_commits[author]:
            if bug_desc.startswith('-'):
                # Remove leading '-' or dch thinks that it's an option
                bug_desc = bug_desc[1:]
            if bug_desc.startswith(' '):
                # Remove leading spaces, there are useless and the result is
                # prettier without them anyway ;)
                bug_desc = bug_desc.strip()
            cmd = ["dch",
                   "--multimaint-merge",
                   "--release-heuristic",
                   "changelog",
                   "-v{}".format(new_package_version),
                   bug_desc]
            process = Popen(
                cmd, stdout=PIPE, stderr=PIPE, cwd=bzr_repo, env=dch_env)
            std_log(process.communicate())

    if tip_bzr_rev is not None:
        commit_message = "{} {}".format(
            settings.REV_STRING_FORMAT, tip_bzr_rev)
        if dest_ppa:
            commit_message += " ({})".format(dest_ppa)
    else:
        commit_message = ""

    dch_env["DEBFULLNAME"] = settings.BOT_DEBFULLNAME
    dch_env["DEBEMAIL"] = settings.BOT_DEBEMAIL
    instance = Popen(
        ['dch', '--release-heuristic', 'changelog',
            '-v{}'.format(new_package_version), commit_message],
        stdout=PIPE, stderr=PIPE, cwd=bzr_repo, env=dch_env)
    (stdout, stderr) = instance.communicate()
    if instance.returncode != 0:
        raise Exception(stderr.decode("utf-8").strip())
    process = Popen(
        ['dch', '-r', '--distribution', series, '--force-distribution', ''],
        stdout=PIPE, stderr=PIPE, cwd=bzr_repo, env=dch_env)
    std_log(process.communicate())

    # in the case of no commit_message and no symbols file change, we have an
    # addition [ DEBFULLNAME ] follow by an empty line better to remove both
    sed_cmd = '/ \[ ' + settings.BOT_DEBFULLNAME + ' \]/{$q; N; /\\n$/d;}'
    process = Popen(
        ['sed', '-i', sed_cmd, 'debian/changelog'],
        stdout=PIPE, stderr=PIPE, cwd=bzr_repo)
    std_log(process.communicate())


def build_source_package(branch_dir, series, distro_version, ppa=None):
    """Build the source package using the internal helper.

    Add the additional ppa inside the chroot if requested.
    """
    chroot_tool_dir = os_path_join_safe(settings.ROOT_CU2D, "chroot-tools")
    buildsource = os_path_join_safe(chroot_tool_dir, "buildsource-chroot")
    log_value_of.branch_dir()
    cowbuilder_env = os.environ.copy()
    cowbuilder_env["HOME"] = chroot_tool_dir  # take the internal .pbuilderrc
    cowbuilder_env["DIST"] = series
    instance = Popen(
        ['./create-chroot'],
        stdout=PIPE, stderr=PIPE, cwd=chroot_tool_dir, env=cowbuilder_env)
    std_log(instance.communicate())
    cmd = ["sudo", "-E", "cowbuilder", "--execute",
           "--bindmounts", SILO_DIR(),
           "--bindmounts", settings.GNUPG_DIR,
           "--", buildsource, branch_dir,
           "--gnupg-parentdir", settings.GNUPG_DIR,
           "--uid", str(os.getuid()), "--gid", str(os.getgid()),
           "--distro-version", distro_version]
    if ppa:
        cmd.extend(["--ppa", ppa])
    instance = Popen(
        cmd, stdout=PIPE, stderr=PIPE, cwd=branch_dir, env=cowbuilder_env)
    std_log(instance.communicate())
    if instance.returncode != 0:
        raise Exception("%r returned: %s." % (cmd, instance.returncode))


def simple_upload(changes, ppa):
    """Upload the package to a ppa."""
    log_value_of.changes('Uploading')
    cmd = ('dput', 'ppa:' + ppa, changes)
    dput = Popen(cmd, stdout=PIPE, stderr=PIPE, cwd=os.path.dirname(changes))
    std_log(dput.communicate())
    if dput.returncode != 0:
        raise Exception('{} returned {}'.format(cmd, dput.returncode))


def refresh_symbol_files(bzr_repo, packaging_version):
    """Refresh the symbols file having REPLACEME_TAG with version of the day.

    Add a changelog entry if needed.
    """
    new_upstream_version = packaging_version.split("-")[0]
    files_replaced = set()
    debian_dir = os_path_join_safe(bzr_repo, 'debian')
    for filename in os.listdir(debian_dir):
        path = os_path_join_safe(debian_dir, filename)
        if not os.path.isfile(path) or os.path.islink(path):
            continue
        for line in fileinput.input(path, inplace=1):
            if settings.REPLACEME_TAG in line:
                files_replaced.add(filename)
                line = line.replace(settings.REPLACEME_TAG,
                                    new_upstream_version)
            sys.stdout.write(line)

    if files_replaced:
        dch_env = os.environ.copy()
        dch_env["DEBFULLNAME"] = settings.BOT_DEBFULLNAME
        dch_env["DEBEMAIL"] = settings.BOT_DEBEMAIL
        for filename in sorted(files_replaced):
            Popen(
                ['dch', '--release-heuristic', 'changelog',
                    'debian/{}: auto-update to released version'.format(
                        filename)],
                stdout=PIPE, stderr=PIPE, cwd=bzr_repo, env=dch_env
            ).communicate()
        process = Popen(
            ['bzr', 'commit', '-m', 'Replaced 0replaceme with real versions.'],
            stdout=PIPE, stderr=PIPE, cwd=bzr_repo)
        std_log(process.communicate())


def has_dont_change_version_flag(bzr_repo):
    """Return true if debian/control has a "don't change versioning" flag."""
    with open(os_path_join_safe(bzr_repo, 'debian', 'control')) as f:
        for line in f:
            if line.startswith(settings.AUTOUPLOAD_CONTROL_TAG):
                if settings.NO_REWRITE_VERSION_TAG in line:
                    return True
    return False


def check_package_reached_destination(silo_state,
                                      packages_in_dest,
                                      ignoremissingprojects,
                                      ignorepackagesnotindest,
                                      for_merge_and_clean=False):
    """Identify where packages are in their migration.

    Return (True, "All packages in destination") if all packages have reached
    the destination. Otherwise, return (False, message).
    """
    # check that all package sources were in the silo configuration
    diff = set(silo_state.all_projects).difference(packages_in_dest)
    log_value_of.diff()

    # additional check for merge and clean, we want all projects in silo
    # configuration to be built and published
    if for_merge_and_clean and diff:
        message = '{} did not get published.'.format(', '.join(diff))
        if ignoremissingprojects:
            logging.info(message + ' So we will not merge those ones.')
            # remove them as we are not going to merge their content.
            for remaining_project in diff:
                silo_state.remove_source(remaining_project)
        else:
            logging.error(message + ' Please check available options.')
            return (False, "Can't merge: " + message)

    logging.info("Check that all package sources are published in destination")
    one_package_not_in_dest = False
    dest = lp.load(silo_state.dest)
    series = lp.load(silo_state.series)
    additional_messages = ""
    for source in packages_in_dest:
        in_dest = is_version_for_series_in_dest(
            source,
            packages_in_dest[source],
            series,
            dest)
        in_updates = is_version_for_series_in_dest(
            source,
            packages_in_dest[source],
            series,
            dest,
            pocket="Updates")
        if not in_dest and not in_updates:
            logging.warning(
                "{} ({}) is not published yet in {} (Release pocket).".format(
                    source, packages_in_dest[source], dest.name))
            one_package_not_in_dest = True

            # if destination is the archive, try to check if it's in proposed
            # or in any queue

            if lp.is_link_distro_archive(silo_state.dest):
                if is_version_for_series_in_dest(
                        source,
                        packages_in_dest[source],
                        series,
                        dest,
                        pocket="Proposed"):
                    in_proposed_msg = (
                        "{} ({}) is in the proposed pocket. ".format(
                            source, packages_in_dest[source]))
                    logging.warning(
                        in_proposed_msg +
                        "You run that job either too quickly or it's stuck "
                        "there. More information available at "
                        "https://wiki.ubuntu.com/ProposedMigration.")
                    additional_messages += in_proposed_msg
                elif is_version_in_queue(
                        source,
                        packages_in_dest[source],
                        series,
                        "New"):
                    in_new_msg = (
                        "{} ({}) is in the NEW queue. ".format(
                            source, packages_in_dest[source]))
                    logging.warning(
                        in_new_msg +
                        "You need an archive admin to review a "
                        "NEW this package.")
                    additional_messages += in_new_msg
                elif is_version_in_queue(
                        source,
                        packages_in_dest[source],
                        series,
                        "Unapproved"):
                    in_unapproved_msg = (
                        "{} ({}) is in the UNAPPROVED queue. ".format(
                            source, packages_in_dest[source]))
                    logging.warning(
                        in_unapproved_msg +
                        "You need a release team member to review an approve "
                        "it. We are probably in a freeze period.")
                    additional_messages += in_unapproved_msg
                else:
                    in_unknown_msg = (
                        "{} ({}) is in no known "
                        "spacetime. ".format(source, packages_in_dest[source]))
                    logging.warning(
                        in_unknown_msg +
                        "We didn't find it anywhere. It could have been "
                        "rejected, or you run it too quickly and it's not "
                        "published in proposed yet, or a network issue "
                        "happened. If this persistent some time after the "
                        "publication, you need to ping the trainguards to "
                        "get more information.")
                    additional_messages += in_unknown_msg

    if one_package_not_in_dest:
        if for_merge_and_clean and ignorepackagesnotindest:
            logging.info("The ignore flag is set.")
        else:
            # this is an error for merge and clean, just a warning for
            # automated check
            message = (
                "One package at least is not available at the "
                "destination. ")
            if for_merge_and_clean:
                logging.error(
                    message +
                    "See above. You can use the ignore package not in dest "
                    "flag. The eventual merge will still be proceeded.")
                return (False, "Can't merge: " + message + additional_messages)
            else:
                logging.warning(
                    message +
                    "See above. You can use the ignore package not in dest "
                    "flag in \"merge and clean\". The eventual merge will "
                    "still be processed.")
                return (False, message + additional_messages)
    return (
        True, "All packages are in destination. You can Merge and Clean now.")


def increment_ubuntu_package_version(version):
    """Bump Ubuntu version in case of rebuild without modifying tarball."""
    regexp = re.compile("([\d]*ubuntu)([\d]*)$")
    result = regexp.findall(version)
    if result:
        # only increment the ubuntu version
        ubuntu_version = int(result[0][1]) + 1
        version = re.sub(regexp, "\\1", version)
        version += str(ubuntu_version)
    else:
        # we need to check if we need to add ubuntu1 or -xubuntu1
        regexp = re.compile("(-[\d]*)$")
        if regexp.findall(version):
            version = re.sub(regexp, "\\1ubuntu1", version)
        else:
            version += "ubuntu1"

    return version
