# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Robert Bruce Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import glob
import os

from six.moves import configparser
from collections import defaultdict

from cupstream2distro.utils import (
    SILO_DIR,
    ignored,
    log_value_of,
    os_path_join_safe,
)


SECTIONS = {
    'dest_current_version': 'Package',
    'packaging_version': 'Package',
    'branch': 'Branch',
    'rev': 'Branch',
}


def section(key):
    """Identify which .project file section the given key is from."""
    return SECTIONS.get(key, 'Misc')


def _get_project_files(infix):
    """Glob for the project files."""
    return glob.glob(os_path_join_safe(SILO_DIR(), '*{}'.format(infix)))


def _get_state(filename):
    """Read the .project file from disk unconditionally and return it."""
    config = configparser.RawConfigParser()
    config.read(filename)
    return config


class DotProject(object):
    """Provide a standard API for accessing and modifying .project files."""
    DIFF_NAME = 'packaging_changes_{}_{}.diff'
    SUFFIX = '.project'
    _state = None

    @staticmethod
    def get_all_packages_uploaded():
        """Get (package, version, rev, branch) of all packages uploaded.

        We do not rely on the .changes files because we need the exact version,
        which can have an epoch. And also because the .changes files are not
        present for manual source uploads.
        """
        result = set()
        project_files = _get_project_files(DotProject.SUFFIX)
        log_value_of.project_files()
        for filename in project_files:
            basename = os.path.basename(filename)
            name = basename.split(DotProject.SUFFIX)[0]
            proj = DotProject(name)
            result.add((name, proj.packaging_version, proj.rev, proj.branch))
        log_value_of.result('Returning')
        return result

    @staticmethod
    def get_all_packages_published():
        """Return a dict of published projects for this silo."""
        result = defaultdict(list)
        project_files = _get_project_files(DotProject.SUFFIX)
        log_value_of.project_files()
        for filename in project_files:
            basename = os.path.basename(filename)
            proj = DotProject(basename.split(DotProject.SUFFIX)[0])
            if proj.published_version:
                result[proj.name].append(proj.published_version)
        log_value_of.result('Returning')
        return dict(result)

    @staticmethod
    def soft_save(name, branch='', rev='', dest='', ours=''):
        """Create a .project file without overwriting values in an old one."""
        proj = DotProject(name)
        proj.branch = branch or proj.branch
        proj.rev = rev or proj.rev
        proj.dest_current_version = dest or proj.dest_current_version
        proj.packaging_version = ours or proj.packaging_version
        proj.write()

    def __init__(self, source_package_name):
        """Identify the location of the .project file."""
        object.__setattr__(self, 'name', source_package_name)
        object.__setattr__(self, 'filename', os_path_join_safe(
            SILO_DIR(), self.name + self.SUFFIX))

    @property
    def state(self):
        """Load the .project file from disk if necessary."""
        if not self._state:
            object.__setattr__(self, '_state', _get_state(self.filename))
        return self._state

    def __getattr__(self, key):
        """Access a value from a .project file, reading from disk if needed."""
        with ignored(configparser.NoSectionError, configparser.NoOptionError):
            return self.state.get(section(key), key)
        return ''

    def __setattr__(self, key, value):
        """Set a value for the .project file in memory only."""
        with ignored(configparser.Error):
            self.state.add_section(section(key))
        self.state.set(section(key), key, value)

    def get_diff_path(self, v=None):
        """Return the packaging diff filename."""
        return os_path_join_safe(
            SILO_DIR(), self.DIFF_NAME.format(
                self.name, v or self.published_version))

    def mark_as_published(self, version):
        """Record the published version in the project file."""
        self.published_version = version
        self.write()
        diff = self.get_diff_path(version)
        log_value_of.diff()
        with ignored(OSError):
            os.rename(diff, '{}.published'.format(diff))

    def write(self):
        """Write the .project file to disk if we have something to write."""
        with open(self.filename, 'wb') as f:
            if self._state:
                self._state.write(f)
