# -*- coding: utf-8 -*-
# Copyright (C) 2014 Canonical
#
# Authors:
#  Robert Bruce Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import json
import logging
import os
import shutil
import signal
import sys

from six.moves import range as gen_range
from six import iteritems

from cupstream2distro.settings import (
    ALL_SILO_NAMES,
    SILO_NAME_LIST,
    SILO_STATUS_RSYNCDIR,
    SILOS_DIR,
)
from cupstream2distro.utils import env, ignored, os_path_join_safe

(SILO_EMPTY, SILO_BUILTCHECKED, SILO_PUBLISHED, SILO_DONE) = gen_range(4)

# All interesting states a silo can be in
SILO_STATE_NONE = 1
SILO_STATE_BUILDING = 2
SILO_STATE_BUILD_SUCCESSFUL = 3
SILO_STATE_BUILD_FAILED = 4
SILO_STATE_RECONFIGURE_FAILED = 5
SILO_STATE_PUBLISHING = 6
SILO_STATE_PUBLISH_FAILED = 7
SILO_STATE_MIGRATING = 8
SILO_STATE_MERGING = 9
SILO_STATE_MERGE_FAILED = 10
SILO_STATE_CLEANING = 11


default_excepthook = sys.excepthook


class SiloState(object):
    """This class stores and manipulates the silo config json blob."""
    CONFIG_FILENAME = 'config'
    _config = None
    _configfile = None

    @staticmethod
    def safe_load(siloname):
        try:
            return SiloState(siloname)
        except IOError as e:
            logging.debug('Loading silo {}: {}'.format(siloname, e))

    @staticmethod
    def find_by_requestid(requestid):
        """Find a silo by it's requestid."""
        logging.info('Checking if {} is in any other silo'.format(requestid))
        for siloname in ALL_SILO_NAMES:
            logging.debug('Checking {} for {}'.format(siloname, requestid))
            silo_state = SiloState.safe_load(siloname)
            if silo_state and silo_state.requestid == requestid:
                return silo_state

    @staticmethod
    def new_blank(siloname, requestid='', dest='', series='', landers=[]):
        """Create a new SiloState object for a newly-assigned silo."""
        silo_state = SiloState(siloname, do_load=False)
        silo_state._config = {
            'requestid': requestid,
            'global': dict(dest=dest, series=series, step='', status=dict(
                url='', message='', ping=True, state=-1)),
            'landers': landers,
            'siloname': siloname,
            'sources': [],
            'mps': {},
        }
        return silo_state

    @staticmethod
    def find_first_available(distribution='ubuntu'):
        """Return the first silo that doesn't have a /config file, or None."""
        for siloname in sorted(SILO_NAME_LIST[distribution]):
            filename = os_path_join_safe(
                SILOS_DIR, siloname, SiloState.CONFIG_FILENAME)
            if not os.path.isfile(filename):
                with ignored(OSError):
                    os.makedirs(os.path.dirname(filename))
                return siloname

    @staticmethod
    def is_project_unique(dest, series, project, reconfigure=None):
        """Return True if the project already exists in a silo."""
        logging.info(
            'Checking if {} {} ({} {}) conflicts with any other silos.'.format(
                project, dest.name, dest.distribution.name, series.name))
        for siloname in SILO_NAME_LIST[dest.distribution.name]:
            if reconfigure == siloname:
                continue
            silo_state = SiloState.safe_load(siloname)
            if silo_state:
                if silo_state.dest == dest.self_link:
                    if silo_state.series == series.self_link:
                        if project in silo_state.all_projects:
                            message = '{} is already in {}'.format(
                                project, siloname)
                            if env.IGNORECONFLICTS == 'true':
                                logging.warning(message)
                            else:
                                logging.error(message)
                            return False
        return True

    def __init__(self, siloname, do_load=True):
        """Create a new SiloState object for an already-assigned silo."""
        self.siloname = siloname
        if do_load:
            self.load_config()

    def load_config(self):
        """Read in the JSON blob."""
        with open(self.configfile) as fd:
            self._config = json.load(fd)

    def save_config(self):
        """Write out the JSON blob."""
        newfile = '{}.new'.format(self.configfile)
        with open(newfile, 'w') as fd:
            json.dump(self._config, fd)
        os.rename(newfile, self.configfile)
        with ignored(OSError):
            os.makedirs(os.path.dirname(self.rsyncfile))
        shutil.copy2(self.configfile, self.rsyncfile)

    def remove_rsync_file(self):
        """Delete the rsync'd copy of the config file."""
        logging.debug('Deleting! ' + self.rsyncfile)
        os.unlink(self.rsyncfile)

    def set_config_status(self, state, status, ping=True):
        """Update to reflect the latest status."""
        self._config['global']['status'] = {
            'message': status,
            'ping':  ping,
            'url': env.BUILD_URL + 'console',
            'state': state,
        }

    def set_package_version_list(self, pkglist):
        """Set the list of published package versions."""
        self._config['global']['pkgversionlist'] = {
            k: sorted(v) for k, v in iteritems(pkglist)}

    @property
    def configfile(self):
        if self._configfile is None:
            self._configfile = os_path_join_safe(
                SILOS_DIR, self.siloname, self.CONFIG_FILENAME)
        return self._configfile

    @property
    def rsyncfile(self):
        return os_path_join_safe(SILO_STATUS_RSYNCDIR, self.siloname)

    @property
    def all_projects(self):
        """Get a list of all projects."""
        return sorted(list(self._config['mps']) + self._config['sources'])

    @property
    def is_built_checked(self):
        return self.config_step == SILO_BUILTCHECKED

    @property
    def is_cleaning(self):
        return self.state == SILO_STATE_CLEANING

    @property
    def is_empty(self):
        return self.config_step == SILO_EMPTY

    @property
    def is_merging(self):
        return self.state == SILO_STATE_MERGING

    @property
    def is_publishing(self):
        return self.state == SILO_STATE_PUBLISHING

    @property
    def is_published(self):
        return self.config_step == SILO_PUBLISHED

    @property
    def is_merge_failed(self):
        return self.state == SILO_STATE_MERGE_FAILED

    def set_building(self):
        self.set_config_status(SILO_STATE_BUILDING, 'Building')

    def set_build_failed(self, message):
        self.set_config_status(SILO_STATE_BUILD_FAILED, message)

    def set_build_successful(self):
        self.set_config_status(SILO_STATE_BUILD_SUCCESSFUL, 'Packages built')

    def set_publishing(self):
        self.set_config_status(SILO_STATE_PUBLISHING, 'Publishing')

    def set_publish_failed(self, message):
        self.set_config_status(SILO_STATE_PUBLISH_FAILED,
                               "Can't publish: " + message)

    def set_cleaning(self, message):
        self.set_config_status(SILO_STATE_CLEANING, message)

    def set_empty(self):
        self.config_step = SILO_EMPTY

    def set_merging(self):
        self.set_config_status(SILO_STATE_MERGING, 'Merging')

    def set_merge_failed(self, message):
        self.set_config_status(SILO_STATE_MERGE_FAILED, message)

    def set_migrating(self, message='Packages migrating to destination'):
        self.set_config_status(SILO_STATE_MIGRATING, message)

    def set_preparing(self):
        self.set_config_status(SILO_STATE_BUILDING, 'Preparing packages')

    def set_ready(self):
        self.set_config_status(SILO_STATE_NONE, 'Silo ready to build')
        self.config_step = SILO_EMPTY

    def set_reconfigure_failed(self, message):
        self.config_step = -1
        self.set_config_status(
            SILO_STATE_RECONFIGURE_FAILED, 'Reconfigure failed: ' + message)

    def set_built_checked(self):
        self.config_step = SILO_BUILTCHECKED

    def set_done(self):
        self.config_step = SILO_DONE

    def set_published(self):
        self.config_step = SILO_PUBLISHED

    @property
    def config_step(self):
        return self._config['global']['step']

    @config_step.setter
    def config_step(self, value):
        self._config['global']['step'] = value

    @property
    def dest(self):
        return self._config['global']['dest']

    @dest.setter
    def dest(self, value):
        self._config['global']['dest'] = value

    @property
    def landers(self):
        return self._config['landers']

    @landers.setter
    def landers(self, value):
        self._config['landers'] = value

    @property
    def mps(self):
        return self._config['mps']

    @mps.setter
    def mps(self, value):
        self._config['mps'] = value

    @property
    def ppa(self):
        return self._config['global']['ppa']

    @ppa.setter
    def ppa(self, value):
        self._config['global']['ppa'] = value

    @property
    def requestid(self):
        return self._config['requestid']

    @requestid.setter
    def requestid(self, value):
        self._config['requestid'] = value

    @property
    def series(self):
        return self._config['global']['series']

    @series.setter
    def series(self, value):
        self._config['global']['series'] = value

    @property
    def sources(self):
        return self._config['sources']

    @sources.setter
    def sources(self, value):
        self._config['sources'] = value

    @property
    def state(self):
        return self._config['global']['status']['state']

    @state.setter
    def state(self, value):
        self._config['global']['status']['state'] = value

    @property
    def source_sync_from_silo(self):
        return self._config.get('source_sync_from_silo')

    @source_sync_from_silo.setter
    def source_sync_from_silo(self, value):
        self._config['source_sync_from_silo'] = value

    @property
    def source_sync_all_contents(self):
        return self._config.get('source_sync_all_contents')

    @source_sync_all_contents.setter
    def source_sync_all_contents(self, value):
        self._config['source_sync_all_contents'] = value

    @property
    def source_archive(self):
        return self._config.get('source_archive')

    @source_archive.setter
    def source_archive(self, value):
        self._config['source_archive'] = value

    @property
    def source_series(self):
        return self._config.get('source_series')

    @source_series.setter
    def source_series(self, value):
        self._config['source_series'] = value

    @property
    def additional_targets(self):
        return self._config.get('additional_targets')

    @additional_targets.setter
    def additional_targets(self, value):
        self._config['additional_targets'] = value

    def append_mp(self, source_package, mp_link):
        self._config['mps'].setdefault(source_package, []).append(mp_link)

    def remove_source(self, source_package):
        if source_package in self.mps:
            del self._config['mps'][source_package]
        if source_package in self.sources:
            self._config['sources'].remove(source_package)

    def get_mtime(self):
        """Return the mtime of our configfile."""
        return os.path.getmtime(self.configfile)


def save_error_message(message):
    """Load the silo state, set an error message, then save it back."""
    logging.error(message)
    if env.SILONAME:
        silo_state = SiloState(env.SILONAME)
        silo_state.set_config_status(-1, message)
        silo_state.save_config()


def cleanup_on_kill(sig, frame):
    """Call save_error_message when we catch a SIGTERM."""
    save_error_message('Caught signal {}, aborting.'.format(sig))


def cleanup_on_exception(e_class, e_instance, traceback):
    """Call save_error_message when we have an unhandled exception."""
    save_error_message('Uncaught exception: {}: {}'.format(
        type(e_instance).__name__,
        e_instance))
    default_excepthook(e_class, e_instance, traceback)


signal.signal(signal.SIGTERM, cleanup_on_kill)
sys.excepthook = cleanup_on_exception
