# -*- coding: utf-8 -*-
# Copyright (C) 2013 Canonical
#
# Authors:
#  Robert Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import sys
import logging

from contextlib import contextmanager
from inspect import currentframe

from cupstream2distro.settings import SILOS_DIR
from cupstream2distro.errors import UnexpectedAbsolutePathError


# this is stolen from python 3.4 :)
@contextmanager
def ignored(*exceptions):
    try:
        yield
    except exceptions:
        pass


def singleton(cls):
    """Decorator that creates a single global instance of a class."""
    class single(cls):
        def __call__(self):
            return self
    return single()


def os_path_join_safe(first, *rest):
    msg = 'Unexpected absolute path after first argument. ' + repr(rest)
    for part in rest:
        if os.path.isabs(part):
            raise UnexpectedAbsolutePathError(msg)
    return os.path.join(first, *rest)


@singleton
class env:
    """Handle some common environment variable manipulations."""
    def __init__(self):
        logging.basicConfig(
            level=logging.DEBUG if self.DEBUG == 'true' else logging.INFO,
            format='%(asctime)s %(levelname)s %(message)s')
        if self.DEBUG == 'true':
            logging.debug(str(self))

    def __getattr__(self, key):
        return os.environ.get(key, '')

    def __setattr__(self, key, value):
        os.environ[key] = str(value)

    def __str__(self):
        from pprint import pformat
        return 'Environment Variables:\n' + pformat(dict(os.environ))


def get_id(merge):
    return merge.source_branch.bzr_identity


def get_pre(merge):
    if merge.prerequisite_branch is not None:
        return merge.prerequisite_branch.bzr_identity
    return None


def reorder_branches(merges):
    """Return merges ordered by their prerequisites, if any."""
    ids = [get_id(m) for m in merges]
    leafs = [m for m in merges if get_pre(m) not in ids]
    ordered = []

    def recurse_upwards(merge):
        """Recurse up the node graph, from leafs to trunks."""
        ordered.append(merge)
        for parent in merges:
            if get_pre(parent) == get_id(merge):
                recurse_upwards(parent)

    [recurse_upwards(leaf) for leaf in leafs]
    return ordered


def SILO_DIR():
    """Return the directory of the currently active silo."""
    return os_path_join_safe(SILOS_DIR, env.SILONAME)


@singleton
class log_value_of:
    """Reach into the calling scope and log the value of a variable.

    This class relies on inspect.currentframe(), which is a CPython
    implementation detail, which means we'll need to delete this if we ever
    want to migrate to Jython (unlikely).

    Invoke this by calling log_value_of.some_local_var('Hello') and it will
    grab the value of some_local_var from the calling scope, and call
    logging.debug with 'Hello: some_local_var: foo' (where foo is the value of
    some_local_var).
    """
    def __getattr__(self, var):
        def debug_log(msg='Variable'):
            value = str(currentframe().f_back.f_locals[var])
            details = [msg, var, value]
            if value.startswith(os.sep):
                details.append('Exists? ' + str(os.path.exists(value)))
            logging.debug(': '.join(details))
        return debug_log


def find_all(infix, start=SILO_DIR(), ignore='build-area'):
    """Recursively find all filenames in the silo containing infix."""
    for root, dirs, files in os.walk(start):
        for name in files:
            if infix in name:
                path = os_path_join_safe(root, name)
                if ignore not in path:
                    yield path


def std_log(std):
    """Log and passthrough stdout and stderr."""
    (stdout, stderr) = std
    if stdout:
        logging.info('\n' + stdout)
    if stderr:
        logging.warning('\n' + stderr)
    return stdout, stderr


def run_script(name, doc, main):
    """Act like a script if we got invoked like a script."""
    if name == '__main__':
        if '--help' in sys.argv or '-h' in sys.argv:
            sys.stdout.write(doc)
        else:
            sys.exit(main())
