# -*- coding: utf-8 -*-
# Copyright: (C) 2013 Canonical
#
# Authors:
#  Didier Roche
#  Robert Bruce Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA


import os
import imp
import time
import shutil
import tempfile
import unittest

from mock import Mock

from cupstream2distro.utils import ignored, env, os_path_join_safe


def clean(line):
    """Strip out parts of a changelog that are allowed to differ."""
    return line.split('>  ')[0].split('; urgency=')[0].strip()


class DirectoryAwareTestCase(unittest.TestCase):
    """This TestCase contains absolute paths to important directories."""
    _tempdir = None
    test_dir = os.path.dirname(os.path.dirname(__file__))
    data_dir = os_path_join_safe(test_dir, 'data')
    changelogs_file_dir = os_path_join_safe(data_dir, 'changelogs')
    result_dir = os_path_join_safe(data_dir, 'results')
    project_file_dir = os_path_join_safe(data_dir, 'project_files')
    root_dir = os.path.dirname(test_dir)
    train_dir = os_path_join_safe(root_dir, 'citrain')
    cup_dir = os_path_join_safe(root_dir, 'cupstream2distro')

    def setUp(self):
        """Time how long each test runs."""
        super(DirectoryAwareTestCase, self).setUp()
        self.start_time = time.time()

    def tearDown(self):
        """Delete that temporary working directory, if it was created."""
        super(DirectoryAwareTestCase, self).tearDown()
        if self._tempdir:
            shutil.rmtree(self._tempdir)
            self._tempdir = None
        for path in ('~/status', '~/silos'):
            self.assertFalse(os.path.exists(os.path.expanduser(path)))
        ms = (time.time() - self.start_time) * 1000
        if env.BEAT_THE_CLOCK and ms > 100:
            raise Exception('Too slow! Test took {:.2f}ms to run.'.format(ms))

    @property
    def tempdir(self):
        """Lazily create a temporary directory for tests to make files in."""
        if not self._tempdir:
            self._tempdir = tempfile.mkdtemp()
        return self._tempdir


class BaseTestCase(DirectoryAwareTestCase):
    """Test module with helpers and keeping track of temp directories."""

    @classmethod
    def addToPath(cls, path):
        """Prepend some data to path, if path is relative, test_dir is used."""
        if not os.path.isabs(path):
            path = os_path_join_safe(cls.test_dir, path)
        os.environ['PATH'] = "{}:{}".format(path, os.environ["PATH"])

    @classmethod
    def removeFromPath(cls, path):
        """Remove some path from the PATH environment variables.

        If path is relative, test_dir is used.
        """
        if not os.path.isabs(path):
            path = os_path_join_safe(cls.test_dir, path)
        elems = []
        for path_elem in os.environ['PATH'].split(':'):
            if path_elem != path:
                elems.append(path_elem)
        os.environ['PATH'] = ':'.join(elems)

    def setUp(self):
        super(BaseTestCase, self).setUp()
        self._dirs_to_remove = []
        os.environ['MOCK_MODE'] = "0"

    def tearDown(self):
        """Remove all temp dirs and return to root test dir."""
        super(BaseTestCase, self).tearDown()
        os.chdir(self.test_dir)
        for dir in self._dirs_to_remove:
            try:
                shutil.rmtree(dir)
            except OSError:
                pass

    def create_temp_workdir(self):
        """Create a temporary work directory."""
        tempdir = tempfile.mkdtemp()
        self._dirs_to_remove.append(tempdir)
        return tempdir

    def get_origin_branch_path(self, target_branch_name):
        """Return the branch url in the data directory, no other changes."""
        return os_path_join_safe(self.data_dir, 'branches', target_branch_name)

    def prep_debian_file(self, filename, content):
        """Prepare a file for modification."""
        path = os_path_join_safe(self.tempdir, 'debian', filename)
        with ignored(OSError):
            os.mkdir(os.path.dirname(path))
        with open(path, 'w') as f:
            f.write(content)
        return path

    def get_ubuntu_source_content_path(self, package_name):
        """Return the ubuntu source package path for package_name."""
        return os_path_join_safe(
            self.data_dir, 'ubuntu_source_packages', package_name)

    def assertFilesAreIdenticals(self, filename1, filename2):
        """Assert if the files contents are identical."""
        self.assertEquals(
            open(filename1).read().strip(),
            open(filename2).read().strip())

    def assertChangelogFilesAreIdenticals(self, a, b):
        """Assert changelog files are identical, removing days/hours."""
        with open(a) as one:
            with open(b) as two:
                for i, line in enumerate(one):
                    self.assertEqual(
                        '{}: {}'.format(i, clean(line)),
                        '{}: {}'.format(i, clean(two.readline())))


class BaseUnitTestCase(BaseTestCase):
    """Base unit test case module, add all mocks to PATH."""
    @classmethod
    def setUpClass(cls):
        super(BaseUnitTestCase, cls).setUpClass()
        cls.addToPath(os_path_join_safe("mocks", "offline"))
        cls.addToPath(os_path_join_safe("mocks", "online"))

    def setUp(self):
        super(BaseUnitTestCase, self).setUp()
        os.chdir(self.tempdir)
        os.environ['MOCK_MODE'] = "0"
        os.environ['MOCK_ERROR_MODE'] = "0"


class CITrainScriptTestCase(DirectoryAwareTestCase):
    """Custom-import citrain script files so they can be tested."""
    scriptname = None
    script = None

    def setUp(self):
        """Re-import the script for each test.

        Allows you to assign mocks to self.script.foo without trampling
        on other tests, as the script is reloaded each time.
        """
        super(CITrainScriptTestCase, self).setUp()
        if self.scriptname:
            self.script = imp.load_source(
                os.path.basename(self.scriptname).split('.')[0],
                os_path_join_safe(self.train_dir, self.scriptname))
            self.script.DotProject = Mock()
            self.script.SILO_DIR = lambda: self.tempdir
            self.script.SiloState = Mock()
            self.script.branchhandling = Mock()
            self.script.find_all = Mock()
            self.script.get_launchpad = Mock()
            self.script.glob = Mock()
            self.script.logging = Mock()
            self.script.lp = Mock()
            self.script.os = Mock()
            self.script.os.path.basename = os.path.basename
            self.script.packagemanager = Mock()
            self.script.shutil = Mock()
            self.script.silo_state = Mock()
            self.script.time = Mock()

    def tearDown(self):
        """Delete .pyc files created by these tests."""
        super(CITrainScriptTestCase, self).tearDown()
        if self.script:
            os.unlink(os_path_join_safe(self.train_dir, self.scriptname + 'c'))

    def test_existence(self):
        """CI Train script has a main() func without SyntaxErrors."""
        if self.script:
            self.assertTrue(self.script.main)
