# -*- coding: utf-8 -*-
# Copyright: (C) 2013 Canonical
#
# Authors:
#  Didier Roche
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from six.moves import range as gen_range
from mock import patch

from tests import strings as s
from tests.unit import BaseUnitTestCase

from cupstream2distro import branchhandling
from cupstream2distro.utils import os_path_join_safe


class BranchHandlingTests(BaseUnitTestCase):
    def setUp(self):
        super(BranchHandlingTests, self).setUp()
        # We want the full diff on failures
        self.maxDiff = None

    @patch('cupstream2distro.branchhandling.Popen')
    def test_branching(self, popenMock):
        """We correcly try to branch a branch."""
        popenMock.return_value.communicate.return_value = ('', '')
        popenMock.return_value.returncode = 0
        source_branch = os_path_join_safe(self.tempdir, 'a')
        branchdir = os_path_join_safe(self.tempdir, 'b')
        branchhandling.get_branch(source_branch, branchdir)
        popenMock.assert_called_once_with(
            ['bzr', 'branch', source_branch, branchdir],
            cwd=self.tempdir, stderr=-1, stdout=-1)
        popenMock().communicate.assert_called_once_with()

    @patch('cupstream2distro.branchhandling.Popen')
    def test_branching_failed(self, popenMock):
        """We raise an exception when failing to branch a branch."""
        popenMock.return_value.communicate.return_value = ('', 'failboat')
        popenMock.return_value.returncode = 42
        source_branch = os_path_join_safe(self.tempdir, 'a')
        branchdir = os_path_join_safe(self.tempdir, 'b')
        with self.assertRaisesRegexp(Exception, 'failboat'):
            branchhandling.get_branch(source_branch, branchdir)
        popenMock.assert_called_once_with(
            ['bzr', 'branch', source_branch, branchdir],
            cwd=self.tempdir, stderr=-1, stdout=-1)
        popenMock().communicate.assert_called_once_with()

    @patch('cupstream2distro.branchhandling.Popen')
    def test_get_tip_bzr_revision(self, popenMock):
        """We extract the tip of bzr revision."""
        popenMock.return_value.communicate.return_value = ('6:foo', '')
        popenMock.return_value.returncode = 0
        self.assertEqual(
            branchhandling.get_tip_bzr_revision(self.tempdir), 6)
        popenMock.assert_called_once_with(
            ['bzr', 'log', '-c', '-1', '--line'],
            cwd=self.tempdir, stderr=-1, stdout=-1)
        popenMock().communicate.assert_called_once_with()

    @patch('cupstream2distro.branchhandling.Popen')
    def test_get_tip_bzr_revision_failed(self, popenMock):
        """We raise an exception when failing to extract the tip rev."""
        popenMock.return_value.communicate.return_value = ('', 'boop')
        popenMock.return_value.returncode = 9000
        with self.assertRaisesRegexp(Exception, 'boop'):
            branchhandling.get_tip_bzr_revision(self.tempdir)
        popenMock.assert_called_once_with(
            ['bzr', 'log', '-c', '-1', '--line'],
            cwd=self.tempdir, stderr=-1, stdout=-1)
        popenMock().communicate.assert_called_once_with()

    def test_log_diff_simple(self):
        """Return the right log diff since a dedicated revision."""
        self.prep_debian_file('changelog', s.NESTED_CHANGELOG)
        result = os_path_join_safe(self.result_dir, 'bzr_log_simple')
        with open(result) as f:
            self.assertEquals(
                branchhandling.log_diff(self.tempdir, 3).strip(),
                f.read().strip())

    def test_log_diff_nested(self):
        """Return the right log diff (with nested elements)."""
        self.prep_debian_file('changelog', s.NESTED_CHANGELOG)
        result = os_path_join_safe(self.result_dir, 'bzr_log_nested')
        with open(result) as f:
            self.assertEquals(
                branchhandling.log_diff(self.tempdir, 1).strip(),
                f.read().strip())

    def test_log_diff_with_remerge_trunk(self):
        """Ensure we return the right log diff.

        Containing some remerge to trunk dating before the previous release.
        """
        self.prep_debian_file('changelog', s.NESTED_CHANGELOG)
        result = os_path_join_safe(self.result_dir, 'bzr_log_remerge_trunk')
        with open(result) as f:
            self.assertEquals(
                branchhandling.log_diff(self.tempdir, 12).strip(),
                f.read().strip())

    def test_extract_authors_simple(self):
        """Extract a single line containing an author."""
        self.assertEquals(
            branchhandling._extract_authors("foo <foo@bar.com>"),
            set(["foo"]))

    def test_extract_authors_without_email(self):
        """Extract a single line containing an author without email."""
        self.assertEquals(
            branchhandling._extract_authors("foo bar"),
            set(["foo bar"]))

    def test_extract_authors_with_multiple_simple(self):
        """Extract a single line containing multiple authors."""
        self.assertEquals(
            branchhandling._extract_authors(
                "foo bar <foo@bar.com>, baz <baz@baz.com>"),
            set(["foo bar", "baz"]))

    def test_extract_authors_with_multiple_missing_email_first(self):
        """Extract a line containing multiple authors, first missing email."""
        self.assertEquals(
            branchhandling._extract_authors("foo bar, baz <baz@baz.com>"),
            set(["foo bar", "baz"]))

    def test_extract_authors_with_multiple_missing_email_last(self):
        """Extract a line containing multiple authors, last missing email."""
        self.assertEquals(
            branchhandling._extract_authors("foo bar <foo@bar.com>, baz"),
            set(["foo bar", "baz"]))

    def test_return_bugs_one(self):
        """Extract the bug number from a simple string."""
        self.assertEquals(
            branchhandling._return_bugs(
                "https://launchpad.net/bugs/1171934"),
            set([1171934]))

    def test_return_bugs_double(self):
        """Extract the bug number from a simple string containing two bugs."""
        self.assertEquals(
            branchhandling._return_bugs(
                "https://launchpad.net/bugs/1171934 https://launchpad.net/"
                "bugs/1171935"),
            set([1171934, 1171935]))

    def test_return_bugs_multiple_lines(self):
        """Extract the bug number from a multi-line string."""
        self.assertEquals(branchhandling._return_bugs("""foo bar baz
            https://launchpad.net/bugs/1171934 blue bi doo
            https://launchpad.net/bugs/1171935 baz
            """), set([1171934, 1171935]))

    def test_return_bugs_nothing(self):
        """Extract no bug number when there is nothing."""
        self.assertEquals(branchhandling._return_bugs("bar baz doo"), set())

    def test_return_bugs_commit_message_bug_pattern(self):
        """Extract bugs from a commit message.

        Make sure we can extract multiple bugs from a typical commit
        message."""
        self.assertEquals(
            branchhandling._return_bugs(
                "Some commit message. This does not matter, only the rest. "
                "Fixes: #12345, #12346, #12347, #12348"),
            set(12345 + x for x in gen_range(4)))

    def test_return_bugs_other_patterns(self):
        """Extract bug numbers with special patterns."""
        self.assertEquals(
            branchhandling._return_bugs(
                """bug #12345, bug#12346, bug12347, bug 12348
    lp: #12349, lp:#12350, lp:12351, lp: 12352
    lp #12353, lp#12354, lp12355, lp 12356,
    Fix #12357, Fix 12358, Fix: 12359, Fix12360, Fix: #12361,
    Fixes #12362, Fixes 12363, Fixes: 12364, Fixes:12365, Fixes: #12366
    BUG: 12367
    #12368 (but not 12369 for false positive) """),
            set(12345 + x for x in gen_range(24)))

    def test_extract_commit_bugs(self):
        """Extra commit and bugs from a merged commit message."""
        self.assertEquals(
            branchhandling._extract_commit_bugs(
                """  UnityWindow: don't draw the panel shadow above
            multiple
            lines      with extra spaces
            and more. Fixes: https://bugs.launchpad.net/bugs/1171934.
              Approved by PS Jenkins bot, Andrea Azzarone."""),
            ("UnityWindow: don't draw the panel shadow above multiple lines "
             "with extra spaces and more.", set([1171934])))

    def test_extract_commit_no_bugs(self):
        """Extract commit and bugs when there are no bugs."""
        self.assertEquals(
            branchhandling._extract_commit_bugs(
                """  UnityWindow: don't draw the panel shadow above
            multiple
            lines      with extra spaces
            and more.
              Approved by PS Jenkins bot, Andrea Azzarone."""),
            ("UnityWindow: don't draw the panel shadow above multiple lines "
             "with extra spaces and more.", set([])))

    def test_extract_commit_manual_merge(self):
        """Extra commit and bugs from a commit message manually merged."""
        self.assertEquals(
            branchhandling._extract_commit_bugs(
                """  UnityWindow: don't draw the panel shadow above
            multiple

            lines      with extra spaces
            and more."""),
            ("UnityWindow: don't draw the panel shadow above multiple lines "
             "with extra spaces and more.", set([])))

    def test_collect_author_commits_starting_with_dash(self):
        """Collect author commits and bugs when the msg starts with '-'."""
        with open(os_path_join_safe(self.data_dir,
                                    'bzrlogs',
                                    'withdashlogdiff')) as f:
            self.assertEquals(
                branchhandling.collect_author_commits(f.read(), set()),
                ({'Diego Sarmentero':
                    ["Update download manager API (BUG: #1224538). "
                        "(LP: #1224538)"],
                  'Manuel de la Pena':
                  ["Update download manager API (BUG: #1224538). "
                      "(LP: #1224538)"]},
                 set([1224538])))

    def test_collect_author_commits_starting_with_star(self):
        """Collect author commits and bugs when the msg starts with '*'."""
        with open(os_path_join_safe(self.data_dir,
                                    'bzrlogs',
                                    'withstarlogdiff')) as f:
            self.assertEquals(
                branchhandling.collect_author_commits(f.read(), set()),
                ({'Jim Hodapp':
                    ["Drop support for thumbnail as we can't depend on "
                     "gstreamer directly until qtmultimedia supports gst1.0 "
                     "(currently only the -touch fork supports it). Remove "
                     "gstreamer related packages from build-dependencies."], },
                 set()))

    def test_collect_author_commits_regular(self):
        """Collect author commits and bugs for a normal bzr log."""
        with open(os_path_join_safe(self.data_dir,
                                    'bzrlogs',
                                    'classiclogdiff')) as f:
            self.assertEquals(
                branchhandling.collect_author_commits(f.read(), set()),
                ({'Sebastien Bacher':
                    ["Use '%s:' string for preview hints, rather than just "
                        "appending ':'. (LP: #1074038)"],
                  'Brandon Schaefer':
                    ['EdgeBarrierController: add multiple types of subscriber '
                     'results, fix edges on autohide'],
                  'Marco Trevisan (Trevi\xc3\xb1o)':
                    ['Simulating direct commit to trunk',
                     'EdgeBarrierController: add multiple types of subscriber '
                     'results, fix edges on autohide',
                     'IconRenderer: preprocess an icon if its emblem has been '
                     'shown/hidden In that way we can '
                     'update its transformation. (LP: #1171476, #1171663)',
                     "UnityWindow: don't draw the panel shadow above a "
                     "fullscreen window. (LP: #1171934)"],
                  "\xc5\x81ukasz 'sil2100' Zemczak":
                    ['Now that we\'re using the new HUD, there have been some '
                     'changes that typically cause test_hud tests to fail. '
                     'Fix the tests to fit the new model. The first one is '
                     'that generally we do not have indicator entries visible '
                     'in the HUD anymore. Only application menu entries are '
                     'in it now. The second one - the way the results are '
                     'displayed is different. Now, instead of "Menu > Entry" '
                     'we have "Entry (Menu)" etc.'],
                  'Andrea Azzarone':
                    ['Disable detail view for webapp icons. (LP: #1169340)']},
                 set([1171476, 1074038, 1169340, 1171934, 1171663])))

    def test_collect_author_commits_with_some_bugs_to_ignore(self):
        """Ignore some bugs from a bzr log."""
        with open(os_path_join_safe(self.data_dir,
                                    'bzrlogs',
                                    'classiclogdiff')) as f:
            self.assertEquals(
                branchhandling.collect_author_commits(
                    f.read(),
                    set([1169340, 1064945])),
                ({'Sebastien Bacher':
                    ["Use '%s:' string for preview hints, rather than just "
                        "appending ':'. (LP: #1074038)"],
                  'Brandon Schaefer':
                    ['EdgeBarrierController: add multiple types of subscriber '
                     'results, fix edges on autohide'],
                  'Marco Trevisan (Trevi\xc3\xb1o)':
                    ['Simulating direct commit to trunk',
                     'EdgeBarrierController: add multiple types of subscriber '
                     'results, fix edges on autohide',
                     'IconRenderer: preprocess an icon if its emblem has been '
                     'shown/hidden In that way we '
                     'can update its transformation. (LP: #1171476, #1171663)',
                     "UnityWindow: don't draw the panel shadow above a "
                     "fullscreen window. (LP: #1171934)"],
                  "\xc5\x81ukasz 'sil2100' Zemczak":
                    ['Now that we\'re using the new HUD, there have been some '
                     'changes that typically cause test_hud tests to fail. '
                     'Fix the tests to fit the new model. The first one is '
                     'that generally we do not have indicator entries visible '
                     'in the HUD anymore. Only application menu entries are '
                     'in it now. The second one - the way the results are '
                     'displayed is different. Now, instead of "Menu > Entry" '
                     'we have "Entry (Menu)" etc.']},
                 set([1171476, 1074038, 1171934, 1171663])))

    def test_collect_author_commits_with_some_commits_to_ignore(self):
        """Ignore some commits from a bzr log."""
        with open(os_path_join_safe(self.data_dir,
                                    'bzrlogs',
                                    'logswithignoredchangelog')) as f:
            self.assertEquals(
                branchhandling.collect_author_commits(f.read(), set()),
                ({'Brandon Schaefer':
                    ['EdgeBarrierController: add multiple types of subscriber '
                     'results, fix edges on autohide'],
                  'Marco Trevisan (Trevi\xc3\xb1o)':
                    ['EdgeBarrierController: add multiple types of subscriber '
                     'results, fix edges on autohide',
                     'IconRenderer: preprocess an icon if its emblem has been '
                     'shown/hidden In that way we can update its '
                     'transformation. (LP: #1171476, #1171663)']},
                 set([1171476, 1171663])))

    def test_collect_author_with_launchpad_bot_commits(self):
        """Ignore launchpad bot commits."""
        with open(os_path_join_safe(self.data_dir,
                                    'bzrlogs',
                                    'logswithlaunchpadcommits')) as f:
            self.assertEquals(
                branchhandling.collect_author_commits(f.read(), set()),
                ({'Sebastien Bacher':
                    ["Use '%s:' string for preview hints, rather than just "
                        "appending ':'. (LP: #1074038)"],
                  'Brandon Schaefer':
                    ['EdgeBarrierController: add multiple types of subscriber '
                        'results, fix edges on autohide'],
                  'Marco Trevisan (Trevi\xc3\xb1o)':
                    ['Simulating direct commit to trunk',
                     'EdgeBarrierController: add multiple types of subscriber '
                     'results, fix edges on autohide',
                     'IconRenderer: preprocess an icon if its emblem has been '
                     'shown/hidden In that way we can '
                     'update its transformation. (LP: #1171476, #1171663)',
                     "UnityWindow: don't draw the panel shadow above a "
                     "fullscreen window. (LP: #1171934)"],
                  "\xc5\x81ukasz 'sil2100' Zemczak":
                    ['Now that we\'re using the new HUD, there have been some '
                     'changes that typically cause test_hud tests to fail. '
                     'Fix the tests to fit the new model. The first one is '
                     'that generally we do not have indicator entries visible '
                     'in the HUD anymore. Only application menu entries are '
                     'in it now. The second one - the way the results are '
                     'displayed is different. Now, instead of "Menu > Entry" '
                     'we have "Entry (Menu)" etc.'],
                  'Andrea Azzarone':
                  ['Disable detail view for webapp icons. (LP: #1169340)']},
                 set([1171476, 1074038, 1169340, 1171934, 1171663])))
