import pep8
import glob
import subprocess

from xml.dom.minidom import parse

from cupstream2distro.utils import os_path_join_safe

from tests.unit import DirectoryAwareTestCase


def call(args):
    return subprocess.Popen(
        args, stderr=subprocess.PIPE, stdout=subprocess.PIPE
    ).communicate()


class StaticCodeAnalysisTestCase(DirectoryAwareTestCase):
    """Static code analysis helps discover common mistakes.

    This module can identify syntax errors in python and xml before they
    get committed to trunk.
    """
    def setUp(self):
        """Identify the locations of all the python files to test."""
        super(StaticCodeAnalysisTestCase, self).setUp()
        self.trainfiles = glob.glob(os_path_join_safe(self.train_dir, '*.py'))
        self.cupfiles = glob.glob(os_path_join_safe(self.cup_dir, '*.py'))
        self.unitfiles = glob.glob(os_path_join_safe(
            self.test_dir, 'unit', '*.py'))
        self.scriptfiles = [os_path_join_safe(self.root_dir, 'copy2distro')]
        self.pyfiles = (self.trainfiles + self.cupfiles +
                        self.unitfiles + self.scriptfiles)

    def test_pep8_conformance(self):
        """PEP8 enforcement helps ensure our code follows a readable style."""
        pep8style = pep8.StyleGuide(show_source=True)
        pep8style.check_files(self.pyfiles)
        self.assertEqual(
            pep8style.options.report.total_errors,
            0,
            "Unexpected output from pep8.")

    def test_pyflakes_conformance(self):
        """Identify all kinds of bad things.

        PyFlakes can identify variables assigned to but never used, unused
        imports, variables referenced before assignment, and syntax errors.
        """
        output = ''.join(call(['pyflakes'] + self.pyfiles))
        print(output)
        self.assertEqual(len(output), 0, "Unexpected output from pyflakes.")

    def test_two_to_three_conformance(self):
        """Identify python2-isms that will be hard to port to python3."""
        stdout, stderr = call(['2to3', '-x', 'future'] + self.pyfiles)
        print(stdout)
        self.assertEqual(
            len(stdout), 0,
            'It looks like you wrote some code that is only compatible with '
            'Python 2. Please consider applying this diff or come up with '
            'an alternate solution that will work in both Python 2 and 3.')

    def test_xml_validity(self):
        """Ensure XML is syntactically correct.

        Does not test whether the XML is valid Jenkins however.
        """
        fileglob = os_path_join_safe(
            self.train_dir, 'jenkins-templates', '*.tmpl')
        for x in glob.glob(fileglob):
            try:
                parse(x)
            except Exception as e:
                self.fail("{}: {}".format(x, e))
