import os

from mock import Mock, patch

from tests.unit import DirectoryAwareTestCase
from cupstream2distro.launchpadmanager import lp, get_launchpad
from cupstream2distro.utils import env


class LaunchpadManagerTestCase(DirectoryAwareTestCase):
    def setUp(self):
        """Initialize a SiloState instance for us to poke at."""
        super(LaunchpadManagerTestCase, self).setUp()
        lp._instance = Mock()

    @patch('cupstream2distro.launchpadmanager.Launchpad')
    def test_get_launchpad(self, lpMock):
        """Test we connect to LP with the correct options."""
        env.JOB_NAME = 'testing'
        self.assertEqual(get_launchpad(), lpMock.login_with.return_value)
        lpMock.login_with.assert_called_once_with(
            service_root='production',
            launchpadlib_dir=os.path.expanduser(
                '~/cu2d/launchpad.cache/testing'),
            allow_access_levels=['WRITE_PRIVATE'],
            application_name='cupstream2distro',
            version='devel')

    @patch('cupstream2distro.launchpadmanager.os')
    @patch('cupstream2distro.launchpadmanager.Launchpad')
    def test_get_launchpad_differently(self, lpMock, osMock):
        """Test we connect to LP with other correct options."""
        env.JOB_NAME = ''
        osMock.path.exists.return_value = True
        self.assertEqual(get_launchpad(True), lpMock.login_with.return_value)
        lpMock.login_with.assert_called_once_with(
            service_root='staging',
            launchpadlib_dir=None,
            allow_access_levels=['WRITE_PRIVATE'],
            credentials_file=os.path.expanduser('~/.launchpad.credentials'),
            application_name='cupstream2distro',
            version='devel')

    @patch('cupstream2distro.launchpadmanager.get_launchpad')
    def test_lp_property(self, lpMock):
        """Ensure lazy-loading doesn't try to connect to LP twice."""
        lp._instance = None
        self.assertEqual(lp._launchpad, lpMock.return_value)
        self.assertEqual(lp._instance, lpMock.return_value)
        lpMock.assert_called_once_with()
        self.assertEqual(lp._launchpad, lpMock.return_value)
        lpMock.assert_called_once_with()

    def test_lp_getattr(self):
        """Undefined attributes are passed through to the _instance."""
        lp.kwyjibo('BART!')
        lp._instance.kwyjibo.assert_called_once_with('BART!')

    def test_lp_get_ppa(self):
        """We call the right methods to get PPAs from LP."""
        people = lp._instance.people.__getitem__ = Mock()
        distro = lp._instance.distributions.__getitem__ = Mock(
            side_effect=lambda x: x)
        ppa = lp.get_ppa('robru/ppa')
        distro.assert_called_once_with('ubuntu')
        people.assert_called_once_with('robru')
        people.return_value.getPPAByName.assert_called_once_with(
            name='ppa', distribution='ubuntu')
        self.assertEqual(ppa, lp._instance.people['robru'].getPPAByName(
            name='ppa', distribution='ubuntu'))

    def test_lp_get_ppa_rtm(self):
        """We call the right methods to get RTM PPAs from LP."""
        people = lp._instance.people.__getitem__ = Mock()
        distro = lp._instance.distributions.__getitem__ = Mock(
            side_effect=lambda x: x)
        ppa = lp.get_ppa('dudeman/ubuntu-rtm/stuff')
        distro.assert_called_once_with('ubuntu-rtm')
        people.assert_called_once_with('dudeman')
        people.return_value.getPPAByName.assert_called_once_with(
            name='stuff', distribution='ubuntu-rtm')
        self.assertEqual(ppa, lp._instance.people['dudeman'].getPPAByName(
            name='stuff', distribution='ubuntu-rtm'))

    def test_lp_is_distro_archive(self):
        """Correctly identify if a series object is a distro archive."""
        series = Mock()
        series.distribution.main_archive_link = 'alpha'
        series.self_link = 'alpha'
        self.assertTrue(lp.is_distro_archive(series))
        series.self_link = 'beta'
        self.assertFalse(lp.is_distro_archive(series))

    def test_lp_is_link_distro_archive(self):
        """Correctly identify if a series URL is a distro archive."""
        self.assertFalse(lp.is_link_distro_archive('example.com'))
        lp._instance.load.assert_called_once_with('example.com')

    def test_lp_load_url(self):
        """Correctly convert MP URLs into API URLs."""
        lp._instance.resource_type_link = \
            'https://api.launchpad.net/devel/#service-root'
        self.assertEqual(lp._instance.load.return_value, lp.load_url(
            'https://code.launchpad.net/~laney/autopilot-gtk/'
            'test-dep-theme/+merge/245112'))
        lp._instance.load.assert_called_once_with(
            'https://api.launchpad.net/devel/~laney/autopilot-gtk/'
            'test-dep-theme/+merge/245112')

    def test_lp_get_series_from_name(self):
        """Correctly get a series archive object from a distro name."""
        distro = lp._instance.distributions.__getitem__ = Mock()
        self.assertEqual(
            lp.get_series('warty', distro_name='ubuntu'),
            distro.return_value.getSeries.return_value)
        distro.assert_called_once_with('ubuntu')
        distro.return_value.getSeries.assert_called_once_with(
            name_or_version='warty')

    def test_lp_get_series_from_dest(self):
        """Correctly get a series archive object from an archive."""
        dest = Mock()
        self.assertEqual(
            lp.get_series('vivid', dest=dest),
            lp._instance.load.return_value.getSeries.return_value)
        lp._instance.load.assert_called_once_with(dest.distribution_link)
        lp._instance.load.return_value.getSeries.assert_called_once_with(
            name_or_version='vivid')

    def test_lp_open_bugs_for_source(self):
        """Sync upstream and downstream bugs."""
        source = 'foopackage'
        series = Mock()
        dest = Mock()
        bugMock = lp._instance.bugs.__getitem__ = Mock()
        lp.open_bugs_for_source(('1',), source, series, dest)
        lp._instance.load.assert_called_once_with(dest.distribution_link)
        bugMock.assert_called_once_with('1')
        bugMock.return_value.addTask.assert_called_once_with(
            target=series.getSourcePackage.return_value)
        bugMock.return_value.lp_save.assert_called_once_with()
        series.getSourcePackage.assert_called_once_with(name='foopackage')

    def test_lp_open_bugs_for_source_from_dest(self):
        """Sync upstream and downstream bugs."""
        source = 'foopackage'
        series = Mock()
        dest = Mock()
        bugMock = lp._instance.bugs.__getitem__ = Mock()
        lp._instance.load.return_value.current_series.name = 'vivid'
        series.name = 'vivid'
        lp.open_bugs_for_source(('1',), source, series, dest)
        lp._instance.load.assert_called_once_with(dest.distribution_link)
        bugMock.assert_called_once_with('1')
        bugMock.return_value.addTask.assert_called_once_with(
            target=lp._instance.load.return_value
            .getSourcePackage.return_value)
        bugMock.return_value.lp_save.assert_called_once_with()
        lp._instance.load.return_value.getSourcePackage\
            .assert_called_once_with(name='foopackage')

    @patch('cupstream2distro.launchpadmanager.logging')
    def test_lp_open_bugs_for_source_exception(self, logMock):
        """Catch exceptions from lp bugs."""
        series = Mock()
        dest = Mock()
        bugMock = lp._instance.bugs.__getitem__ = Mock()
        bugMock.side_effect = KeyError('Bug not found')
        lp.open_bugs_for_source(('1',), 'foo', series, dest)
        logMock.info.assert_called_once_with(
            'Failed to sync upstream/downstream bugs for bug #1.')
