# -*- coding: utf-8 -*-
# Copyright: (C) 2013-2014 Canonical
#
# Authors:
#  Didier Roche
#  Rodney Dawes
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

import os
import shutil

from six.moves.urllib.parse import quote
from six.moves import cStringIO as StringIO
from datetime import datetime
from mock import Mock, patch, call

from tests.unit import BaseUnitTestCase
from tests import strings as s

from cupstream2distro import packagemanager
from cupstream2distro.utils import env, os_path_join_safe


class PackageManagerTests(BaseUnitTestCase):
    def setUp(self):
        """Set SILO_DIR to tempdir so we can create files safely."""
        super(PackageManagerTests, self).setUp()
        packagemanager.SILO_DIR = lambda: self.tempdir

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_current_version_for_series_distro(self, lpMock):
        """Get the most recent publication in any pocket."""
        source1 = Mock()
        source1.source_package_version = '83.09.14-0ubuntu1'
        source1.date_created = datetime(2014, 1, 1)
        source1.status = 'Published'
        source2 = Mock()
        source2.source_package_version = '83.09.13-0ubuntu1'
        source2.date_created = datetime(2014, 2, 1)
        source2.status = 'Published'
        dest = lpMock.main_archive
        dest.getPublishedSources.return_value = [source1, source2]

        return_version = packagemanager.get_current_version_for_series(
            'foo', 'rolling')

        self.assertEquals(lpMock.get_ppa.call_count, 0)
        lpMock.get_series.assert_called_with('rolling', dest=dest)
        lpMock.main_archive.getPublishedSources.assert_called_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value)
        self.assertEquals('83.09.13-0ubuntu1', return_version)

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_current_version_for_series_none_in_distro(self, lpMock):
        """Get the newest version, but was never published in distro."""
        dest = lpMock.main_archive
        dest.getPublishedSources.return_value = []

        return_version = packagemanager.get_current_version_for_series(
            'foo', 'rolling')

        self.assertEquals(lpMock.get_ppa.call_count, 0)
        lpMock.get_series.assert_called_with('rolling', dest=dest)
        lpMock.main_archive.getPublishedSources.assert_called_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value)
        self.assertEquals('0', return_version)

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_current_version_for_series_ppa(self, lpMock):
        """Get the version from a ppa."""
        source_in_ppa = Mock()
        source_in_ppa.source_package_version = '83.09.13-0ubuntu1'
        source_in_ppa.status = 'Published'
        dest = lpMock.get_ppa.return_value
        dest.getPublishedSources.return_value = [source_in_ppa]

        return_version = packagemanager.get_current_version_for_series(
            'foo', 'rolling', 'didppa')

        lpMock.get_series.assert_called_with('rolling', dest=dest)
        lpMock.get_ppa.assert_called_once_with('didppa')
        lpMock.get_ppa.return_value.getPublishedSources.assert_called_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value)
        self.assertEquals('83.09.13-0ubuntu1', return_version)

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_current_version_for_series_ppa_ignoring_invalid(self, lpMock):
        """Make sure to ignore Deleted, Obsolete and Superseded packages."""
        source1 = Mock()
        source1.source_package_version = '83.09.13-0ubuntu4'
        source1.status = 'Deleted'
        source2 = Mock()
        source2.source_package_version = '83.09.13-0ubuntu3'
        source2.status = 'Obsolete'
        source3 = Mock()
        source3.source_package_version = '83.09.13-0ubuntu2'
        source3.status = 'Superseded'
        source4 = Mock()
        source4.source_package_version = '83.09.13-0ubuntu1'
        source4.status = 'Published'

        lpMock.get_ppa.return_value.getPublishedSources.return_value = [
            source1, source2, source3, source4]

        return_version = packagemanager.get_current_version_for_series(
            'foo', 'rolling', 'didppa')
        self.assertEquals('83.09.13-0ubuntu1', return_version)

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_current_version_for_series_none_in_ppa(self, lpMock):
        """Get the newest version, but was never published in ppa."""
        dest = lpMock.get_ppa.return_value
        dest.getPublishedSources.return_value = []

        return_version = packagemanager.get_current_version_for_series(
            'foo', 'rolling', 'didppa')

        lpMock.get_series.assert_called_with('rolling', dest=dest)
        lpMock.get_ppa.assert_called_once_with('didppa')
        lpMock.get_ppa.return_value.getPublishedSources.assert_called_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value)
        self.assertEquals('0', return_version)

    def test_lower_version(self):
        """Matching expectations for different cases of lower/upper version."""
        self.assertTrue(
            packagemanager.is_version1_higher_than_version2(
                '2-0ubuntu1',
                '1-0ubuntu1'))
        self.assertTrue(
            packagemanager.is_version1_higher_than_version2(
                '2-0ubuntu1',
                '2~daily13.10.1-0ubuntu1'))
        self.assertFalse(
            packagemanager.is_version1_higher_than_version2(
                '2-0ubuntu1',
                '2daily13.10.1-0ubuntu1'))
        self.assertTrue(
            packagemanager.is_version1_higher_than_version2(
                '2dailyrelease13.10.1.1-0ubuntu1',
                '2dailyrelease13.10.1-0ubuntu1'))

    def test_is_version_in_changelog_found(self):
        """We find the desired version from changelog."""
        self.assertTrue(packagemanager.is_version_in_changelog(
            '42.0daily83.09.13-0ubuntu1', StringIO(s.NESTED_CHANGELOG)))

    def test_unknown_version_is_not_in_changelog(self):
        """We don't find any unexisting version in changelog."""
        self.assertFalse(packagemanager.is_version_in_changelog(
            'version_which_dont_exist', StringIO(s.NESTED_CHANGELOG)))

    def test_unreleased_version_in_changelog(self):
        """Return false if the version we pass is UNRELEASED in changelog."""
        self.assertFalse(packagemanager.is_version_in_changelog(
            '42.0daily83.09.13-0ubuntu2', StringIO(s.NESTED_CHANGELOG)))

    def test_never_released_version_in_changelog(self):
        """We return true if the version we pass is 0."""
        self.assertTrue(packagemanager.is_version_in_changelog(
            '0', StringIO(s.NESTED_CHANGELOG)))

    def test_is_version_with_symbols_in_changelog_found(self):
        """We find the desired version, containing symbols from changelog."""
        self.assertTrue(packagemanager.is_version_in_changelog(
            '42.0daily83.09.13-0ubuntu1', StringIO(s.NESTED_CHANGELOG)))

    def test_get_latest_upstream_bzr_rev_in_last_changelog(self):
        """Always get the latest upstream bzr rev version from a changelog."""
        self.assertEquals(10, packagemanager.get_latest_upstream_bzr_rev(
            StringIO(s.RELEASED_CHANGELOG)))

    def test_get_latest_upstream_bzr_rev_with_only_bootstrap(self):
        """Get the latest bzr version from the bootstrap version."""
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'distro_with_only_bootstrap')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(f), 42)

    def test_get_latest_upstream_bzr_rev_in_previous_changelog(self):
        """Get bzr changelog version when marker not in newest changelog."""
        self.assertEquals(3, packagemanager.get_latest_upstream_bzr_rev(
            StringIO(s.CHANGEBETWEEN_CHANGELOG)))

    def test_get_latest_upstream_bzr_rev_with_two_in_changelog(self):
        """We always get the latest upstream bzr rev version from a changelog.

        We have two marker in the changelog, last one is taken.
        """
        self.assertEquals(42, packagemanager.get_latest_upstream_bzr_rev(
            StringIO(s.TWO_MARKERS_CHANGELOG)))

    def test_get_latest_upstream_bzr_rev_with_dest_ppa_without_distro(self):
        """We always get the latest upstream bzr rev version.

        When we have a dest ppa without a distro version.
        """
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'destppa_without_distro')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(
                    f,
                    'ubuntu-unity/next'),
                8)

    def test_get_latest_upstream_bzr_rev_with_dest_ppa_and_distro(self):
        """We always get the latest upstream bzr rev version.

        When we have a dest ppa and a distro version.
        """
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'destppa_with_distro')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(
                    f,
                    'ubuntu-unity/next'),
                42)

    def test_get_latest_upstream_bzr_rev_with_dest_ppa_and_distro_first(self):
        """We always get the latest upstream bzr rev version.

        When we have a dest ppa and a distro version appearing first in the
        result.
        """
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'destppa_with_distro_first')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(
                    f,
                    'ubuntu-unity/next'),
                8)

    def test_latest_bzr_rev_with_dest_ppa_distro_without_dest_ppa_rev(self):
        """We always get the latest upstream bzr rev version.

        When we have a dest ppa and a distro version without dest_ppa rev.
        """
        with open(os_path_join_safe(
                self.changelogs_file_dir,
                'destppa_with_distro_without_dest_ppa')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(
                    f,
                    'ubuntu-unity/next'),
                42)

    def test_get_latest_upstream_bzr_rev_with_dest_ppa_with_2_versions(self):
        """We always get the latest upstream bzr rev version.

        When we have a dest ppa with 2 marker version in the same changelog.
        """
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'destppa_with_2_versions')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(
                    f,
                    'ubuntu-unity/next'),
                43)

    def test_get_latest_upstream_bzr_rev_without_commit_tag(self):
        """We will get rev 0 if we don't have any commit tag."""
        self.assertEquals(0, packagemanager.get_latest_upstream_bzr_rev(
            StringIO(s.DUMMY_CHANGELOG)))

    def test_latest_upstream_bzr_rev_with_dest_ppa_marker_on_two_lines(self):
        """We always get the latest upstream bzr rev version.

        When we have a dest ppa with a marker shown on 2 lines.
        """
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'destppa_with_marker_two_lines')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(
                    f,
                    'ubuntu-unity/experimental-certified'),
                42)

    def test_get_latest_upstream_bzr_rev_with_dest_ppa_with_another_ppa_marker(
            self):
        """We always get the latest upstream bzr rev version.

        When we have a dest ppa with another dest ppa marker.
        """
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'destppa_with_another_dest_ppa')) as f:
            self.assertEquals(
                packagemanager.get_latest_upstream_bzr_rev(
                    f,
                    'ubuntu-unity/next'),
                8)

    def test_list_packages_info_in_str(self):
        """We return the packages info in a string."""
        package1 = Mock()
        package1.source_name = 'foo'
        package1.version = '42'
        package2 = Mock()
        package2.source_name = 'bar'
        package2.version = '44'
        packages_set = set([package1, package2])
        # sets are accessed randomly, so just ensure we have our strings in it.
        self.assertIn(
            'foo (42)',
            packagemanager.list_packages_info_in_str(packages_set))
        self.assertIn(
            'bar (44)',
            packagemanager.list_packages_info_in_str(packages_set))

    def test_get_packaging_version(self):
        """Get latest packaging version."""
        self.prep_debian_file('changelog', s.NESTED_CHANGELOG)
        self.assertEquals(
            packagemanager.get_packaging_version(self.tempdir),
            '42.0daily83.09.13-0ubuntu2')

    def test_list_packages_info_in_str_no_package(self):
        """We return no packaging info if we get no package parameter."""
        self.assertEquals(packagemanager.list_packages_info_in_str(set()), '')

    def get_source_files_for_package(self, package_name, for_download=False):
        """Small tool to get a source package files in package_name dir."""
        result_files_path = []
        package_source_dir = self.get_ubuntu_source_content_path(package_name)
        for filename in os.listdir(package_source_dir):
            if for_download:
                filename = quote(filename)
            result_files_path.append(
                os_path_join_safe(package_source_dir, filename))
        return result_files_path

    def get_dsc_for_package(self, package_name):
        """Only return the .dsc file for a package source name."""
        for file in self.get_source_files_for_package(package_name):
            if file.endswith('.dsc'):
                return file

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest(self, lpMock):
        """We grab the correct source from dest."""
        env.SILONAME = 'ubuntu/landing-234'
        dest = Mock()
        dest.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        dest.getPublishedSources.return_value = [source1]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_package', for_download=True))

        source_package_dir = packagemanager.get_source_package_from_dest(
            'foo', dest, '42.0daily83.09.13.2-0ubuntu1', 'rolling')

        lpMock.get_series.assert_called_with('rolling', dest=dest)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='42.0daily83.09.13.2-0ubuntu1')
        source1.sourceFileUrls.assert_called_once()
        self.assertTrue(os.path.isdir(source_package_dir))

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_two_foo_packages(self, lpMock):
        """Only unpack the right source package by version."""
        env.SILONAME = 'ubuntu/landing-123'
        dest1 = Mock()
        dest1.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        dest2 = Mock()
        dest2.distribution.name = 'ubuntu'
        source2 = Mock()
        source2.status = 'Published'
        dest1.getPublishedSources.return_value = [source1]
        dest2.getPublishedSources.return_value = [source2]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_package', for_download=True))
        source2.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_package_with_upstream_changes', for_download=True))

        source_package_dir1 = packagemanager.get_source_package_from_dest(
            'foo', dest1, '42.0daily83.09.13.2-0ubuntu1', 'rolling')
        self.assertTrue(os.path.isdir(source_package_dir1))

        # to make sure that it's not extracting both .dsc's during the
        # get_source_package_from_dest step
        shutil.rmtree(source_package_dir1)

        source_package_dir2 = packagemanager.get_source_package_from_dest(
            'foo', dest2, '42.0+13.10.20130709-0ubuntu1', 'rolling')
        self.assertFalse(os.path.isdir(source_package_dir1))
        self.assertTrue(os.path.isdir(source_package_dir2))

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_multiple_candidates(self, lpMock):
        """We grab the correct source from dest, even if more than one."""
        env.SILONAME = 'hello'
        dest = Mock()
        dest.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        source1.date_created = datetime(2014, 1, 1)
        source2 = Mock()
        source2.status = 'Published'
        source1.date_created = datetime(2014, 2, 1)
        dest.getPublishedSources.return_value = [source1, source2]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package('foo_package'))

        source_package_dir = packagemanager.get_source_package_from_dest(
            'foo', dest, '42.0daily83.09.13.2-0ubuntu1', 'rolling')

        lpMock.get_series.assert_called_with('rolling', dest=dest)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='42.0daily83.09.13.2-0ubuntu1')
        source1.sourceFileUrls.assert_called_once()
        self.assertFalse(source2.sourceFileUrls.called)
        self.assertTrue(os.path.isdir(source_package_dir))

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_with_epoc(self, lpMock):
        """We grab the correct source from dest with epoch."""
        env.SILONAME = 'i-am-an-invalid-silo-name-but-thats-ok'
        dest = Mock()
        dest.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        dest.getPublishedSources.return_value = [source1]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_package', for_download=True))

        source_package_dir = packagemanager.get_source_package_from_dest(
            'foo', dest, '1:42.0daily83.09.13.2-0ubuntu1', 'rolling')

        lpMock.get_series.assert_called_with('rolling', dest=dest)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='1:42.0daily83.09.13.2-0ubuntu1')
        source1.sourceFileUrls.assert_called_once()
        self.assertTrue(os.path.isdir(source_package_dir))
        self.assertNotIn(':', source_package_dir)

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_for_native(self, lpMock):
        """We grab the correct source from dest for native packages."""
        env.SILONAME = 'frumpbuntu/flailing-987'
        dest = Mock()
        dest.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        dest.getPublishedSources.return_value = [source1]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_native_package', for_download=True))

        source_package_dir = packagemanager.get_source_package_from_dest(
            'foo', dest, '42.0daily83.09.13.2', 'rolling')

        lpMock.get_series.assert_called_with('rolling', dest=dest)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='42.0daily83.09.13.2')
        source1.sourceFileUrls.assert_called_once()
        self.assertTrue(os.path.isdir(source_package_dir))

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_with_debian_version(
            self, lpMock):
        """We grab the correct source from dest with only a debian version."""
        env.SILONAME = 'ubuntu/landing-000'
        dest = Mock()
        dest.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        dest.getPublishedSources.return_value = [source1]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_debian_package', for_download=True))

        source_package_dir = packagemanager.get_source_package_from_dest(
            'foo', dest, '42.0daily83.09.13.2-1', 'rolling')

        lpMock.get_series.assert_called_with('rolling', dest=dest)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='42.0daily83.09.13.2-1')
        source1.sourceFileUrls.assert_called_once()
        self.assertTrue(os.path.isdir(source_package_dir))

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_with_special_chars(
            self, lpMock):
        """Grab the correct source from dest with chars like ~ and +."""
        env.SILONAME = 'ubuntu/landing-000'
        dest = Mock()
        dest.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        dest.getPublishedSources.return_value = [source1]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_specialchars_package', for_download=True))

        source_package_dir = packagemanager.get_source_package_from_dest(
            'foo', dest, '42.0~daily83.09.13.2+0-0ubuntu1', 'rolling')

        lpMock.get_series.assert_called_once_with('rolling', dest=dest)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='42.0~daily83.09.13.2+0-0ubuntu1')
        source1.sourceFileUrls.assert_called_once()
        self.assertTrue(os.path.isdir(source_package_dir))

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_not_published(self, lpMock):
        """We return none if the package was never published into the dest."""
        self.assertIsNone(packagemanager.get_source_package_from_dest(
            'foo', None, '0', 'rolling'))
        self.assertEqual(os.listdir('.'), [])
        self.assertFalse(lpMock.get_series.called)

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_with_ubuntu_native_version(
            self, lpMock):
        """We grab the correct source from dest with ubuntu native version."""
        dest = Mock()
        dest.distribution.name = 'ubuntu'
        source1 = Mock()
        source1.status = 'Published'
        dest.getPublishedSources.return_value = [source1]
        source1.sourceFileUrls.return_value = (
            self.get_source_files_for_package(
                'foo_native_ubuntu_version', for_download=True))

        source_package_dir = packagemanager.get_source_package_from_dest(
            'foo', dest, '42ubuntu1', 'rolling')

        lpMock.get_series.assert_called_once_with('rolling', dest=dest)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='42ubuntu1')
        source1.sourceFileUrls.assert_called_once()
        self.assertTrue(os.path.isdir(source_package_dir))

    def test_release_with_no_ubuntu_upload(self):
        """We release if there has been no upload before."""
        self.assertTrue(
            packagemanager.is_new_content_relevant(self.tempdir, None))

    @patch('cupstream2distro.packagemanager.Popen')
    def test_release_if_content_committed_before_commit(self, popenMock):
        """Release if there is at least one change including an upstream."""
        popenMock.return_value.communicate.return_value = ('diffblorp', '')
        a = os_path_join_safe(self.tempdir, 'a')
        b = os_path_join_safe(self.tempdir, 'b')
        stdout = popenMock.return_value.stdout
        self.assertTrue(packagemanager.is_new_content_relevant(a, b))
        expected_calls = [
            call(['diff', '-Nrup', '.', b], cwd=a, stderr=-1, stdout=-1),
            call(['filterdiff', '--clean', '-x', '*po', '-x', '*pot', '-x',
                  '*local-options'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['lsdiff'], stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['grep', '-Ev', '.bzr|.pc'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(),
        ]
        popenMock.return_value.communicate.assert_called_once_with()
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(kall, popenMock.mock_calls[i])

    @patch('cupstream2distro.packagemanager.Popen')
    def test_release_if_new_content_with_manual_uploads(self, popenMock):
        """Release if there has been some manual uploads backported."""
        popenMock.return_value.communicate.return_value = (
            '> debian/control\ndebian/changelog\n', '')
        a = os_path_join_safe(self.tempdir, 'a')
        b = os_path_join_safe(self.tempdir, 'b')
        a_control = os_path_join_safe(a, 'debian', 'control')
        b_control = os_path_join_safe(b, 'debian', 'control')
        stdout = popenMock.return_value.stdout
        self.assertTrue(packagemanager.is_new_content_relevant(a, b))
        expected_calls = [
            call(['diff', '-Nrup', '.', b], cwd=a, stderr=-1, stdout=-1),
            call(['filterdiff', '--clean', '-x', '*po', '-x', '*pot', '-x',
                  '*local-options'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['lsdiff'], stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['grep', '-Ev', '.bzr|.pc'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call().communicate(),
            call(['diff', a_control, b_control], cwd=a, stderr=-1, stdout=-1),
            call(),
        ]
        self.assertEqual(popenMock.return_value.communicate.call_count, 2)
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(kall, popenMock.mock_calls[i])

    @patch('cupstream2distro.packagemanager.Popen')
    def test_dont_release_if_content_matches(self, popenMock):
        """Don't release if upstream and downstream content matches."""
        popenMock.return_value.communicate.return_value = ('', '')
        a = os_path_join_safe(self.tempdir, 'a')
        b = os_path_join_safe(self.tempdir, 'b')
        stdout = popenMock.return_value.stdout
        self.assertFalse(packagemanager.is_new_content_relevant(a, b))
        expected_calls = [
            call(['diff', '-Nrup', '.', b], cwd=a, stderr=-1, stdout=-1),
            call(['filterdiff', '--clean', '-x', '*po', '-x', '*pot', '-x',
                  '*local-options'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['lsdiff'], stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['grep', '-Ev', '.bzr|.pc'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(),
        ]
        popenMock.return_value.communicate.assert_called_once_with()
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(kall, popenMock.mock_calls[i])

    @patch('cupstream2distro.packagemanager.Popen')
    def test_dont_release_if_vcs_bzr_change_with_changelog(self, popenMock):
        """Don't release if only a vcs change, even with a changelog change."""
        popenMock.return_value.communicate.return_value = (
            'debian/control\ndebian/changelog\n', '')
        a = os_path_join_safe(self.tempdir, 'a')
        b = os_path_join_safe(self.tempdir, 'b')
        a_control = os_path_join_safe(a, 'debian', 'control')
        b_control = os_path_join_safe(b, 'debian', 'control')
        stdout = popenMock.return_value.stdout
        self.assertFalse(packagemanager.is_new_content_relevant(a, b))
        expected_calls = [
            call(['diff', '-Nrup', '.', b], cwd=a, stderr=-1, stdout=-1),
            call(['filterdiff', '--clean', '-x', '*po', '-x', '*pot', '-x',
                  '*local-options'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['lsdiff'], stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['grep', '-Ev', '.bzr|.pc'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call().communicate(),
            call(['diff', a_control, b_control], cwd=a, stderr=-1, stdout=-1),
            call(),
        ]
        self.assertEqual(popenMock.return_value.communicate.call_count, 2)
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(kall, popenMock.mock_calls[i])

    @patch('cupstream2distro.packagemanager.Popen')
    def test_release_even_if_changelog_change(self, popenMock):
        """Release even if the only change is a debian/changelog change."""
        popenMock.return_value.communicate.return_value = (
            'debian/changelog\n', '')
        a = os_path_join_safe(self.tempdir, 'a')
        b = os_path_join_safe(self.tempdir, 'b')
        stdout = popenMock.return_value.stdout
        self.assertTrue(packagemanager.is_new_content_relevant(a, b))
        expected_calls = [
            call(['diff', '-Nrup', '.', b], cwd=a, stderr=-1, stdout=-1),
            call(['filterdiff', '--clean', '-x', '*po', '-x', '*pot', '-x',
                  '*local-options'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['lsdiff'], stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(['grep', '-Ev', '.bzr|.pc'],
                 stdin=stdout, cwd=a, stderr=-1, stdout=-1),
            call(),
        ]
        popenMock.return_value.communicate.assert_called_once_with()
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(kall, popenMock.mock_calls[i])

    def test_package_wrongly_diffing(self):
        """Package only has the automatic bump and d/changelog has nothing."""
        for file in self.get_source_files_for_package('foo_package'):
            if not os.path.isdir(file):
                shutil.copy2(file, self.tempdir)
        dest_version_source = self.get_ubuntu_source_content_path(
            'ubuntu_foo_package_with_one_less_release')
        self.assertFalse(
            packagemanager.is_relevant(
                'foo_42.0daily83.09.13.2-0ubuntu1.dsc',
                dest_version_source))

    def test_package_rightly_diffing(self):
        """Detect if a package has changes justifying the daily release."""
        for file in self.get_source_files_for_package('foo_package'):
            if not os.path.isdir(file):
                shutil.copy2(file, self.tempdir)
        dest_version_source = self.get_ubuntu_source_content_path(
            'ubuntu_foo_package_with_two_less_release')
        self.assertTrue(
            packagemanager.is_relevant(
                'foo_42.0daily83.09.13.2-0ubuntu1.dsc',
                dest_version_source))

    def test_detect_packaging_changes_since_last_release(self):
        """We detect packaging changes since last release."""
        self.assertTrue(
            packagemanager._packaging_changes_between_dsc(
                self.get_dsc_for_package('foo_package'),
                self.get_dsc_for_package(
                    'foo_package_with_upstream_and_packaging_changes')))

    def test_with_upstream_and_changelog_only_change(self):
        """Detect no packaging change when changelog changed only."""
        self.assertFalse(
            packagemanager._packaging_changes_between_dsc(
                self.get_dsc_for_package('foo_package'),
                self.get_dsc_for_package('foo_package_with_upstream_changes')))

    def test_always_diff_when_no_previous_version_in_distro(self):
        """Detect packaging change if no previous version was in distro."""
        self.assertTrue(
            packagemanager._packaging_changes_between_dsc(
                None, self.get_dsc_for_package('foo_package')))

    def test_generate_diff(self):
        """We generate the right diff."""
        packagemanager.generate_diff_between_dsc(
            'foo.diff',
            self.get_dsc_for_package('foo_package'),
            self.get_dsc_for_package(
                'foo_package_with_upstream_and_packaging_changes'))
        canonical_filepath = os_path_join_safe(
            self.result_dir,
            'foo_package_foo_package_with_upstream_and_packaging_changes.diff')
        self.assertFilesAreIdenticals('foo.diff', canonical_filepath)

    def test_diff_warn_for_new_package(self):
        """We warn in the generate diff for new packages."""
        packagemanager.generate_diff_between_dsc(
            'foo.diff',
            None,
            self.get_dsc_for_package('foo_package'))
        canonical_filepath = os_path_join_safe(
            self.result_dir, 'new_foo_package.diff')
        self.assertFilesAreIdenticals('foo.diff', canonical_filepath)

    def test_diff_warn_for_new_binary_package(self):
        """We warn in the generate diff for new binary packages."""
        packagemanager.generate_diff_between_dsc(
            'foo.diff',
            self.get_dsc_for_package('foo_package'),
            self.get_dsc_for_package('foo_package_with_new_binary_package'))
        canonical_filepath = os_path_join_safe(
            self.result_dir,
            'foo_package_foo_package_with_new_binary_package.diff')
        self.assertFilesAreIdenticals('foo.diff', canonical_filepath)

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_regular(self, datetimeMock):
        """Create a new packaging version after a regular daily release."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830912-0ubuntu1',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_epoch(self, datetimeMock):
        """Create a new version after a daily release with epoch."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '1:42+13.10.19830912-0ubuntu1',
                '13.10'),
            '1:42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_manual_upload(
            self, datetimeMock):
        """Create a new version after an daily release with manual upload."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830912-0ubuntu3',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_first_rerelease_sameday(
            self, datetimeMock):
        """Create a new version after already released same day."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830913-0ubuntu1',
                '13.10'),
            '42+13.10.19830913.1-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_continous_rerelease_sameday(
            self, datetimeMock):
        """Create new packaging version after several released same day."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830913.2-0ubuntu1',
                '13.10'),
            '42+13.10.19830913.3-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_manual_upload_same_day(
            self, datetimeMock):
        """Create a new version after some manual uploads same day."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830913.2-0ubuntu3',
                '13.10'),
            '42+13.10.19830913.3-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_transition_releases(
            self, datetimeMock):
        """Create a new version for a new series."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.04.19830913-0ubuntu1',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_without_daily(self, datetimeMock):
        """Create a new version following a non-daily-release version."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42-0ubuntu1',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_native(self, datetimeMock):
        """Verify that native package versions are supported."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '13.10+13.10.19830912',
                '13.10'),
            '13.10+13.10.19830913')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_native_epoch(self, datetimeMock):
        """Verify that native package versions with epoch are supported."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '1:13.10+13.10.19830912',
                '13.10'),
            '1:13.10+13.10.19830913')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_native_first_rerelease_sameday(
            self, datetimeMock):
        """Verify that native versions re-release on same day is supported."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '13.10+13.10.19830913',
                '13.10'),
            '13.10+13.10.19830913.1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_native_continous_rerelease_sameday(
            self, datetimeMock):
        """Verify that native versions continuoed re-release is supported."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '13.10+13.10.19830913.2',
                '13.10'),
            '13.10+13.10.19830913.3')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_native_ppa_dest(self, datetimeMock):
        """Verify that native versions with a ppa dest are supported."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '13.10+13.10.19830912',
                '13.10',
                'didrocks/my-ppa_mine'),
            '13.10+13.10.19830913didrocks.my.ppa.mine')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_from_native(self, datetimeMock):
        """We create a new packaging version after a native version."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version('42', '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_sync_debian(self, datetimeMock):
        """Create a new packaging version after a debian-synced version."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version('42-1', '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_debian_based_manual(
            self, datetimeMock):
        """Create a new version after a manual based on a debian version."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42-1ubuntu2',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_debian_based_daily(
            self, datetimeMock):
        """Create a new version after a manual based on a debian version."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830912-1ubuntu2',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_with_ppa_dest(self, datetimeMock):
        """Create a new version after with a ppa destination."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830912-0ubuntu1',
                '13.10',
                'didrocks/my-ppa_mine'),
            '42+13.10.19830913didrocks.my.ppa.mine-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_with_ppa_dest_after_a_ppa_dest(
            self, datetimeMock):
        """Create a new version after another ppa_dest."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830912didrocks.my.ppa.mine-0ubuntu1',
                '13.10',
                'didrocks/my-ppa_mine'),
            '42+13.10.19830913didrocks.my.ppa.mine-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_new_packaging_version_after_a_ppa_dest_with_intermediate_version(
            self, datetimeMock):
        """Create new version after another ppa_dest released the same day."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830913didrocks.my.ppa.mine-0ubuntu1',
                '13.10',
                'didrocks/my-ppa_mine'),
            '42+13.10.19830913.1didrocks.my.ppa.mine-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_with_ppa_dest_to_distro(
            self, datetimeMock):
        """Create a new version after having a ppa destination to distro."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830912didrocks.my.ppa.mine-0ubuntu1',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_wrong_native(self, datetimeMock):
        """Create a new version after a wrong native with ubuntu in it."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42ubuntu1',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_from_old_format(self, datetimeMock):
        """Create a new version transitionned from previous daily format."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42daily83.09.12-1ubuntu2',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_from_old_format_complex(
            self, datetimeMock):
        """Create a new version transitionned with maintenance branch."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42daily83.09.12~13.10-1ubuntu2',
                '13.10'),
            '42+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_from_old_format_sameday(
            self, datetimeMock):
        """Create a new version transitionned from previous format same day."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '20830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42daily83.09.13~13.10-1ubuntu2',
                '13.10'),
            '42+13.10.20830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_from_old_format_sameday_with_minor(
            self, datetimeMock):
        """Create new version from previous with multiple releases same day."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '20830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42daily83.09.13.3~13.10-1ubuntu2',
                '13.10'),
            '42+13.10.20830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_new_packaging_version_from_old_format_with_maintenance_branch(
            self, datetimeMock):
        """Create a new version transitionned from previous with branch."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42daily83.09.12didrocks.my.ppa.mine-0ubuntu2',
                '13.10',
                'didrocks/my-ppa_mine'),
            '42+13.10.19830913didrocks.my.ppa.mine-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_new_packaging_version_for_rtm(
            self, datetimeMock):
        """Create a version for ubuntu-rtm."""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+13.10.19830912-0ubuntu1',
                '14.09'),
            '42+15.04.19830913~rtm-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    @patch('cupstream2distro.packagemanager.datetime')
    def test_create_new_packaging_version_with_bzr_rev(
            self, datetimeMock):
        """Create a new version for ubuntu when the upstream part contains
        a +bzr string"""
        strftime_call = datetimeMock.date.today.return_value.strftime
        strftime_call.side_effect = lambda date: '19830913'
        self.assertEqual(
            packagemanager.create_new_packaging_version(
                '42+bzr123+13.10.19830912-0ubuntu1',
                '13.10'),
            '42+bzr123+13.10.19830913-0ubuntu1')
        strftime_call.assert_called_with('%Y%m%d')

    def test_collect_bugs_simple(self):
        """Simple case: some bugs in an unreleased changelog."""
        self.assertEquals(
            packagemanager.collect_bugs(StringIO(s.NESTED_CHANGELOG), 'foo'),
            set(['56789', '67890', '34567890', '12345']))

    def test_collect_bugs_extras(self):
        """Extras syntax case: some bugs in an unreleased changelog."""
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'different_bugs_pattern')) as f:
            self.assertEquals(
                packagemanager.collect_bugs(f, 'foo'),
                set(['56789', '67890', '34567', '12345', '567890',
                     '678901', '123456', '1234567']))

    def test_collect_bugs_no_bugs(self):
        """No bugs, but an unreleased changelog."""
        self.assertEqual(set(), packagemanager.collect_bugs(
            StringIO(s.BASIC_CHANGELOG), 'foo'))

    def test_collect_bugs_no_unrelease_no_manual(self):
        """No manual upload, no unreleased changelog."""
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'no_unrelease_no_manual')) as f:
            self.assertEqual(packagemanager.collect_bugs(f, 'foo'), set())

    def test_collect_bugs_multiple_unrelease(self):
        """No bug from it in the first snapshot."""
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'multiple_unrelease')) as f:
            self.assertEqual(
                packagemanager.collect_bugs(f, 'foo'),
                set(['34567', '12345', '23456']))

    def test_collect_bugs_multiple_unrelease_manual_upload(self):
        """Only collect bugs until the last manual upload."""
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'multiple_unrelease_with_release')) as f:
            self.assertEqual(
                packagemanager.collect_bugs(f, 'foo'),
                set(['23456']))

    def test_collect_bugs_only_one_unreleased_changelog(self):
        """Collect the bugs from the one unreleased changelog."""
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'one_unreleased')) as f:
            self.assertEqual(
                packagemanager.collect_bugs(f, 'foo'),
                set(['12345']))

    def test_collect_bugs_only_one_unreleased_changelog_no_bug(self):
        """We have no bug from the only one unreleased changelog."""
        with open(os_path_join_safe(self.changelogs_file_dir,
                                    'one_unreleased_no_bug')) as f:
            self.assertEqual(packagemanager.collect_bugs(f, 'foo'), set())

    def test_update_changelog_simple(self):
        """Update a changelog from one author with a bug and bzr rev."""
        changelog = self.prep_debian_file('changelog', s.DUMMY_CHANGELOG)
        authors = {'Foo': ['One fix for LP: #12345']}
        packagemanager.update_changelog(
            self.tempdir, '1.2daily83.09.14-0ubuntu1', 'raring', 42, authors)
        result_file = os_path_join_safe(
            self.result_dir, 'simple_changelog_update')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_simple_no_author(self):
        """Update a changelog from a list with no bug fix."""
        changelog = self.prep_debian_file('changelog', s.DUMMY_CHANGELOG)
        packagemanager.update_changelog(
            self.tempdir, '1.2daily83.09.14-0ubuntu1', 'raring', 42, {})
        result_file = os_path_join_safe(
            self.result_dir, 'no_bug_changelog_update')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_one_author_multiple_bugs(self):
        """Update a changelog from a list of one author with multiple bugs."""
        changelog = self.prep_debian_file('changelog', s.DUMMY_CHANGELOG)
        authors = {
            'Foo': ['One fix for LP: #12345', 'Another fix for LP: #23456']
        }
        packagemanager.update_changelog(
            self.tempdir, '1.2daily83.09.14-0ubuntu1', 'raring', 42, authors)
        result_file = os_path_join_safe(
            self.result_dir, 'multiple_bugs_one_author_changelog_update')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_multiple_authors(self):
        """Update a changelog from multiple authors with multiple bugs."""
        changelog = self.prep_debian_file('changelog', s.DUMMY_CHANGELOG)
        authors = {
            'Foo': ['One fix for LP: #12345', 'Another fix for LP: #23456'],
            'Bar': ['Another fix for LP: #23456'],
            'Baz baz': ['and another Fix on LP: #34567']
        }
        packagemanager.update_changelog(
            self.tempdir, '1.2daily83.09.14-0ubuntu1', 'raring', 42, authors)
        result_file = os_path_join_safe(
            self.result_dir, 'multiple_authors_bugs_changelog_update')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_with_msg_starting_with_dash(self):
        """Update a changelog with a message starting with '-'."""
        changelog = self.prep_debian_file('changelog', s.BASIC_CHANGELOG)
        authors = {'Foo': ['- One fix for LP: #12345']}
        packagemanager.update_changelog(
            self.tempdir, '42.0daily83.09.14-0ubuntu1',
            'raring', 42, authors)
        result_file = os_path_join_safe(
            self.result_dir, 'message_starting_with_dash')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_with_existing_content(self):
        """Update a changelog when we already have existing content."""
        changelog = self.prep_debian_file('changelog', s.BASIC_CHANGELOG)
        authors = {
            'Foo': ['One fix for LP: #12345', 'Another fix for LP: #23456'],
            'Bar': ['Another fix for LP: #23456'],
            'Baz baz': ['and another Fix on LP: #34567']
        }
        packagemanager.update_changelog(
            self.tempdir, '42.0daily83.09.14-0ubuntu1', 'raring', 42, authors)
        result_file = os_path_join_safe(
            self.result_dir, 'existing_content_changelog_update')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_with_existing_content_existing_author(self):
        """Update a changelog when we already have existing author."""
        changelog = self.prep_debian_file('changelog', s.BASIC_CHANGELOG)
        authors = {
            'Didier Roche': [
                'One fix for LP: #12345', 'Another fix for LP: #23456'],
            'Bar': ['Another fix for LP: #23456'],
            'Baz baz': ['and another Fix on LP: #34567']
        }
        packagemanager.update_changelog(
            self.tempdir, '42.0daily83.09.14-0ubuntu1', 'raring', 42, authors)
        result_file = os_path_join_safe(
            self.result_dir, 'existing_content_author_changelog_update')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_with_existing_content_existing_multiple_authors(
            self):
        """Update a changelog with existing content and multiple authors."""
        changelog = self.prep_debian_file(
            'changelog', s.BASIC_MULTIPLE_CONTENTS)
        authors = {'Didier Roche': ['One fix for LP: #12345',
                                    'Another fix for LP: #23456'],
                   'Bar': ['Another fix for LP: #23456'],
                   'Baz baz': ['and another Fix on LP: #34567']}
        packagemanager.update_changelog(
            self.tempdir, '42.0daily83.09.14-0ubuntu1', 'raring', 42, authors)
        result_file = os_path_join_safe(
            self.result_dir,
            'existing_content_multiple_authors_changelog_update')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_update_changelog_with_dest_ppa(self):
        """Update a changelog from a list with a destination ppa."""
        changelog = self.prep_debian_file('changelog', s.DUMMY_CHANGELOG)
        packagemanager.update_changelog(
            self.tempdir,
            '1.2daily83.09.14.ubuntu.unity.next-0ubuntu1',
            'raring',
            42,
            {},
            'ubuntu-unity/next')
        result_file = os_path_join_safe(
            self.result_dir, 'changelog_with_dest_ppa')
        self.assertChangelogFilesAreIdenticals(result_file, changelog)

    def test_increment_ubuntu_package_version_ubuntu(self):
        """Increment the ubuntu version for a package with ubuntu changes."""
        version = '1.2-0ubuntu1'
        new_version = packagemanager.increment_ubuntu_package_version(version)
        self.assertEquals('1.2-0ubuntu2', new_version)

    def test_increment_ubuntu_package_version_debian(self):
        """Increment the ubuntu version for a package with debian changes."""
        version = '1.2-1'
        new_version = packagemanager.increment_ubuntu_package_version(version)
        self.assertEquals('1.2-1ubuntu1', new_version)

    def test_increment_ubuntu_package_version_native(self):
        """Increment the ubuntu version number for a native package version."""
        version = '1.2'
        new_version = packagemanager.increment_ubuntu_package_version(version)
        self.assertEquals('1.2ubuntu1', new_version)

    @patch('cupstream2distro.packagemanager.lp')
    def test_is_upstream_version_for_series_in_dest_with_ubuntu_versioning(
            self, lpMock):
        """Check if the archive has the upstream, strip off ubuntu version."""
        source_in_ppa = Mock()
        source_in_ppa.source_package_version = '1.2-0ubuntu2'
        source_in_ppa.source_package_name = 'foo'
        source_in_ppa.status = 'Published'
        dest = lpMock.get_ppa.return_value
        dest.getPublishedSources.return_value = [source_in_ppa]

        self.assertTrue(packagemanager.is_upstream_version_for_series_in_dest(
            'foo', '1.2-0ubuntu1', 'rolling', dest))

    def test_simple_upload(self):
        with patch('cupstream2distro.packagemanager.Popen') as popen_mock:
            popen_mock().returncode = 0
            popen_mock().communicate.return_value = ('', '')
            packagemanager.simple_upload('/path/to/changes', 'my/ubuntu/ppa')
            popen_mock.assert_called_with(
                ('dput', 'ppa:my/ubuntu/ppa', '/path/to/changes'),
                cwd='/path/to', stderr=-1, stdout=-1)

    def test_simple_upload_failed(self):
        with patch('cupstream2distro.packagemanager.Popen') as popen_mock:
            popen_mock().returncode = 1
            popen_mock().communicate.return_value = ('', '')
            with self.assertRaisesRegexp(Exception, 'returned 1'):
                packagemanager.simple_upload(
                    '/path/to/changes', 'my/ubuntu/ppa')
                popen_mock.assert_called_with(
                    ('dput', 'ppa:my/ubuntu/ppa', '/path/to/changes'),
                    cwd='/path/to', stderr=-1, stdout=-1)


class PackageManagerOfflineTests(BaseUnitTestCase):
    """System tests not depending on online services."""

    def test_refresh_symbols_files_simple(self):
        """Update the symbols file with the new version."""
        changelog = self.prep_debian_file('changelog', s.BASE_CHANGELOG)
        symbols = self.prep_debian_file('foo.symbols', s.BASIC_SYMBOLS)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        result_symbols = os_path_join_safe(
            self.result_dir, 'simple_update.symbols')
        result_changelog = os_path_join_safe(
            self.result_dir, 'simple_foo_update.changelog')
        self.assertFilesAreIdenticals(symbols, result_symbols)
        self.assertChangelogFilesAreIdenticals(result_changelog, changelog)

    def test_refresh_symbols_files_multiple_replace_tag_in_symbol(self):
        """Update symbols file replacing multiple symbols with new version."""
        changelog = self.prep_debian_file('changelog', s.BASE_CHANGELOG)
        symbols = self.prep_debian_file('foo.symbols', s.MULTIPLE_SYMBOLS)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        result_symbols = os_path_join_safe(
            self.result_dir, 'multiplesymbols_update.symbols')
        result_changelog = os_path_join_safe(
            self.result_dir, 'simple_foo_update.changelog')
        self.assertFilesAreIdenticals(symbols, result_symbols)
        self.assertChangelogFilesAreIdenticals(result_changelog, changelog)

    def test_refresh_symbols_files_simple_with_existing_changelog(self):
        """Update symbols file with the version in an existing changelog."""
        changelog = self.prep_debian_file(
            'changelog', s.BASIC_SYMBOLS_WITH_CHANGELOG)
        symbols = self.prep_debian_file('foo.symbols', s.BASIC_SYMBOLS)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        result_symbols = os_path_join_safe(
            self.result_dir, 'simple_update.symbols')
        result_changelog = os_path_join_safe(
            self.result_dir,
            'simple_foo_update_with_existing_content.changelog')
        self.assertFilesAreIdenticals(symbols, result_symbols)
        self.assertChangelogFilesAreIdenticals(result_changelog, changelog)

    def test_refresh_symbols_files_with_no_symbol_files(self):
        """Update symbols file with the new version having no symbol file."""
        changelog = self.prep_debian_file('changelog', s.NESTED_CHANGELOG)
        orig = self.prep_debian_file('changelog', s.NESTED_CHANGELOG)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        self.assertChangelogFilesAreIdenticals(orig, changelog)

    def test_refresh_symbols_files_with_no_symbol_files_to_update(self):
        """No update for symbols files having nothing to be updated for."""
        changelog = self.prep_debian_file('changelog', s.BASE_CHANGELOG)
        symbols = self.prep_debian_file('foo.symbols', s.NO_SYMBOLS)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        result_changelog = self.prep_debian_file('orig_log', s.BASE_CHANGELOG)
        result_symbols = self.prep_debian_file('orig_sym', s.NO_SYMBOLS)
        self.assertFilesAreIdenticals(symbols, result_symbols)
        self.assertChangelogFilesAreIdenticals(result_changelog, changelog)

    def test_refresh_symbols_files_with_multiple_symbol_files(self):
        """Update all the symbols file with the new version."""
        changelog = self.prep_debian_file('changelog', s.BASE_CHANGELOG)
        foo = self.prep_debian_file('foo.symbols', s.MULTIPLE_SYMBOLS)
        bar = self.prep_debian_file('bar.symbols', s.MULTIPLE_SYMBOLS)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        result_symbols = os_path_join_safe(
            self.result_dir, 'multiplesymbols_update.symbols')
        result_changelog = os_path_join_safe(
            self.result_dir, 'multiple_update.changelog')
        self.assertFilesAreIdenticals(foo, result_symbols)
        self.assertFilesAreIdenticals(bar, result_symbols)
        self.assertChangelogFilesAreIdenticals(result_changelog, changelog)

    def test_refresh_symbols_files_alone_symbol(self):
        """Update a symbols (not <package>.symbols) file."""
        changelog = self.prep_debian_file('changelog', s.BASE_CHANGELOG)
        symbols = self.prep_debian_file('symbols', s.BASIC_SYMBOLS)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        result_symbols = os_path_join_safe(
            self.result_dir, 'simple_update.symbols')
        result_changelog = os_path_join_safe(
            self.result_dir, 'simple_update.changelog')
        self.assertFilesAreIdenticals(symbols, result_symbols)
        self.assertChangelogFilesAreIdenticals(result_changelog, changelog)

    def test_refresh_symbols_files_with_arch_file(self):
        """Update the symbols file with the new version."""
        changelog = self.prep_debian_file('changelog', s.BASE_CHANGELOG)
        symbols = self.prep_debian_file('foo.symbols.amd64', s.BASIC_SYMBOLS)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        result_symbols = os_path_join_safe(
            self.result_dir, 'simple_update.symbols')
        result_changelog = os_path_join_safe(
            self.result_dir, 'simple_update_arch.changelog')
        self.assertChangelogFilesAreIdenticals(result_changelog, changelog)
        self.assertFilesAreIdenticals(symbols, result_symbols)

    def test_dont_refresh_symbols_files_for_directory(self):
        """We don't fail for directories under debian/."""
        self.prep_debian_file('changelog', s.DUMMY_CHANGELOG)
        debian_source_path = os_path_join_safe(
            self.tempdir, 'debian', 'source')
        os.mkdir(debian_source_path)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        self.assertTrue(os.path.isdir(debian_source_path))

    def test_dont_refresh_symbols_files_for_symlink(self):
        """We don't touch symlinks under debian/."""
        self.prep_debian_file('changelog', s.BASE_CHANGELOG)
        self.prep_debian_file('foo.symbols', s.BASIC_SYMBOLS)
        symlink_path = os_path_join_safe(
            self.tempdir, 'debian', 'foo.symbols.link')
        os.symlink('foo.symbols', symlink_path)
        packagemanager.refresh_symbol_files(
            self.tempdir, '42.0daily83.09.14-0ubuntu1')
        self.assertTrue(os.path.islink(symlink_path))
        self.assertEqual('foo.symbols', os.readlink(symlink_path))


class PackageManagerOnlineTests(BaseUnitTestCase):
    # FIXME: This shit should be mocked out regardless. There's no reason
    # a test should fail just because the network happens to be down.
    """Test that uses online services, but read-only."""

    @classmethod
    def setUpClass(cls):
        super(PackageManagerOnlineTests, cls).setUpClass()
        cls.original_settings = packagemanager.settings

    def setup_settings_mock(self, settings_mock):
        """Setup the settings mock for the build source package method.

        Ensure that GNUPG_DIR is not a parent dir of the branch directory.
        """
        settings_mock.ROOT_CU2D = self.original_settings.ROOT_CU2D
        # create another temp dir not parent of the branch directory to ensure
        # we don't mix bindmount and fix magically potential code issues ;)
        settings_mock.GNUPG_DIR = self.create_temp_workdir()

    @patch('cupstream2distro.packagemanager.os.getuid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.getgid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.environ.copy')
    @patch('cupstream2distro.packagemanager.Popen')
    def test_build_source_package(self, popenMock, copyMock):
        """Call cowbuilder and build a source package."""
        home = os.path.expanduser('~')
        bsc = os_path_join_safe(
            self.root_dir, 'chroot-tools', 'buildsource-chroot')
        copyMock.return_value = {}
        popenMock.return_value.communicate.return_value = ('', '')
        popenMock.return_value.returncode = 0
        packagemanager.SILO_DIR = lambda: self.tempdir
        packagemanager.build_source_package(
            self.tempdir, 'raring', '1.1-0ubuntu1')
        copyMock.assert_called_once_with()
        self.assertEqual(
            copyMock(), dict(DIST='raring', HOME=os.path.dirname(bsc)))
        expected_calls = [
            call(['./create-chroot'],
                 env=copyMock(),
                 cwd=os_path_join_safe(self.root_dir, 'chroot-tools'),
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
            call(['sudo', '-E', 'cowbuilder', '--execute', '--bindmounts',
                  self.tempdir, '--bindmounts', home, '--', bsc,
                  self.tempdir, '--gnupg-parentdir', home, '--uid', '1000',
                  '--gid', '1000', '--distro-version', '1.1-0ubuntu1'],
                 env=copyMock(),
                 cwd=self.tempdir,
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
        ]
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(popenMock.mock_calls[i], kall)

    @patch('cupstream2distro.packagemanager.os.getuid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.getgid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.environ.copy')
    @patch('cupstream2distro.packagemanager.Popen')
    def test_build_source_package_with_ppa(self, popenMock, copyMock):
        """Build a source package (unsigned for tests) for an older distro."""
        home = os.path.expanduser('~')
        bsc = os_path_join_safe(
            self.root_dir, 'chroot-tools', 'buildsource-chroot')
        copyMock.return_value = {}
        popenMock.return_value.communicate.return_value = ('', '')
        popenMock.return_value.returncode = 0
        packagemanager.SILO_DIR = lambda: self.tempdir
        packagemanager.build_source_package(
            self.tempdir, 'raring', '1.1-0ubuntu1', 'ubuntu-unity/next')
        copyMock.assert_called_once_with()
        self.assertEqual(
            copyMock(), dict(DIST='raring', HOME=os.path.dirname(bsc)))
        expected_calls = [
            call(['./create-chroot'],
                 env=copyMock(),
                 cwd=os_path_join_safe(self.root_dir, 'chroot-tools'),
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
            call(['sudo', '-E', 'cowbuilder', '--execute', '--bindmounts',
                  self.tempdir, '--bindmounts', home, '--', bsc,
                  self.tempdir, '--gnupg-parentdir', home, '--uid', '1000',
                  '--gid', '1000', '--distro-version', '1.1-0ubuntu1', '--ppa',
                  'ubuntu-unity/next'],
                 env=copyMock(),
                 cwd=self.tempdir,
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
        ]
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(popenMock.mock_calls[i], kall)

    @patch('cupstream2distro.packagemanager.os.getuid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.getgid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.environ.copy')
    @patch('cupstream2distro.packagemanager.Popen')
    def test_build_with_other_distro_version(self, popenMock, copyMock):
        """Build a source package (unsigned for tests) for an older distro."""
        home = os.path.expanduser('~')
        bsc = os_path_join_safe(
            self.root_dir, 'chroot-tools', 'buildsource-chroot')
        copyMock.return_value = {}
        popenMock.return_value.communicate.return_value = ('', '')
        popenMock.return_value.returncode = 0
        packagemanager.SILO_DIR = lambda: self.tempdir
        packagemanager.build_source_package(
            self.tempdir, 'precise', '1.1-0ubuntu1')
        copyMock.assert_called_once_with()
        self.assertEqual(
            copyMock(), dict(DIST='precise', HOME=os.path.dirname(bsc)))
        expected_calls = [
            call(['./create-chroot'],
                 env=copyMock(),
                 cwd=os_path_join_safe(self.root_dir, 'chroot-tools'),
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
            call(['sudo', '-E', 'cowbuilder', '--execute', '--bindmounts',
                  self.tempdir, '--bindmounts', home, '--', bsc,
                  self.tempdir, '--gnupg-parentdir', home, '--uid', '1000',
                  '--gid', '1000', '--distro-version', '1.1-0ubuntu1'],
                 env=copyMock(),
                 cwd=self.tempdir,
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
        ]
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(popenMock.mock_calls[i], kall)

    @patch('cupstream2distro.packagemanager.os.getuid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.getgid', lambda: '1000')
    @patch('cupstream2distro.packagemanager.os.environ.copy')
    @patch('cupstream2distro.packagemanager.Popen')
    def test_build_source_package_raises_exception(self, popenMock, copyMock):
        """Raise exception if subprocess returns error."""
        home = os.path.expanduser('~')
        bsc = os_path_join_safe(
            self.root_dir, 'chroot-tools', 'buildsource-chroot')
        copyMock.return_value = {}
        popenMock.return_value.communicate.return_value = ('', '')
        popenMock.return_value.returncode = 30
        packagemanager.SILO_DIR = lambda: self.tempdir
        with self.assertRaisesRegexp(Exception, 'returned: 30'):
            packagemanager.build_source_package(
                self.tempdir, 'raring', '1.1-0ubuntu1')
        copyMock.assert_called_once_with()
        self.assertEqual(
            copyMock(), dict(DIST='raring', HOME=os.path.dirname(bsc)))
        expected_calls = [
            call(['./create-chroot'],
                 env=copyMock(),
                 cwd=os_path_join_safe(self.root_dir, 'chroot-tools'),
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
            call(['sudo', '-E', 'cowbuilder', '--execute', '--bindmounts',
                  self.tempdir, '--bindmounts', home, '--', bsc,
                  self.tempdir, '--gnupg-parentdir', home, '--uid', '1000',
                  '--gid', '1000', '--distro-version', '1.1-0ubuntu1'],
                 env=copyMock(),
                 cwd=self.tempdir,
                 stderr=-1,
                 stdout=-1),
            call().communicate(),
        ]
        self.assertEqual(len(expected_calls), len(popenMock.mock_calls))
        for i, kall in enumerate(expected_calls):
            self.assertEqual(popenMock.mock_calls[i], kall)

    def get_source_files_for_package(self, package_name):
        """Small tool to get a source package files in package_name dir."""
        result_files_path = []
        package_source_dir = self.get_ubuntu_source_content_path(package_name)
        for filename in os.listdir(package_source_dir):
            result_files_path.append(
                os_path_join_safe(package_source_dir, filename))
        return result_files_path

    def test_raise_exception_if_diff_unexisting_dsc(self):
        """We raise an exception if we try to debdiff an unexisting dsc."""
        with self.assertRaises(Exception):
            packagemanager._packaging_changes_between_dsc('foo.dsc', 'bar.dsc')

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_no_source(self, lpMock):
        """Assert when we can't find the correct source in dest we download."""
        dest_archive = Mock()
        dest_archive.getPublishedSources.return_value = []
        with self.assertRaises(Exception):
            packagemanager.get_source_package_from_dest(
                'foo', dest_archive, '42.0daily83.09.13.2', 'rolling')

    @patch('cupstream2distro.packagemanager.lp')
    def test_get_source_package_from_dest_wrong_content_source(
            self, lpMock):
        """Didn't download properly expected source (no .dsc file)."""

        def get_source_files_for_package_but_dsc(package_name):
            """Small tool to get a source package files in package_name dir."""
            result_files_path = []
            package_source_dir = self.get_ubuntu_source_content_path(
                package_name)
            for filename in os.listdir(package_source_dir):
                if not filename.endswith('.dsc'):
                    result_files_path.append(
                        os_path_join_safe(package_source_dir, filename))
            return result_files_path

        dest_archive = Mock()
        source1 = Mock()
        dest_archive.getPublishedSources.return_value = [source1]
        source1.sourceFileUrls.return_value = (
            get_source_files_for_package_but_dsc('foo_package'))

        with self.assertRaises(Exception):
            packagemanager.get_source_package_from_dest(
                'foo', dest_archive, '42.0daily83.09.13.2-0ubuntu1', 'rolling')

        dest_archive.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=lpMock.get_series.return_value,
            version='42.0daily83.09.13.2-0ubuntu1')
        source1.sourceFileUrls.assert_called_once()

    def test_create_new_packaging_version_with_wrong_daily_format(self):
        """Raise exception if previous version is in wrong format."""
        with self.assertRaises(Exception):
            packagemanager.create_new_packaging_version(
                '42daily83.0garbage9.12-1ubuntu2')

    def test_update_changelog_simple_lower(self):
        """Raise an exception if we update a changelog with a lower version."""
        self.prep_debian_file('changelog', s.DUMMY_CHANGELOG)
        packagemanager.SILO_DIR = lambda: self.tempdir
        with self.assertRaisesRegexp(Exception, 'is less than\nthe current'):
            packagemanager.update_changelog(
                self.tempdir, '1.1daily83.09.13-0ubuntu1', 'rolling', 42, {})
