# -*- coding: utf-8 -*-
# Copyright: (C) 2014 Canonical
#
# Authors:
#  Robert Bruce Park
#
# This program is free software; you can redistribute it and/or modify it under
# the terms of the GNU General Public License as published by the Free Software
# Foundation; version 3.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA

from tests.unit import DirectoryAwareTestCase

from cupstream2distro import project
from cupstream2distro.project import DotProject
from cupstream2distro.utils import env, os_path_join_safe

import os
import shutil


class ProjectTests(DirectoryAwareTestCase):
    def setUp(self):
        """Mock out SILO_DIR and SILOS_DIR for testing."""
        super(ProjectTests, self).setUp()
        project.SILO_DIR = lambda: self.tempdir
        env.SILONAME = ''

    def test_section(self):
        """Do we correctly identify which sections each key belong to."""
        self.assertEqual(project.section('rev'), 'Branch')
        self.assertEqual(project.section('branch'), 'Branch')
        self.assertEqual(project.section('dest_current_version'), 'Package')
        self.assertEqual(project.section('packaging_version'), 'Package')
        self.assertEqual(project.section('unknown'), 'Misc')
        self.assertEqual(project.section('published_version'), 'Misc')

    def test_file_creation(self):
        """Do we create an empty file when saving an empty project?"""
        proj = DotProject('hello')
        proj.write()
        self.assertTrue(os.path.isfile(
            os_path_join_safe(self.tempdir, 'hello.project')))
        with open(os_path_join_safe(self.tempdir, 'hello.project')) as f:
            self.assertEqual(f.readlines(), [])

    def test_file_writing(self):
        """Does the written file have the expected values in it?"""
        proj = DotProject('yeehaw')
        proj.rev = '5'
        proj.branch = 'lp:example'
        proj.dest_current_version = '0.99.alpha+15.04.20141225-0ubuntu1'
        proj.packaging_version = '1.0+15.04.20150101-0ubuntu1'
        proj.write()
        with open(os_path_join_safe(self.tempdir, 'yeehaw.project')) as f:
            self.assertEqual(
                f.readlines(),
                ['[Branch]\n',
                 'rev = 5\n',
                 'branch = lp:example\n',
                 '\n',
                 '[Package]\n',
                 'dest_current_version = 0.99.alpha+15.04.20141225-0ubuntu1\n',
                 'packaging_version = 1.0+15.04.20150101-0ubuntu1\n',
                 '\n',
                 ])

    def test_soft_save_initial(self):
        """Does the soft_save method create the expected values in the file?"""
        DotProject.soft_save(
            'yessir', branch='lp:foo', rev='5', dest='0.1', ours='0.2')
        with open(os_path_join_safe(self.tempdir, 'yessir.project')) as f:
            self.assertEqual(
                f.readlines(),
                ['[Branch]\n',
                 'branch = lp:foo\n',
                 'rev = 5\n',
                 '\n',
                 '[Package]\n',
                 'dest_current_version = 0.1\n',
                 'packaging_version = 0.2\n',
                 '\n',
                 ])

    def test_soft_save_initial_with_blanks(self):
        """Does soft_save create the expected blank values in the file?"""
        DotProject.soft_save('yessir', dest='0.1', ours='0.2')
        with open(os_path_join_safe(self.tempdir, 'yessir.project')) as f:
            self.assertEqual(
                f.readlines(),
                ['[Branch]\n',
                 'branch = \n',
                 'rev = \n',
                 '\n',
                 '[Package]\n',
                 'dest_current_version = 0.1\n',
                 'packaging_version = 0.2\n',
                 '\n',
                 ])

    def test_soft_save_twice(self):
        """Does a second call to soft_save preserve original values?"""
        DotProject.soft_save(
            'yessir', branch='lp:foo', rev='5', dest='0.1', ours='0.2')
        DotProject.soft_save('yessir', dest='0.2', ours='0.3')
        with open(os_path_join_safe(self.tempdir, 'yessir.project')) as f:
            self.assertEqual(
                f.readlines(),
                ['[Branch]\n',
                 'branch = lp:foo\n',
                 'rev = 5\n',
                 '\n',
                 '[Package]\n',
                 'dest_current_version = 0.2\n',
                 'packaging_version = 0.3\n',
                 '\n',
                 ])

    def test_soft_save_twice_again(self):
        """Does a second call to soft_save preserve those other values?"""
        DotProject.soft_save(
            'yessir', branch='lp:foo', rev='5', dest='0.1', ours='0.2')
        DotProject.soft_save('yessir', branch='lp:crap', rev='9000')
        with open(os_path_join_safe(self.tempdir, 'yessir.project')) as f:
            self.assertEqual(
                f.readlines(),
                ['[Branch]\n',
                 'branch = lp:crap\n',
                 'rev = 9000\n',
                 '\n',
                 '[Package]\n',
                 'dest_current_version = 0.1\n',
                 'packaging_version = 0.2\n',
                 '\n',
                 ])

    def test_file_reading(self):
        """If we create a file manually, can DotProject read those values?"""
        with open(os_path_join_safe(self.tempdir, 'fuddy.project'), 'w') as f:
            f.write(
                '[Branch]\n'
                'rev = 27\n'
                'branch = lp:robru-desktop\n\n'
                '[Package]\n'
                'dest_current_version = 0.99\n'
                'packaging_version = 1.0\n\n'
                '[Misc]\n'
                'wut = yo\n\n')
        proj = DotProject('fuddy')
        self.assertEqual(proj.rev, '27')
        self.assertEqual(proj.branch, 'lp:robru-desktop')
        self.assertEqual(proj.dest_current_version, '0.99')
        self.assertEqual(proj.packaging_version, '1.0')
        self.assertEqual(proj.wut, 'yo')

    def test_get_previous_distro_version_from_config(self):
        """Can we access dest_current_version?"""
        shutil.copy2(
            os_path_join_safe(self.project_file_dir, 'foo.project'),
            self.tempdir)
        self.assertEqual(
            DotProject('foo').dest_current_version,
            '6.12.0-0ubuntu1')

    def test_get_all_packages_uploaded(self):
        """Does get_all_packages_uploaded find one file?"""
        shutil.copy2(
            os_path_join_safe(self.project_file_dir, 'foo.project'),
            self.tempdir)
        self.assertEqual(
            DotProject.get_all_packages_uploaded(),
            set([('foo', '6.12.0daily13.02.27-0ubuntu1', '42', 'lp:foo')]))

    def test_get_all_packages_uploaded_multi(self):
        """Does get_all_packages_uploaded find three files?"""
        shutil.copy2(
            os_path_join_safe(self.project_file_dir, 'foo.project'),
            self.tempdir)
        with open(os_path_join_safe(self.tempdir, 'f.project'), 'w') as f:
            f.write(
                '[Branch]\n'
                'rev = 27\n'
                'branch = lp:robru-desktop\n\n'
                '[Package]\n'
                'dest_current_version = 0.99\n'
                'packaging_version = 1.0\n\n'
                '[Misc]\n'
                'published_version = 2.0\n'
                'wut = yo\n\n')
        proj = DotProject('bar')
        proj.rev = '300'
        proj.packaging_version = 'OVER NINE THOUSAND'
        proj.branch = 'lp:bar'
        proj.write()
        self.assertEqual(
            DotProject.get_all_packages_uploaded(),
            set([
                ('f', '1.0', '27', 'lp:robru-desktop'),
                ('foo', '6.12.0daily13.02.27-0ubuntu1', '42', 'lp:foo'),
                ('bar', 'OVER NINE THOUSAND', '300', 'lp:bar'),
                ]))

    def test_mark_project_as_published(self):
        """Confirm project file creation to avoid re-publishing projects."""
        shutil.copy2(os_path_join_safe(self.project_file_dir, 'foo.project'),
                     self.tempdir)
        DotProject('foo').mark_as_published('4.2dailysomething-0ubuntu1')
        filename = os_path_join_safe(self.tempdir, 'foo.project')
        self.assertTrue(os.path.isfile(filename))
        with open(filename) as f:
            self.assertEqual(
                f.readlines(), [
                    '[Branch]\n',
                    'branch = lp:foo\n',
                    'rev = 42\n',
                    '\n',
                    '[Package]\n',
                    'dest_current_version = 6.12.0-0ubuntu1\n',
                    'packaging_version = 6.12.0daily13.02.27-0ubuntu1\n',
                    '\n',
                    '[Misc]\n',
                    'published_version = 4.2dailysomething-0ubuntu1\n',
                    '\n',
                ])

    def test_mark_project_and_diff_as_published(self):
        """Confirm diff file creation to prevent project republishing."""
        shutil.copy2(os_path_join_safe(self.project_file_dir, 'foo.project'),
                     self.tempdir)
        shutil.copy2(os_path_join_safe(self.project_file_dir, 'foo.project'),
                     os_path_join_safe(
                         self.tempdir, 'packaging_changes_foo_4.2'
                         'dailysomething-0ubuntu1.diff'))
        DotProject('foo').mark_as_published('4.2dailysomething-0ubuntu1')
        self.assertTrue(os.path.isfile(
            os_path_join_safe(
                self.tempdir,
                'packaging_changes_foo_4.2dailysomething'
                '-0ubuntu1.diff.published')))

    def test_one_project_published(self):
        """Get one project published with one version."""
        shutil.copy2(os_path_join_safe(self.project_file_dir, 'foo.project'),
                     self.tempdir)
        shutil.copy2(os_path_join_safe(self.tempdir, 'foo.project'),
                     os_path_join_safe(self.tempdir, 'bar.project'))
        DotProject('foo').mark_as_published('42-0ubuntu1')
        self.assertEquals(
            DotProject.get_all_packages_published(),
            {'foo': ['42-0ubuntu1']})

    def test_one_project_published_multiple_times(self):
        """Get one project published with multiple published versions."""
        shutil.copy2(os_path_join_safe(self.project_file_dir, 'foo.project'),
                     self.tempdir)
        DotProject('foo').mark_as_published('42-0ubuntu1')
        shutil.copy2(os_path_join_safe(self.project_file_dir, 'foo.project'),
                     self.tempdir)
        DotProject('foo').mark_as_published('43-0ubuntu1')
        published = DotProject.get_all_packages_published()
        self.assertEqual(list(published), ['foo'])
        self.assertEqual(sorted(published['foo']), ['43-0ubuntu1'])

    def test_multiple_projects_published(self):
        """Get one project published with multiple published versions."""
        shutil.copy2(os_path_join_safe(self.project_file_dir, 'foo.project'),
                     self.tempdir)
        DotProject('foo').mark_as_published('42-0ubuntu1')
        shutil.copy2(
            os_path_join_safe(self.project_file_dir, 'foo.project'),
            os_path_join_safe(self.tempdir, 'bar.project'))
        DotProject('bar').mark_as_published('43-0ubuntu1')
        self.assertDictEqual(
            DotProject.get_all_packages_published(),
            {'foo': ['42-0ubuntu1'], 'bar': ['43-0ubuntu1']})

    def test_packaging_diff_filename(self):
        """Return the right packaging diff name."""
        self.assertEquals(
            DotProject('foo').get_diff_path('1.0daily~-0ubuntu1'),
            os_path_join_safe(
                self.tempdir, 'packaging_changes_foo_1.0daily~-0ubuntu1.diff'))
        proj = DotProject('bar')
        proj.mark_as_published('2.0')
        self.assertEquals(
            DotProject('bar').get_diff_path(),
            os_path_join_safe(self.tempdir, 'packaging_changes_bar_2.0.diff'))
