from tests.unit import CITrainScriptTestCase

from mock import call, patch, Mock, MagicMock

# Unused import necessary for code coverage reporting
from citrain import build
build


class source_archive:
    class distribution:
        name = 'ubuntu'


class BuildTestCase(CITrainScriptTestCase):
    scriptname = 'build.py'

    def setUp(self):
        super(BuildTestCase, self).setUp()

        # Setup the test environment
        self.test_archive = Mock()
        self.test_archive.copyPackages.return_value = True

        self.test_source_archive = source_archive
        self.test_source_series = None

        self.test_series = Mock()
        self.test_series.name = '14.09'
        self.test_target_archive_name = 'ubuntu-rtm'

    def test_correct_changelog(self):
        """Successfully mangle the series/version in the changelog."""
        self.script.open = Mock()
        self.script.open.return_value = MagicMock(spec=file)
        self.script.open().__enter__().readlines.return_value = [
            'foo-bar (0.1) utopic; urgency=low\n',
            '\n',
            '  * stuff\n',
            '\n',
            ' -- Jim Bo <jim@example.com>  Tue, 29 Apr 2014 16:37:24 +0200\n',
            '\n',
            'only-first-line-mangled (0.01) utopic; urgency=low\n',
        ]
        self.script.correct_changelog('/branch', 'v1.0', 'warty')
        calls = self.script.open.mock_calls
        self.assertIn(call('/branch/debian/changelog'), calls)
        self.assertIn(call('/branch/debian/changelog', 'w'), calls)
        self.script.open().__enter__().writelines.assert_called_once_with([
            'foo-bar (v1.0) warty; urgency=low\n',
            '\n',
            '  * stuff\n',
            '\n',
            ' -- Jim Bo <jim@example.com>  Tue, 29 Apr 2014 16:37:24 +0200\n',
            '\n',
            'only-first-line-mangled (0.01) utopic; urgency=low\n',
        ])

    def test_also_build_for(self):
        self.script.shutil = Mock()
        self.script.correct_changelog = Mock()
        self.script.SILO_DIR = lambda: '/silo/dir'
        series_dir = self.script.also_build_for(
            self.script.SILO_DIR() + '/branch', 'warty', 'v0.99')
        self.assertEqual(series_dir, '/silo/dir/warty/branch')
        self.script.shutil.rmtree.assert_called_once_with(
            series_dir, ignore_errors=True)
        self.script.shutil.copytree.assert_called_once_with(
            '/silo/dir/branch', series_dir)
        self.script.correct_changelog.assert_called_once_with(
            series_dir, 'v0.99~warty', 'warty')

    def test_append_rtm_to_version_native(self):
        """Append the ~rtm tag to the packaging version for native pkgs."""
        new_version = self.script.append_tag_to_version('1.2', 'rtm')
        self.assertEqual(new_version, '1.2~rtm')

    def test_append_rtm_to_version_citrain(self):
        """Append the ~rtm tag to the packaging version for ubuntu pkgs."""
        new_version = self.script.append_tag_to_version(
            '1.2+14.10.20140923-0ubuntu1', 'rtm')
        self.assertEqual(new_version, '1.2+14.10.20140923~rtm-0ubuntu1')

    def test_append_tag_to_version_multi(self):
        """Append a tag to the version string with multiple hyphens."""
        new_version = self.script.append_tag_to_version(
            '1-2-3-0ubuntu1', 'hello')
        self.assertEqual(new_version, '1-2-3~hello-0ubuntu1')

    @patch('cupstream2distro.packagemanager.get_source_package_from_dest')
    def test_binary_sync(self, getFromDest):
        """Try performing a binary sync for a dummy package."""
        self.script.packagemanager\
            .is_upstream_version_for_series_in_dest = Mock(return_value=False)
        source = Mock()
        source.source_package_version = '1.1'
        source.status = 'Published'

        self.test_archive.getPublishedSources.return_value = [source]

        getFromDest.return_value = '/foo/bar/foo-1.1'

        (dest, version) = self.script.prepare_binary_sync(
            'foo',
            '1.2',
            self.test_source_archive,
            self.test_source_series,
            self.test_archive,
            self.test_series,
            self.test_target_archive_name,
            '1.1')

        self.assertEqual(version, '1.2')
        self.test_archive.copyPackage.assert_called_once_with(
            from_archive=self.test_source_archive,
            from_pocket="Release",
            from_series=self.test_source_series,
            include_binaries=True,
            to_pocket="Release",
            to_series=self.test_series.name,
            source_name='foo',
            version='1.2')

    def test_no_source_sync_if_package_up_to_date(self):
        """Make sure no source sync is made when archive is up-to-date."""
        source = Mock()
        source.source_package_version = '1.2'
        source.status = 'Published'
        self.test_archive.getPublishedSources.return_value = [source]

        self.script.packagemanager = Mock()
        (dest, version) = self.script.prepare_source_sync(
            'foo',
            '1.2',
            self.test_source_archive,
            self.test_source_series,
            self.test_archive,
            self.test_series,
            self.test_target_archive_name,
            '1.1')

        self.assertIsNone(version)

    def test_no_binary_sync_if_package_up_to_date(self):
        """Make sure no binary sync is made when archive is up-to-date."""
        source = Mock()
        source.source_package_version = '1.2'
        source.status = 'Published'
        self.test_archive.getPublishedSources.return_value = [source]

        (dest, version) = self.script.prepare_binary_sync(
            'foo',
            '1.2',
            self.test_source_archive,
            self.test_source_series,
            self.test_archive,
            self.test_series,
            self.test_target_archive_name,
            '1.1')

        self.assertIsNone(version)
