from collections import defaultdict
from mock import Mock, call

from tests.unit.test_script_watch_ppa import SourceFake
from tests.unit import CITrainScriptTestCase
from tests import strings as s

from cupstream2distro.utils import env, os_path_join_safe

from citrain.publisher import PublishError


class MergeFake:
    def __init__(self, status):
        self.queue_status = status
        self.web_link = 'example.com'


class PublisherTestCase(CITrainScriptTestCase):
    scriptname = 'publisher.py'

    def setUp(self):
        super(PublisherTestCase, self).setUp()
        env.SILONAME = 'ubuntu/landing-789'
        env.DISTRO = 'ubuntu'
        env.SERIES = 'vivid'

    def test_generate_xml_artefacts_no_issue(self):
        """Generate the xml jenkins artefacts when no issue occured."""
        self.script.generate_xml_artifacts([])
        with open(os_path_join_safe(self.tempdir, 'publisher.xml')) as f:
            self.assertEqual(f.read().split('\n'), s.XML_NO_ERROR.split('\n'))

    def test_generate_xml_artefacts_one_failure(self):
        """Generate the xml jenkins artefacts when there is one failure."""
        self.script.generate_xml_artifacts(['one issue'])
        with open(os_path_join_safe(self.tempdir, 'publisher.xml')) as f:
            self.assertEqual(f.read().split('\n'), s.XML_1_ERROR.split('\n'))

    def test_generate_xml_artefacts_two_failures(self):
        """Generate the xml artefacts when there is more than one failure."""
        self.script.generate_xml_artifacts(['one issue', 'a second issue'])
        with open(os_path_join_safe(self.tempdir, 'publisher.xml')) as f:
            self.assertEqual(f.read().split('\n'), s.XML_2_ERRORS.split('\n'))

    def test_get_distro_and_series(self):
        """Given a series URL, identify the distro and series."""
        env.DISTRO, env.SERIES = ('', '')
        self.script.get_distro_and_series(
            'https://api.launchpad.net/devel/ubuntu/vivid')
        self.assertEqual(env.DISTRO, 'ubuntu')
        self.assertEqual(env.SERIES, 'vivid')
        self.script.get_distro_and_series(
            'https://api.launchpad.net/devel/ubuntu-rtm/14.09')
        self.assertEqual(env.DISTRO, 'ubuntu-rtm')
        self.assertEqual(env.SERIES, '14.09')

    def test_packagelist(self):
        """Make sure the format for each row of the packagelist is correct."""
        self.assertEqual(
            self.script.PackageList.TEMPLATE,
            '{src_ppa}\t{src_pocket}\t{src_series}\t{target_pocket}\t'
            '{target_series}\t{source_name}\t{version}\t{version_in_dest}\t'
            '{publisher_nick}\t{target_distribution}\n')

    def test_packagelist_init(self):
        """PackageList should start with the correct values."""
        self.script.get_distro_and_series = Mock()
        self.script.get_distro_and_series.return_value = ('ubuntu', 'vivid')
        packagelist = self.script.PackageList()
        self.assertEqual(packagelist.silo, 'landing-789')

    def test_packagelist_publish_source(self):
        """We can add a row to the packagelist lines array."""
        self.script.DotProject = Mock()
        self.script.DotProject.return_value.dest_current_version = '0.9'
        env.PUBLISHER = 'robru'
        packagelist = self.script.PackageList()
        source = SourceFake(
            'foo', '1.0-0ubuntu1',
            'https://api.launchpad.net/devel/ubuntu/vivid')
        packagelist.publish_source(source)
        self.assertEqual(
            packagelist.lines,
            ['ci-train-ppa-service/ubuntu/landing-789\tRelease\tvivid'
             '\tProposed\tvivid\tfoo\t1.0-0ubuntu1\t0.9\trobru\tubuntu\n'])
        source = SourceFake(
            'bar', '2.0-0ubuntu1',
            'https://api.launchpad.net/devel/ubuntu/vivid')
        packagelist.publish_source(source)
        self.assertEqual(
            packagelist.lines,
            ['ci-train-ppa-service/ubuntu/landing-789\tRelease\tvivid'
             '\tProposed\tvivid\tfoo\t1.0-0ubuntu1\t0.9\trobru\tubuntu\n',
             'ci-train-ppa-service/ubuntu/landing-789\tRelease\tvivid'
             '\tProposed\tvivid\tbar\t2.0-0ubuntu1\t0.9\trobru\tubuntu\n'])

    def test_packagelist_write(self):
        """Ensure we can correctly write the packagelist to disk."""
        packagelist = self.script.PackageList()
        lines = packagelist.lines = [
            'ci-train-ppa-service/ubuntu/landing-789\tRelease\tvivid'
            '\tProposed\tvivid\tfoo\t1.0-0ubuntu1\t0.9\trobru\tubuntu\n',
            'ci-train-ppa-service/ubuntu/landing-789\tRelease\tvivid'
            '\tProposed\tvivid\tbar\t2.0-0ubuntu1\t0.9\trobru\tubuntu\n',
        ]
        packagelist.write()
        self.script.logging.debug.assert_called_once_with(
            '\n' + ''.join(lines))
        result_path = os_path_join_safe(
            self.tempdir, 'packagelist_rsync_landing-789-vivid')
        with open(result_path) as f:
            self.assertEqual(f.read(), ''.join(lines))

    def test_push_branches_success(self):
        """Ensure that push_branches() can succeed."""
        env.DISTRO = 'ubuntu'
        env.SERIES = 'vivid'
        mps = {
            'foo': ['example.com/a', 'example.com/b'],
            'bar': ['example.com/c', 'example.com/d'],
        }
        lp_load = Mock()
        lp_load.return_value.target_branch.project.name = 'fooproject'
        self.script.branchhandling.push_to_branch.return_value = True
        self.script.push_branches(mps, lp_load)
        self.assertEqual(
            lp_load.mock_calls,
            [call('example.com/c'), call('example.com/a')])
        self.assertEqual(
            self.script.branchhandling.push_to_branch.mock_calls,
            [call(os_path_join_safe(self.tempdir, 'bar'),
                  'lp:~/fooproject/ubuntu-vivid-proposed',
                  overwrite=True),
             call(os_path_join_safe(self.tempdir, 'foo'),
                  'lp:~/fooproject/ubuntu-vivid-proposed',
                  overwrite=True)])
        self.assertEqual(self.script.logging.error.call_count, 0)

    def test_push_branches_failure(self):
        """Ensure that push_branches() handles failure."""
        env.DISTRO = 'ubuntu-rtm'
        env.SERIES = '14.09'
        mps = {
            'foo': ['example.com/a', 'example.com/b'],
            'bar': ['example.com/c', 'example.com/d'],
        }
        lp_load = Mock()
        lp_load.return_value.target_branch.project.name = 'fooproject'
        self.script.branchhandling.push_to_branch.return_value = False
        with self.assertRaisesRegexp(PublishError, 'Failed push'):
            self.script.push_branches(mps, lp_load)
        self.assertEqual(lp_load.mock_calls, [call('example.com/c')])
        self.assertEqual(
            self.script.branchhandling.push_to_branch.mock_calls,
            [call(os_path_join_safe(self.tempdir, 'bar'),
                  'lp:~/fooproject/ubuntu-rtm-14.09-proposed',
                  overwrite=True)])
        self.script.branchhandling.push_to_branch.assert_called_once_with(
            os_path_join_safe(self.tempdir, 'bar'),
            'lp:~/fooproject/ubuntu-rtm-14.09-proposed',
            overwrite=True)
        self.assertEqual(self.script.logging.error.call_count, 0)

    def test_initiate_publication_rsync(self):
        """Ensure that we determine when to rsync."""
        sources = set([
            SourceFake('foo', '1.0-0ubuntu1'),
            SourceFake('bar', '2.0-0ubuntu1'),
        ])
        dest, ppa, set_versions = Mock(), Mock(), Mock()
        self.script.lp.is_distro_archive = lambda x: True
        self.script.write_rsync_file = Mock()
        self.script.copy_to_dest = Mock()
        self.script.find_ppa_sources = Mock()
        self.script.find_ppa_sources.side_effect = \
            lambda x, y, z: y.update(sources)
        self.script.initiate_publication(dest, ppa, set_versions)
        self.script.find_ppa_sources.assert_called_once_with(
            ppa, sources, defaultdict(set))
        self.script.write_rsync_file.assert_called_once_with(sources, dest)
        self.assertEqual(self.script.copy_to_dest.call_count, 0)
        set_versions.assert_called_once_with(defaultdict(set))

    def test_initiate_publication_copy(self):
        """Ensure that we determine when to copy."""
        sources = set([
            SourceFake('foo', '1.0-0ubuntu1'),
            SourceFake('bar', '2.0-0ubuntu1'),
        ])
        dest, ppa, set_versions = Mock(), Mock(), Mock()
        self.script.lp.is_distro_archive = lambda x: False
        self.script.write_rsync_file = Mock()
        self.script.copy_to_dest = Mock()
        self.script.find_ppa_sources = Mock()
        self.script.find_ppa_sources.side_effect = \
            lambda x, y, z: y.update(sources)
        self.script.initiate_publication(dest, ppa, set_versions)
        self.script.find_ppa_sources.assert_called_once_with(
            ppa, sources, defaultdict(set))
        self.script.copy_to_dest.assert_called_once_with(sources, dest)
        self.assertEqual(self.script.write_rsync_file.call_count, 0)
        set_versions.assert_called_once_with(defaultdict(set))

    def test_write_rsync_file(self):
        """Ensure that we write the rsync packagelist correctly."""
        dest = Mock()
        sources = [
            SourceFake('foo', '1.0-0ubuntu1'),
            SourceFake('bar', '2.0-0ubuntu1'),
        ]
        self.script.PackageList = Mock()
        self.script.PackageList.return_value.series = 'vivid'
        self.script.write_rsync_file(sources, dest)
        self.script.PackageList.assert_called_once_with()
        publish_source = self.script.PackageList().publish_source
        self.assertEqual(publish_source.mock_calls,
                         [call(sources[0]), call(sources[1])])
        self.assertEqual(
            self.script.DotProject.mock_calls,
            [call('foo'), call().mark_as_published('1.0-0ubuntu1'),
             call('bar'), call().mark_as_published('2.0-0ubuntu1')])
        self.assertEqual(self.script.logging.error.call_count, 0)
        self.script.PackageList().write.assert_called_once_with()

    def test_write_rsync_file_no_dupes(self):
        """Ensure that rsync file does not contain duplicate publishing."""
        dest = Mock()
        sources = [
            SourceFake('foo', '1.0-0ubuntu1'),
            SourceFake('bar', '2.0-0ubuntu1'),
        ]
        self.script.packagemanager.get_current_version_for_series\
            .return_value = '2.0-0ubuntu1'
        self.script.PackageList = Mock()
        self.script.PackageList.return_value.series = 'vivid'
        self.script.write_rsync_file(sources, dest)
        self.script.PackageList.assert_called_once_with()
        publish_source = self.script.PackageList().publish_source
        self.assertEqual(publish_source.mock_calls,
                         [call(sources[0])])
        self.assertEqual(
            self.script.DotProject.mock_calls,
            [call('foo'), call().mark_as_published('1.0-0ubuntu1')])
        self.assertEqual(self.script.logging.error.call_count, 0)
        self.script.PackageList().write.assert_called_once_with()

    def test_write_rsync_file_handles_utopic_and_vivid(self):
        """write_rsync_file() should put only a single series in the list.

        And it should call copy_to_dest with the rest.
        """
        dest = Mock()
        trusty = 'https://api.launchpad.net/devel/ubuntu/trusty'
        utopic = 'https://api.launchpad.net/devel/ubuntu/utopic'
        vivid = 'https://api.launchpad.net/devel/ubuntu/vivid'
        sources = [
            SourceFake('foo', '1.0-0ubuntu1', vivid),
            SourceFake('foo', '1.0~utopic-0ubuntu1', utopic),
            SourceFake('bar', '2.0-0ubuntu1', vivid),
            SourceFake('bar', '2.0~utopic-0ubuntu1', utopic),
            SourceFake('bar', '2.0~trusty-0ubuntu1', trusty),  # Ignored.
        ]
        self.script.copy_to_dest = Mock()
        self.script.PackageList = Mock()
        self.script.write_rsync_file(sources, dest)
        self.script.PackageList.assert_called_once_with()
        publish_source = self.script.PackageList().publish_source
        self.assertEqual(publish_source.call_count, 2)
        for source in sources:
            args = (call(source), publish_source.mock_calls)
            if source.distro_series_link == vivid:
                self.assertIn(*args)
            else:
                self.assertNotIn(*args)
        # TODO: This is part of the utopic+vivid hack, enable if necessary.
        # self.assertEqual(
        #     self.script.copy_to_dest.mock_calls,
        #     [call([sources[1]], dest=self.script.UTOPIC_DEVEL_PPA),
        #      call([sources[3]], dest=self.script.UTOPIC_DEVEL_PPA)])
        self.assertEqual(self.script.logging.error.call_count, 0)
        self.script.PackageList().write.assert_called_once_with()

    def test_prevent_duplicate_publishing_ok(self):
        """Ensure we can publish ok the first time."""
        self.script.prevent_duplicate_publishing(False)
        self.script.os.remove.assert_called_once_with(
            os_path_join_safe(self.tempdir, 'publisher.xml'))

    def test_prevent_duplicate_publishing(self):
        """Ensure main function cowardly refuses to run twice."""
        with self.assertRaisesRegexp(PublishError, 'already run'):
            self.script.prevent_duplicate_publishing(True)

    def test_check_versions_at_destination_override(self):
        """Ensure that users can disable checking versions at dest."""
        env.IGNORE_VERSIONDESTINATION = 'true'
        dest = Mock()
        sources = ('foo', 'bar')
        self.script.check_versions_at_destination(dest, sources)
        self.assertEqual(self.script.lp.load.call_count, 0)
        self.assertEqual(
            self.script.packagemanager.get_current_version_for_series
            .call_count,
            0)
        self.assertEqual(self.script.DotProject.call_count, 0)
        self.script.logging.warning.assert_called_once_with(
            'Honoring your request to ignore dest versions.')

    def test_check_versions_at_destination_all_good(self):
        """Ensure that matching versions at dest don't fail the publish."""
        env.SERIES = 'vivid'
        env.IGNORE_VERSIONDESTINATION = 'false'
        dest = Mock()
        self.script.lp.is_distro_archive = Mock(return_value=True)
        all_projects = ('foo', 'bar', 'grill')
        self.script.DotProject.return_value.dest_current_version = '1.0'
        self.script.packagemanager.get_current_version_for_series\
            .return_value = '1.0'
        self.script.check_versions_at_destination(dest, all_projects)
        self.script.lp.is_distro_archive.assert_called_once_with(dest)
        self.assertEqual(
            self.script.packagemanager.get_current_version_for_series
            .mock_calls, [
                call('foo', 'vivid', dest=dest),
                call('bar', 'vivid', dest=dest),
                call('grill', 'vivid', dest=dest),
            ])
        self.assertEqual(
            self.script.DotProject.mock_calls, [
                call('foo'),
                call('bar'),
                call('grill'),
            ])

    def test_check_versions_at_destination_failure(self):
        """Ensure that version mismatch at dest prevents publication."""
        env.SERIES = 'vivid'
        env.IGNORE_VERSIONDESTINATION = 'false'
        self.script.lp.is_distro_archive = Mock(return_value=True)
        dest = Mock()
        all_projects = ('foo', 'bar', 'grill')
        self.script.DotProject.return_value.dest_current_version = '0.9'
        self.script.packagemanager.get_current_version_for_series\
            .return_value = '1.0'
        exception = 'Version of foo at dest changed from 0.9 to 1.0'
        with self.assertRaisesRegexp(PublishError, exception):
            self.script.check_versions_at_destination(dest, all_projects)
        self.script.lp.is_distro_archive.assert_called_once_with(dest)
        self.script.packagemanager.get_current_version_for_series\
            .assert_called_once_with('foo', 'vivid', dest=dest)
        self.script.DotProject.assert_called_once_with('foo')

    def test_check_for_unapproved_override(self):
        """check_for_unapproved() does nothing with ALLOW_UNAPPROVED=true"""
        env.ALLOW_UNAPPROVED = 'true'
        mps = {'foo': ['merge1', 'merge2']}
        lp_load = Mock()
        self.script.check_for_unapproved(mps, lp_load)
        self.assertEqual(lp_load.call_count, 0)
        self.assertEqual(self.script.logging.error.call_count, 0)
        self.script.logging.warning.assert_called_once_with(
            'Honoring your request to publish unapproved MPs.')

    def test_check_for_unapproved_all_good(self, status='Approved'):
        """check_for_unapproved() likes approved merges."""
        mps = {'foo': ['merge1', 'merge2']}
        lp_load = Mock(return_value=MergeFake(status))
        self.script.check_for_unapproved(mps, lp_load)
        self.assertEqual(lp_load.mock_calls, [call('merge1'), call('merge2')])
        self.assertEquals(self.script.logging.error.call_count, 0)

    def test_check_for_unapproved_all_merged(self):
        """check_for_unapproved() likes merged merges."""
        self.test_check_for_unapproved_all_good('Merged')

    def test_check_for_unapproved_needs_review(self):
        """check_for_unapproved() hates you!"""
        mps = {'foo': ['merge1', 'merge2']}
        lp_load = Mock(return_value=MergeFake('Needs review'))
        with self.assertRaisesRegexp(PublishError, 'merges are unapproved'):
            self.script.check_for_unapproved(mps, lp_load)
        self.assertEqual(lp_load.mock_calls, [call('merge1'), call('merge2')])
        self.assertEquals(
            self.script.logging.error.mock_calls,
            [call('These merges must be approved before publishing:'),
             call('example.com'),
             call('example.com')])

    def test_confirm_publisher_identity_success(self):
        """Ensure that we can correctly identify the publisher."""
        lp_people = defaultdict(str)
        env.PUBLISHER = ''
        env.BUILD_USER_ID = 'vorlon'
        self.script.confirm_publisher_identity(lp_people)
        self.assertEqual(lp_people, {'vorlon': ''})
        self.assertEqual(env.PUBLISHER, 'vorlon')

    def test_confirm_publisher_identity_failure(self):
        """Ensure that we fail to publish without identifying the publisher."""
        lp_people = dict()
        env.PUBLISHER = ''
        env.BUILD_USER_ID = 'vorlon'
        with self.assertRaisesRegexp(PublishError, 'not a valid launchpad'):
            self.script.confirm_publisher_identity(lp_people)
        self.assertEqual(lp_people, {})
        self.assertEqual(env.PUBLISHER, '')

    def test_ack_packaging_diffs_acked(self):
        """ack_packaging_diffs acknowledges ACK."""
        self.script.generate_xml_artifacts = Mock()
        env.ACK_PACKAGING = 'true'
        self.script.ack_packaging_diffs()
        self.assertEqual(self.script.glob.glob.call_count, 0)
        self.assertEqual(self.script.logging.error.call_count, 0)
        self.assertEqual(self.script.generate_xml_artifacts.call_count, 0)
        self.script.logging.warning.assert_called_once_with(
            'Honoring your request to ACK packaging changes.')

    def test_ack_packaging_diffs_none_to_ack(self):
        """ack_packaging_diffs approved of no diffs present."""
        env.ACK_PACKAGING = 'false'
        self.script.generate_xml_artifacts = Mock()
        self.script.glob.glob.return_value = []
        self.script.ack_packaging_diffs()
        self.script.DotProject.assert_called_once_with('*')
        self.script.DotProject().get_diff_path.assert_called_once_with('*')
        self.script.glob.glob.assert_called_once_with(
            self.script.DotProject().get_diff_path())
        self.assertEqual(self.script.logging.error.call_count, 0)
        self.assertEqual(self.script.generate_xml_artifacts.call_count, 0)

    def test_ack_packaging_diffs_halts_when_diff_found(self):
        """ack_packaging_diffs hates those diffs you have!"""
        env.ACK_PACKAGING = 'false'
        self.script.generate_xml_artifacts = Mock()
        self.script.glob.glob.return_value = ['/my/diff1', '/my/diff2']
        with self.assertRaisesRegexp(PublishError, 'need manual ACKing'):
            self.script.ack_packaging_diffs()
        self.script.DotProject.assert_called_once_with('*')
        self.script.DotProject().get_diff_path.assert_called_once_with('*')
        self.script.glob.glob.assert_called_once_with(
            self.script.DotProject().get_diff_path())
        self.assertEquals(
            self.script.logging.error.mock_calls,
            [call('Packaging changes found: diff1'),
             call('Packaging changes found: diff2')])
        self.script.generate_xml_artifacts.assert_called_once_with(
            self.script.glob.glob.return_value)

    def test_copy_to_dest(self):
        """Ensure that we poke the LP API correctly for copyPackage."""
        sources = [
            SourceFake('foo', '0.9', 'http://api.example.com/vivid'),
            SourceFake('bar', '0.8', 'http://api.example.com/vivid'),
        ]
        dest = Mock()
        self.script.copy_to_dest(sources, dest, target_pocket='Jeans')
        self.assertEqual(
            dest.copyPackage.mock_calls, [
                call(to_series='vivid', source_name='foo', version='0.9',
                     from_series='vivid', from_archive='http://api.example.com'
                     '/ubuntu/+archive/primary', to_pocket='Jeans',
                     include_binaries=True, from_pocket='Release'),
                call(to_series='vivid', source_name='bar', version='0.8',
                     from_series='vivid', from_archive='http://api.example.com'
                     '/ubuntu/+archive/primary', to_pocket='Jeans',
                     include_binaries=True, from_pocket='Release')])

    def mock_main(self):
        """Ensure enough of the publisher is mocked out for testing main()."""
        self.script.prevent_duplicate_publishing = Mock()
        self.script.check_versions_at_destination = Mock()
        self.script.check_for_unapproved = Mock()
        self.script.confirm_publisher_identity = Mock()
        self.script.ack_packaging_diffs = Mock()
        self.script.push_branches = Mock()
        self.script.initiate_publication = Mock()
        self.script.generate_xml_artifacts = Mock()
        self.script.SiloState.safe_load.return_value.series = \
            'https://api.launchpad.net/devel/ubuntu/vivid'

    def test_main_success(self):
        """publisher.main() should call the right functions."""
        self.mock_main()
        state = self.script.SiloState.safe_load()
        dest = self.script.lp.load()
        ppa = self.script.lp.load()
        self.assertEqual(self.script.main(), 0)
        self.script.prevent_duplicate_publishing.assert_called_once_with(
            state.is_publishing)
        state.set_publishing.assert_called_once_with()
        self.script.check_versions_at_destination.assert_called_once_with(
            dest, state.all_projects)
        self.script.check_for_unapproved.assert_called_once_with(
            state.mps, self.script.lp.load)
        self.script.confirm_publisher_identity.assert_called_once_with(
            self.script.lp.people)
        self.script.ack_packaging_diffs.assert_called_once_with()
        self.script.push_branches.assert_called_once_with(
            state.mps, self.script.lp.load)
        self.script.initiate_publication.assert_called_once_with(
            dest, ppa, state.set_package_version_list)
        self.script.generate_xml_artifacts.assert_called_once_with()
        state.set_published.assert_called_once_with()
        state.set_migrating.assert_called_once_with()
        self.assertEqual(state.save_config.mock_calls, [call(), call()])

    def test_main_failure(self):
        """publisher.main() should handle errors."""
        self.mock_main()
        state = self.script.SiloState.safe_load()
        dest = self.script.lp.load()
        self.script.push_branches = Mock(side_effect=PublishError('Whoops'))
        self.assertEqual(self.script.main(), 1)
        self.script.prevent_duplicate_publishing.assert_called_once_with(
            state.is_publishing)
        state.set_publishing.assert_called_once_with()
        self.script.check_versions_at_destination.assert_called_once_with(
            dest, state.all_projects)
        self.script.check_for_unapproved.assert_called_once_with(
            state.mps, self.script.lp.load)
        self.script.confirm_publisher_identity.assert_called_once_with(
            self.script.lp.people)
        self.script.push_branches.assert_called_once_with(
            state.mps, self.script.lp.load)
        self.assertEqual(self.script.initiate_publication.call_count, 0)
        self.script.logging.error.assert_called_once_with('Whoops')
        state.set_publish_failed.assert_called_once_with('Whoops')
        self.assertEqual(state.save_config.mock_calls, [call(), call()])
