import os
import glob

from mock import Mock, call

from tests.unit import CITrainScriptTestCase

from cupstream2distro.utils import os_path_join_safe

# Unused import necessary for code coverage reporting
from citrain import setup_citrain
setup_citrain


class SetupCITrainTestCase(CITrainScriptTestCase):
    scriptname = 'setup_citrain.py'

    def test_template_names(self):
        """Ensure we can find template filenames properly."""
        self.assertTrue(self.script.template)
        self.assertEqual(self.script.template('foobar'), 'foobar.xml.tmpl')

    def test_load_template(self):
        """Ensure we can load and populate templates correctly."""
        self.assertTrue(self.script.load_template)
        CONTEXT = {
            'SILO_NAME': 'ubuntu/landing-123',
            'SILOS_DIR': '/var/lib/jenkins/silos',
            'BINDIR':    '~/citrain/citrain',
        }
        result = self.script.load_template('build.xml.tmpl', CONTEXT)
        self.assertTrue('"ubuntu/landing-123" projects' in result)
        self.assertTrue('/var/lib/jenkins/silos/ubuntu/landing-123' in result)
        self.assertTrue('~/citrain/citrain/build.py' in result)
        self.assertTrue('~/citrain/citrain/mark_as_builtchecked.py' in result)

    def test_setup_job(self):
        """Ensure we write templates to disk correctly."""
        self.script.load_template = Mock(return_value='hello there!')
        self.script.JOBS_DIR = self.tempdir
        self.script.os.makedirs = os.makedirs
        self.script.setup_job('prepare-silo')
        config = os_path_join_safe(self.tempdir, 'prepare-silo', 'config.xml')
        with open(config) as f:
            self.assertEqual(f.read(), 'hello there!')
        self.script.load_template.assert_called_once_with(
            'prepare-silo.xml.tmpl',
            {'BINDIR': self.train_dir, 'RSYNC_OUTDIR': '~/out'})

    def test_setup_silo(self):
        """Ensure that we create all silo-related jobs correctly."""
        self.script.setup_job = Mock()
        self.script.JOBS_DIR = self.tempdir
        self.script.setup_silo('ubuntu/landing-000')
        self.assertEqual(
            self.script.setup_job.mock_calls, [
                call('ubuntu-landing-000-0-reconfigure',
                     'reconfigure.xml.tmpl',
                     {'SILO_NAME': 'ubuntu/landing-000',
                      'SILOS_DIR': os.path.expanduser('~/silos')}),
                call('ubuntu-landing-000-1-build',
                     'build.xml.tmpl',
                     {'SILO_NAME': 'ubuntu/landing-000',
                      'SILOS_DIR': os.path.expanduser('~/silos')}),
                call('ubuntu-landing-000-2-publish',
                     'publish.xml.tmpl',
                     {'SILO_NAME': 'ubuntu/landing-000',
                      'SILOS_DIR': os.path.expanduser('~/silos')}),
                call('ubuntu-landing-000-3-merge-clean',
                     'merge-clean.xml.tmpl',
                     {'SILO_NAME': 'ubuntu/landing-000',
                      'SILOS_DIR': os.path.expanduser('~/silos')})
            ])

    def test_main(self):
        """Tie deployment all together."""
        self.script.setup_silo = Mock()
        self.script.setup_job = Mock()
        self.assertEqual(self.script.main(), 0)
        self.assertEqual(self.script.setup_silo.mock_calls, [
            call('ubuntu/landing-000'),
            call('ubuntu/landing-001'),
            call('ubuntu/landing-002'),
            call('ubuntu/landing-003'),
            call('ubuntu/landing-004'),
            call('ubuntu/landing-005'),
            call('ubuntu/landing-006'),
            call('ubuntu/landing-007'),
            call('ubuntu/landing-008'),
            call('ubuntu/landing-009'),
            call('ubuntu/landing-010'),
            call('ubuntu/landing-011'),
            call('ubuntu/landing-012'),
            call('ubuntu/landing-013'),
            call('ubuntu/landing-014'),
            call('ubuntu/landing-015'),
            call('ubuntu/landing-016'),
            call('ubuntu/landing-017'),
            call('ubuntu/landing-018'),
            call('ubuntu/landing-019'),
            call('ubuntu/landing-020'),
            call('ubuntu/landing-021'),
            call('ubuntu/landing-022'),
            call('ubuntu/landing-023'),
            call('ubuntu/landing-024'),
            call('ubuntu/landing-025'),
            call('ubuntu/landing-026'),
            call('ubuntu/landing-027'),
            call('ubuntu/landing-028'),
            call('ubuntu/landing-029'),
            call('ubuntu/landing-030'),
            call('ubuntu-rtm/landing-000'),
            call('ubuntu-rtm/landing-001'),
            call('ubuntu-rtm/landing-002'),
            call('ubuntu-rtm/landing-003'),
            call('ubuntu-rtm/landing-004'),
            call('ubuntu-rtm/landing-005'),
            call('ubuntu-rtm/landing-006'),
            call('ubuntu-rtm/landing-007'),
            call('ubuntu-rtm/landing-008'),
            call('ubuntu-rtm/landing-009'),
            call('ubuntu-rtm/landing-010'),
            call('ubuntu-rtm/landing-011'),
            call('ubuntu-rtm/landing-012'),
            call('ubuntu-rtm/landing-013'),
            call('ubuntu-rtm/landing-014'),
            call('ubuntu-rtm/landing-015'),
            call('ubuntu-rtm/landing-016'),
            call('ubuntu-rtm/landing-017'),
            call('ubuntu-rtm/landing-018'),
            call('ubuntu-rtm/landing-019'),
            call('ubuntu-rtm/landing-020'),
            call('ubuntu-rtm/landing-021'),
            call('ubuntu-rtm/landing-022'),
            call('ubuntu-rtm/landing-023'),
            call('ubuntu-rtm/landing-024'),
            call('ubuntu-rtm/landing-025'),
            call('ubuntu-rtm/landing-026'),
            call('ubuntu-rtm/landing-027'),
            call('ubuntu-rtm/landing-028'),
            call('ubuntu-rtm/landing-029'),
            call('ubuntu-rtm/landing-030')])
        self.assertEqual(self.script.setup_job.mock_calls, [
            call('prepare-silo'),
            call('check-publication-migration'),
            call('upgrade-chroot'),
            call('apt-get-clean'),
            call('pbuilder-clean')])

    def test_integration(self):
        """A basic integration test for jenkins config deployment."""
        self.script.JOBS_DIR = self.tempdir
        self.script.os.makedirs = os.makedirs
        self.assertEqual(self.script.main(), 0)
        expected = [
            'apt-get-clean/config.xml',
            'check-publication-migration/config.xml',
            'pbuilder-clean/config.xml',
            'prepare-silo/config.xml',
            'ubuntu-landing-000-0-reconfigure/config.xml',
            'ubuntu-landing-000-1-build/config.xml',
            'ubuntu-landing-000-2-publish/config.xml',
            'ubuntu-landing-000-3-merge-clean/config.xml',
            'ubuntu-landing-001-0-reconfigure/config.xml',
            'ubuntu-landing-001-1-build/config.xml',
            'ubuntu-landing-001-2-publish/config.xml',
            'ubuntu-landing-001-3-merge-clean/config.xml',
            'ubuntu-landing-002-0-reconfigure/config.xml',
            'ubuntu-landing-002-1-build/config.xml',
            'ubuntu-landing-002-2-publish/config.xml',
            'ubuntu-landing-002-3-merge-clean/config.xml',
            'ubuntu-landing-003-0-reconfigure/config.xml',
            'ubuntu-landing-003-1-build/config.xml',
            'ubuntu-landing-003-2-publish/config.xml',
            'ubuntu-landing-003-3-merge-clean/config.xml',
            'ubuntu-landing-004-0-reconfigure/config.xml',
            'ubuntu-landing-004-1-build/config.xml',
            'ubuntu-landing-004-2-publish/config.xml',
            'ubuntu-landing-004-3-merge-clean/config.xml',
            'ubuntu-landing-005-0-reconfigure/config.xml',
            'ubuntu-landing-005-1-build/config.xml',
            'ubuntu-landing-005-2-publish/config.xml',
            'ubuntu-landing-005-3-merge-clean/config.xml',
            'ubuntu-landing-006-0-reconfigure/config.xml',
            'ubuntu-landing-006-1-build/config.xml',
            'ubuntu-landing-006-2-publish/config.xml',
            'ubuntu-landing-006-3-merge-clean/config.xml',
            'ubuntu-landing-007-0-reconfigure/config.xml',
            'ubuntu-landing-007-1-build/config.xml',
            'ubuntu-landing-007-2-publish/config.xml',
            'ubuntu-landing-007-3-merge-clean/config.xml',
            'ubuntu-landing-008-0-reconfigure/config.xml',
            'ubuntu-landing-008-1-build/config.xml',
            'ubuntu-landing-008-2-publish/config.xml',
            'ubuntu-landing-008-3-merge-clean/config.xml',
            'ubuntu-landing-009-0-reconfigure/config.xml',
            'ubuntu-landing-009-1-build/config.xml',
            'ubuntu-landing-009-2-publish/config.xml',
            'ubuntu-landing-009-3-merge-clean/config.xml',
            'ubuntu-landing-010-0-reconfigure/config.xml',
            'ubuntu-landing-010-1-build/config.xml',
            'ubuntu-landing-010-2-publish/config.xml',
            'ubuntu-landing-010-3-merge-clean/config.xml',
            'ubuntu-landing-011-0-reconfigure/config.xml',
            'ubuntu-landing-011-1-build/config.xml',
            'ubuntu-landing-011-2-publish/config.xml',
            'ubuntu-landing-011-3-merge-clean/config.xml',
            'ubuntu-landing-012-0-reconfigure/config.xml',
            'ubuntu-landing-012-1-build/config.xml',
            'ubuntu-landing-012-2-publish/config.xml',
            'ubuntu-landing-012-3-merge-clean/config.xml',
            'ubuntu-landing-013-0-reconfigure/config.xml',
            'ubuntu-landing-013-1-build/config.xml',
            'ubuntu-landing-013-2-publish/config.xml',
            'ubuntu-landing-013-3-merge-clean/config.xml',
            'ubuntu-landing-014-0-reconfigure/config.xml',
            'ubuntu-landing-014-1-build/config.xml',
            'ubuntu-landing-014-2-publish/config.xml',
            'ubuntu-landing-014-3-merge-clean/config.xml',
            'ubuntu-landing-015-0-reconfigure/config.xml',
            'ubuntu-landing-015-1-build/config.xml',
            'ubuntu-landing-015-2-publish/config.xml',
            'ubuntu-landing-015-3-merge-clean/config.xml',
            'ubuntu-landing-016-0-reconfigure/config.xml',
            'ubuntu-landing-016-1-build/config.xml',
            'ubuntu-landing-016-2-publish/config.xml',
            'ubuntu-landing-016-3-merge-clean/config.xml',
            'ubuntu-landing-017-0-reconfigure/config.xml',
            'ubuntu-landing-017-1-build/config.xml',
            'ubuntu-landing-017-2-publish/config.xml',
            'ubuntu-landing-017-3-merge-clean/config.xml',
            'ubuntu-landing-018-0-reconfigure/config.xml',
            'ubuntu-landing-018-1-build/config.xml',
            'ubuntu-landing-018-2-publish/config.xml',
            'ubuntu-landing-018-3-merge-clean/config.xml',
            'ubuntu-landing-019-0-reconfigure/config.xml',
            'ubuntu-landing-019-1-build/config.xml',
            'ubuntu-landing-019-2-publish/config.xml',
            'ubuntu-landing-019-3-merge-clean/config.xml',
            'ubuntu-landing-020-0-reconfigure/config.xml',
            'ubuntu-landing-020-1-build/config.xml',
            'ubuntu-landing-020-2-publish/config.xml',
            'ubuntu-landing-020-3-merge-clean/config.xml',
            'ubuntu-landing-021-0-reconfigure/config.xml',
            'ubuntu-landing-021-1-build/config.xml',
            'ubuntu-landing-021-2-publish/config.xml',
            'ubuntu-landing-021-3-merge-clean/config.xml',
            'ubuntu-landing-022-0-reconfigure/config.xml',
            'ubuntu-landing-022-1-build/config.xml',
            'ubuntu-landing-022-2-publish/config.xml',
            'ubuntu-landing-022-3-merge-clean/config.xml',
            'ubuntu-landing-023-0-reconfigure/config.xml',
            'ubuntu-landing-023-1-build/config.xml',
            'ubuntu-landing-023-2-publish/config.xml',
            'ubuntu-landing-023-3-merge-clean/config.xml',
            'ubuntu-landing-024-0-reconfigure/config.xml',
            'ubuntu-landing-024-1-build/config.xml',
            'ubuntu-landing-024-2-publish/config.xml',
            'ubuntu-landing-024-3-merge-clean/config.xml',
            'ubuntu-landing-025-0-reconfigure/config.xml',
            'ubuntu-landing-025-1-build/config.xml',
            'ubuntu-landing-025-2-publish/config.xml',
            'ubuntu-landing-025-3-merge-clean/config.xml',
            'ubuntu-landing-026-0-reconfigure/config.xml',
            'ubuntu-landing-026-1-build/config.xml',
            'ubuntu-landing-026-2-publish/config.xml',
            'ubuntu-landing-026-3-merge-clean/config.xml',
            'ubuntu-landing-027-0-reconfigure/config.xml',
            'ubuntu-landing-027-1-build/config.xml',
            'ubuntu-landing-027-2-publish/config.xml',
            'ubuntu-landing-027-3-merge-clean/config.xml',
            'ubuntu-landing-028-0-reconfigure/config.xml',
            'ubuntu-landing-028-1-build/config.xml',
            'ubuntu-landing-028-2-publish/config.xml',
            'ubuntu-landing-028-3-merge-clean/config.xml',
            'ubuntu-landing-029-0-reconfigure/config.xml',
            'ubuntu-landing-029-1-build/config.xml',
            'ubuntu-landing-029-2-publish/config.xml',
            'ubuntu-landing-029-3-merge-clean/config.xml',
            'ubuntu-landing-030-0-reconfigure/config.xml',
            'ubuntu-landing-030-1-build/config.xml',
            'ubuntu-landing-030-2-publish/config.xml',
            'ubuntu-landing-030-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-000-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-000-1-build/config.xml',
            'ubuntu-rtm-landing-000-2-publish/config.xml',
            'ubuntu-rtm-landing-000-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-001-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-001-1-build/config.xml',
            'ubuntu-rtm-landing-001-2-publish/config.xml',
            'ubuntu-rtm-landing-001-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-002-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-002-1-build/config.xml',
            'ubuntu-rtm-landing-002-2-publish/config.xml',
            'ubuntu-rtm-landing-002-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-003-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-003-1-build/config.xml',
            'ubuntu-rtm-landing-003-2-publish/config.xml',
            'ubuntu-rtm-landing-003-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-004-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-004-1-build/config.xml',
            'ubuntu-rtm-landing-004-2-publish/config.xml',
            'ubuntu-rtm-landing-004-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-005-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-005-1-build/config.xml',
            'ubuntu-rtm-landing-005-2-publish/config.xml',
            'ubuntu-rtm-landing-005-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-006-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-006-1-build/config.xml',
            'ubuntu-rtm-landing-006-2-publish/config.xml',
            'ubuntu-rtm-landing-006-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-007-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-007-1-build/config.xml',
            'ubuntu-rtm-landing-007-2-publish/config.xml',
            'ubuntu-rtm-landing-007-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-008-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-008-1-build/config.xml',
            'ubuntu-rtm-landing-008-2-publish/config.xml',
            'ubuntu-rtm-landing-008-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-009-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-009-1-build/config.xml',
            'ubuntu-rtm-landing-009-2-publish/config.xml',
            'ubuntu-rtm-landing-009-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-010-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-010-1-build/config.xml',
            'ubuntu-rtm-landing-010-2-publish/config.xml',
            'ubuntu-rtm-landing-010-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-011-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-011-1-build/config.xml',
            'ubuntu-rtm-landing-011-2-publish/config.xml',
            'ubuntu-rtm-landing-011-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-012-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-012-1-build/config.xml',
            'ubuntu-rtm-landing-012-2-publish/config.xml',
            'ubuntu-rtm-landing-012-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-013-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-013-1-build/config.xml',
            'ubuntu-rtm-landing-013-2-publish/config.xml',
            'ubuntu-rtm-landing-013-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-014-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-014-1-build/config.xml',
            'ubuntu-rtm-landing-014-2-publish/config.xml',
            'ubuntu-rtm-landing-014-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-015-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-015-1-build/config.xml',
            'ubuntu-rtm-landing-015-2-publish/config.xml',
            'ubuntu-rtm-landing-015-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-016-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-016-1-build/config.xml',
            'ubuntu-rtm-landing-016-2-publish/config.xml',
            'ubuntu-rtm-landing-016-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-017-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-017-1-build/config.xml',
            'ubuntu-rtm-landing-017-2-publish/config.xml',
            'ubuntu-rtm-landing-017-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-018-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-018-1-build/config.xml',
            'ubuntu-rtm-landing-018-2-publish/config.xml',
            'ubuntu-rtm-landing-018-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-019-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-019-1-build/config.xml',
            'ubuntu-rtm-landing-019-2-publish/config.xml',
            'ubuntu-rtm-landing-019-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-020-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-020-1-build/config.xml',
            'ubuntu-rtm-landing-020-2-publish/config.xml',
            'ubuntu-rtm-landing-020-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-021-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-021-1-build/config.xml',
            'ubuntu-rtm-landing-021-2-publish/config.xml',
            'ubuntu-rtm-landing-021-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-022-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-022-1-build/config.xml',
            'ubuntu-rtm-landing-022-2-publish/config.xml',
            'ubuntu-rtm-landing-022-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-023-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-023-1-build/config.xml',
            'ubuntu-rtm-landing-023-2-publish/config.xml',
            'ubuntu-rtm-landing-023-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-024-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-024-1-build/config.xml',
            'ubuntu-rtm-landing-024-2-publish/config.xml',
            'ubuntu-rtm-landing-024-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-025-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-025-1-build/config.xml',
            'ubuntu-rtm-landing-025-2-publish/config.xml',
            'ubuntu-rtm-landing-025-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-026-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-026-1-build/config.xml',
            'ubuntu-rtm-landing-026-2-publish/config.xml',
            'ubuntu-rtm-landing-026-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-027-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-027-1-build/config.xml',
            'ubuntu-rtm-landing-027-2-publish/config.xml',
            'ubuntu-rtm-landing-027-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-028-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-028-1-build/config.xml',
            'ubuntu-rtm-landing-028-2-publish/config.xml',
            'ubuntu-rtm-landing-028-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-029-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-029-1-build/config.xml',
            'ubuntu-rtm-landing-029-2-publish/config.xml',
            'ubuntu-rtm-landing-029-3-merge-clean/config.xml',
            'ubuntu-rtm-landing-030-0-reconfigure/config.xml',
            'ubuntu-rtm-landing-030-1-build/config.xml',
            'ubuntu-rtm-landing-030-2-publish/config.xml',
            'ubuntu-rtm-landing-030-3-merge-clean/config.xml',
            'upgrade-chroot/config.xml',
        ]
        result = sorted(glob.glob(os_path_join_safe(self.tempdir, '*', '*')))
        self.assertEqual(len(result), len(expected))
        for expectation in expected:
            self.assertIn(os_path_join_safe(self.tempdir, expectation), result)
