from tests.unit import CITrainScriptTestCase

from mock import Mock, call
from collections import defaultdict

# Unused import necessary for code coverage reporting
from citrain import watch_ppa
watch_ppa


class DotProjectFake:
    def __init__(self, items):
        self.items = items

    def __call__(self, name):
        self.name = name
        return self

    def __getattr__(self, key):
        return self.items.pop()


class BuildFake:
    def __init__(self, arch, status=''):
        self.arch_tag = arch
        self.buildstate = status
        self.build_log_url = 'example.com'


class SourceFake:
    _builds = []

    def __init__(self, name, version, series='vivid'):
        self.source_package_name = name
        self.source_package_version = version
        self.distro_series_link = series
        self.display_name = ' '.join([name, version, 'in', series])
        self.pocket = 'Release'
        self.archive_link = 'http://api.example.com/ubuntu/+archive/primary'

    def getBuilds(self):
        return self._builds.pop()


class distro_arch_series:
    def __init__(self):
        self._arches = ['arm64', 'ppc64el', 'i386', 'amd64', 'powerpc']

    @property
    def architecture_tag(self):
        return self._arches.pop()


class BinaryFake:
    def __init__(self, specific=True):
        self.distro_arch_series = distro_arch_series()
        self.architecture_specific = specific


class WatchPPATestCase(CITrainScriptTestCase):
    scriptname = 'watch_ppa.py'

    def test_watch_arch(self):
        """Ignore the correct arches."""
        watcher = self.script.watch_arch(['foo', 'i386', 'amd64'])
        self.assertTrue(watcher(BuildFake('foo')))
        self.assertTrue(watcher(BuildFake('i386')))
        self.assertTrue(watcher(BuildFake('amd64')))
        self.assertFalse(watcher(BuildFake('powerpc')))
        self.assertFalse(watcher(BuildFake('arm64')))
        self.assertFalse(watcher(BuildFake('ppc64el')))

    def test_watch_arch_any(self):
        """An empty arch list means we care about all arches."""
        watcher = self.script.watch_arch([])
        for arch in ('foo', 'i386', 'amd64', 'arm64', 'powerpc', 'ppc64el'):
            self.assertTrue(watcher(BuildFake(arch)))

    def test_find_supported_arches(self):
        """Ensure we care which are the right arches to watch."""
        source = Mock()
        source.getPublishedBinaries.return_value = [BinaryFake()] * 5
        dest = Mock()
        dest.getPublishedSources.return_value = [source]
        self.script.packagemanager.sort_by_date_created.return_value = [source]
        self.assertEqual(
            sorted(self.script.find_supported_arches(dest, 'foo')),
            ['amd64', 'arm64', 'i386', 'powerpc', 'ppc64el'])
        self.script.lp.load.assert_called_once_with(
            self.script.silo_state.series)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=self.script.lp.load(),
            status='Published',
        )
        self.script.packagemanager.sort_by_date_created\
            .assert_called_once_with(dest.getPublishedSources())

    def test_find_supported_arches_any(self):
        """For arch:any packages, find_supported_arches() returns set()."""
        source = Mock()
        source.getPublishedBinaries.return_value = [BinaryFake(False)] * 3
        dest = Mock()
        dest.getPublishedSources.return_value = [source]
        self.script.packagemanager.sort_by_date_created.return_value = [source]
        self.assertEqual(
            sorted(self.script.find_supported_arches(dest, 'foo')), [])
        self.script.lp.load.assert_called_once_with(
            self.script.silo_state.series)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=self.script.lp.load(),
            status='Published',
        )
        self.script.packagemanager.sort_by_date_created\
            .assert_called_once_with(dest.getPublishedSources())

    def test_find_supported_arches_some(self):
        """Some binary packages don't support all arches."""
        binaries = [BinaryFake(False)] * 3 + [BinaryFake(True)] * 3
        source = Mock()
        source.getPublishedBinaries.return_value = binaries
        dest = Mock()
        dest.getPublishedSources.return_value = [source]
        self.script.packagemanager.sort_by_date_created.return_value = [source]
        self.assertEqual(
            sorted(self.script.find_supported_arches(dest, 'foo')),
            ['amd64', 'i386', 'powerpc'])
        self.script.lp.load.assert_called_once_with(
            self.script.silo_state.series)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=self.script.lp.load(),
            status='Published',
        )
        self.script.packagemanager.sort_by_date_created\
            .assert_called_once_with(dest.getPublishedSources())

    def test_find_supported_arches_none(self):
        """Return set() when no package exists at destination."""
        binaries = []
        source = Mock()
        source.getPublishedBinaries.return_value = binaries
        dest = Mock()
        dest.getPublishedSources.return_value = [source]
        self.script.packagemanager.sort_by_date_created.return_value = []
        self.assertEqual(
            sorted(self.script.find_supported_arches(dest, 'foo')),
            [])
        self.script.lp.load.assert_called_once_with(
            self.script.silo_state.series)
        dest.getPublishedSources.assert_called_once_with(
            exact_match=True,
            source_name='foo',
            distro_series=self.script.lp.load(),
            status='Published',
        )
        self.script.packagemanager.sort_by_date_created\
            .assert_called_once_with(dest.getPublishedSources())

    def test_find_all_uploaded(self):
        """Find everything that has been uploaded from this silo."""
        self.script.find_all.return_value = [
            '/path/to/foo_0.1_source.ppa.upload',
            '/path/to/bar_0.2_source.ppa.upload',
        ]
        self.script.DotProject = DotProjectFake(['0.2', '0.1'])
        self.assertEqual(
            self.script.find_all_uploaded(),
            {'bar': set(['0.2']), 'foo': set(['0.1'])})

    def test_find_all_uploaded_multiple_series(self):
        """We can handle uploading multiple series from a single silo."""
        self.script.find_all.return_value = [
            '/path/to/foo_0.1_source.ppa.upload',
            '/path/to/foo_0.1~utopic_source.ppa.upload',
            '/path/to/bar_0.2_source.ppa.upload',
            '/path/to/bar_0.2~utopic_source.ppa.upload',
        ]
        self.script.DotProject = DotProjectFake(
            ['0.2~utopic', '0.2', '0.1~utopic', '0.1'])
        self.assertEqual(
            self.script.find_all_uploaded(),
            {'bar': set(['0.2', '0.2~utopic']),
             'foo': set(['0.1', '0.1~utopic'])})

    def test_find_ppa_sources(self):
        """Use launchpadlib to correctly get the sources from a PPA."""
        ppa = Mock()
        ppa.getPublishedSources.return_value = [
            SourceFake('a', '0.9'),
            SourceFake('b', '1.0'),
            SourceFake('c', '1.1'),
        ]
        sources = set()
        self.script.find_ppa_sources(ppa, sources, defaultdict(set))
        for source in ppa.getPublishedSources.return_value:
            self.assertIn(source, sources)
        self.assertEqual(
            ppa.getPublishedSources.mock_calls,
            [call(status='Published'), call(status='Pending')])

    def test_find_ppa_sources_versions(self):
        """Use launchpadlib to correctly get multiple series from a PPA."""
        ppa = Mock()
        ppa.getPublishedSources.return_value = [
            SourceFake('a', '0.9'),
            SourceFake('b', '1.0'),
            SourceFake('c', '1.1'),
            SourceFake('c', '1.1~utopic'),
        ]
        sources = set()
        versions = defaultdict(set)
        self.script.find_ppa_sources(ppa, sources, versions)
        self.assertEqual(
            versions, {
                'a': set(['0.9']),
                'b': set(['1.0']),
                'c': set(['1.1', '1.1~utopic']),
            })

    def test_compare_versions_matches(self):
        """Correctly determine whether the PPA contains what we expected."""
        uploaded = defaultdict(set)
        versions = defaultdict(set)
        # Exact match is great
        uploaded['foo'].add('1.0')
        versions['foo'].add('1.0')
        self.assertTrue(self.script.compare_versions(versions, uploaded, []))
        # Extra manual uploads in the PPA are fine, too
        versions['bar'].add('3.0')
        self.assertTrue(self.script.compare_versions(versions, uploaded, []))
        # Manual sources are present
        self.assertTrue(self.script.compare_versions(versions, {}, ['bar']))
        # Superceded versions not yet disappeared shouldn't break stuff
        versions['foo'].add('0.9')
        self.assertTrue(self.script.compare_versions(versions, uploaded, []))

    def test_compare_versions_no_match(self):
        """Correctly identify when the PPA does not contain what we expect."""
        uploaded = defaultdict(set)
        versions = defaultdict(set)
        # False if 'foo' doesn't (yet) appear in the ppa
        uploaded['foo'].add('1.0')
        self.assertFalse(self.script.compare_versions(versions, uploaded, []))
        # False if 'foo' appears in the ppa with wrong version
        versions['foo'].add('0.9')
        self.assertFalse(self.script.compare_versions(versions, uploaded, []))
        # Manual upload we're expecting is not found
        self.assertFalse(self.script.compare_versions(versions, {}, ['none']))

    def test_pretty_mismatch(self):
        """Print PPA version mismatches attractively."""
        uploaded = defaultdict(set)
        versions = defaultdict(set)
        self.assertEqual(
            self.script.pretty_mismatch(versions, uploaded, []),
            'Versions in PPA did not match our expectations.')
        uploaded['foo'].add('1.0')
        self.assertEqual(
            self.script.pretty_mismatch(versions, uploaded, []),
            'Versions in PPA did not match our expectations.\n'
            'Missing any version of: foo')
        versions['foo'].add('0.9')
        self.assertEqual(
            self.script.pretty_mismatch(versions, uploaded, []),
            'Versions in PPA did not match our expectations.\n'
            'Expected: foo 1.0\n'
            'Found:    foo 0.9')
        self.assertEqual(
            self.script.pretty_mismatch(versions, {}, ['bar']),
            'Versions in PPA did not match our expectations.\n'
            'Missing any version of: bar')
        self.assertEqual(
            self.script.pretty_mismatch(versions, uploaded, ['bar']),
            'Versions in PPA did not match our expectations.\n'
            'Missing any version of: bar\n'
            'Expected: foo 1.0\n'
            'Found:    foo 0.9')
        uploaded['foo'].add('1.0~utopic')
        uploaded['fizzbuzz'].add('ZZ9 Plural Z Alpha')
        versions['fizzbuzz'].add('ZZ9 Plural Y Alpha')
        self.assertEqual(
            self.script.pretty_mismatch(versions, uploaded, ['bar', 'grill']),
            'Versions in PPA did not match our expectations.\n'
            'Missing any version of: bar, grill\n'
            'Expected: fizzbuzz ZZ9 Plural Z Alpha\n'
            'Found:    fizzbuzz ZZ9 Plural Y Alpha\n'
            'Expected: foo 1.0, 1.0~utopic\n'
            'Found:    foo 0.9')

    def test_wait_for_ppa_instantly(self):
        """Handle a PPA that has the packages immediately.

        This never actually happens, but it was the simplest test to write.
        """
        self.script.compare_versions = Mock()
        self.script.find_ppa_sources = Mock()
        uploaded = {
            'foo': set(['1.0-0ubuntu1', '1.0~utopic-0ubuntu1']),
            'bar': set(['9000-0ubuntu1', '9000~utopic-0ubuntu1']),
        }
        sources = self.script.wait_for_ppa(uploaded)
        self.script.lp.load.assert_called_once_with(self.script.silo_state.ppa)
        self.script.compare_versions.assert_called_once_with(
            defaultdict(set), uploaded, self.script.silo_state.sources)
        self.assertEqual(self.script.find_ppa_sources.mock_calls, [])
        self.assertEqual(sources, set())

    def test_wait_for_ppa_timeout(self):
        """Handle a failed upload and the PPA times out."""
        times = [self.script.PPA_TIMEOUT*2, self.script.PPA_TIMEOUT/2, 0]
        self.script.time.time = lambda: times.pop()
        self.script.compare_versions = Mock()
        self.script.compare_versions.return_value = False
        self.script.pretty_mismatch = Mock()
        self.script.find_ppa_sources = Mock()
        uploaded = {
            'foo': set(['1.0-0ubuntu1', '1.0~utopic-0ubuntu1']),
            'bar': set(['9000-0ubuntu1', '9000~utopic-0ubuntu1']),
        }
        sources = self.script.wait_for_ppa(uploaded)
        self.script.lp.load.assert_called_once_with(self.script.silo_state.ppa)
        self.assertEqual(self.script.compare_versions.call_count, 2)
        self.script.compare_versions.assert_called_with(
            defaultdict(set), uploaded, self.script.silo_state.sources)
        self.script.time.sleep.assert_called_once_with(300)
        self.script.pretty_mismatch.assert_called_once_with(
            defaultdict(set), uploaded, self.script.silo_state.sources)
        self.script.logging.error.assert_called_once_with(
            self.script.pretty_mismatch())
        self.script.find_ppa_sources.assert_called_once_with(
            self.script.lp.load(),
            set(), defaultdict(set))
        self.assertIsNone(sources)

    def test_wait_for_ppa_normal(self):
        """Handle the normal case from a launchpad PPA."""
        self.script.time.time.return_value = 0
        comparisons = [True, False, False]
        self.script.compare_versions = lambda x, y, z: comparisons.pop()
        self.script.pretty_mismatch = Mock()
        self.script.find_ppa_sources = Mock()
        uploaded = {
            'foo': set(['1.0-0ubuntu1', '1.0~utopic-0ubuntu1']),
            'bar': set(['9000-0ubuntu1', '9000~utopic-0ubuntu1']),
        }
        sources = self.script.wait_for_ppa(uploaded)
        self.script.lp.load.assert_called_once_with(self.script.silo_state.ppa)
        self.assertEqual(self.script.time.sleep.mock_calls,
                         [call(300), call(300)])
        self.assertEqual(self.script.pretty_mismatch.call_count, 0)
        self.assertEqual(self.script.find_ppa_sources.call_count, 2)
        self.script.find_ppa_sources.assert_called_with(
            self.script.lp.load(),
            set(), defaultdict(set))
        # The set is populated by find_ppa_sources which we've mocked out,
        # so we really do expect it to be empty even though this test is
        # testing the "successful case"
        self.assertEqual(sources, set())

    def test_main_null(self):
        """watch_ppa.main does some basic things in the simplest case."""
        self.script.find_all_uploaded = Mock()
        self.script.wait_for_ppa = Mock()
        self.script.wait_for_ppa.return_value = None
        self.assertEqual(self.script.main(), 1)

    def test_main_ignore_stale_sources(self):
        """Watch_ppa shouldn't fail over stale builds."""
        uploaded = {
            'foo': set(['1.0-0ubuntu1', '1.0~utopic-0ubuntu1']),
            'bar': set(['9000-0ubuntu1', '9000~utopic-0ubuntu1']),
        }
        self.script.find_supported_arches = Mock(return_value=['i386'])
        self.script.find_all_uploaded = Mock()
        self.script.find_all_uploaded.return_value = uploaded
        self.script.wait_for_ppa = Mock()
        sources = [
            SourceFake('foo', '0.9-0ubuntu1'),
            SourceFake('foo', '1.0-0ubuntu1'),
            SourceFake('bar', '9000-0ubuntu1'),
        ]
        sources[0]._builds = [[BuildFake('i386', 'Failed to build')]]
        sources[1]._builds = [[BuildFake('i386', 'Successfully built')]]
        sources[2]._builds = [[BuildFake('i386', 'Successfully built')]]
        self.script.wait_for_ppa.return_value = set(sources)
        self.assertEqual(self.script.main(), 0)
        self.script.lp.load.assert_called_once_with(
            self.script.silo_state.dest)
        self.script.find_all_uploaded.assert_called_once_with()
        self.script.wait_for_ppa.assert_called_once_with(uploaded)
        self.assertEqual(self.script.wait_for_ppa(), set())
        self.assertEqual(self.script.logging.error.call_count, 0)
        dest = self.script.lp.load()
        calls = self.script.packagemanager\
            .get_current_version_for_series.mock_calls
        self.assertEqual(len(calls), 2)
        self.assertIn(call('foo', 'vivid', dest=dest), calls)
        self.assertIn(call('bar', 'vivid', dest=dest), calls)
        dest_ver = self.script.packagemanager.get_current_version_for_series()
        calls = self.script.DotProject.soft_save.mock_calls
        # Note '0.9-0ubuntu1' not in uploaded and thus ignored.
        self.assertEqual(len(calls), 2)
        self.assertIn(call('foo', dest=dest_ver, ours='1.0-0ubuntu1'), calls)
        self.assertIn(call('bar', dest=dest_ver, ours='9000-0ubuntu1'), calls)

    def test_main_build_failed(self):
        """Watch_ppa shouldn't fail over stale builds."""
        uploaded = {
            'foo': set(['1.0-0ubuntu1', '1.0~utopic-0ubuntu1']),
            'bar': set(['9000-0ubuntu1', '9000~utopic-0ubuntu1']),
        }
        self.script.find_supported_arches = Mock(return_value=['i386'])
        self.script.find_all_uploaded = Mock()
        self.script.find_all_uploaded.return_value = uploaded
        self.script.wait_for_ppa = Mock()
        sources = [
            SourceFake('foo', '0.9-0ubuntu1'),
            SourceFake('foo', '1.0-0ubuntu1'),
            SourceFake('bar', '9000-0ubuntu1'),
        ]
        sources[0]._builds = [[BuildFake('i386', 'Failed to build')]]
        sources[1]._builds = [[BuildFake('i386', 'Failed to build')]]
        sources[2]._builds = [[BuildFake('i386', 'Failed to build')]]
        self.script.wait_for_ppa.return_value = set(sources)
        self.assertEqual(self.script.main(), 2)
        self.script.lp.load.assert_called_once_with(
            self.script.silo_state.dest)
        self.script.find_all_uploaded.assert_called_once_with()
        self.script.wait_for_ppa.assert_called_once_with(uploaded)
        self.assertEqual(self.script.wait_for_ppa(), set())
        self.assertEqual(
            self.script.logging.error.mock_calls,
            [call('example.com'), call('example.com')])
        calls = self.script.packagemanager\
            .get_current_version_for_series.mock_calls
        self.assertEqual(len(calls), 0)
        self.assertEqual(len(self.script.DotProject.soft_save.mock_calls), 0)

    def test_main_go_for_the_gold(self):
        """Test sleeping in watch_ppa.main for 100% test coverage."""
        uploaded = {
            'foo': set(['1.0-0ubuntu1', '1.0~utopic-0ubuntu1']),
            'bar': set(['9000-0ubuntu1', '9000~utopic-0ubuntu1']),
        }
        self.script.find_all_uploaded = Mock()
        self.script.find_all_uploaded.return_value = uploaded
        self.script.wait_for_ppa = Mock()
        sources = [
            SourceFake('foo', '1.0-0ubuntu1'),
        ]
        sources[0]._builds = [
            [BuildFake('i386', 'Successfully built')],
            [BuildFake('i386', 'Currently building')],
        ]
        self.script.find_supported_arches = Mock(return_value=['i386'])
        self.script.wait_for_ppa.return_value = set(sources)
        self.assertEqual(self.script.main(), 0)
        self.script.lp.load.assert_called_once_with(
            self.script.silo_state.dest)
        self.script.find_all_uploaded.assert_called_once_with()
        self.script.wait_for_ppa.assert_called_once_with(uploaded)
        self.assertEqual(self.script.wait_for_ppa(), set())
        self.assertEqual(self.script.logging.error.call_count, 0)
        self.script.time.sleep.assert_called_once_with(300)
