import os
import sys
import time
import signal

from mock import Mock, patch, call
from collections import defaultdict

from tests.unit import DirectoryAwareTestCase
from cupstream2distro import silomanager
from cupstream2distro.silomanager import SiloState
from cupstream2distro.utils import env, os_path_join_safe


class dest:
    name = 'primary'
    self_link = 'https://api.launchpad.net/devel/ubuntu/+archive/primary'

    class distribution:
        name = 'ubuntu'


class series:
    name = 'vivid'
    self_link = 'https://api.launchpad.net/devel/ubuntu/vivid'


class SiloManagerTestCase(DirectoryAwareTestCase):
    maxDiff = None

    def setUp(self):
        """Initialize a SiloState instance for us to poke at."""
        super(SiloManagerTestCase, self).setUp()
        self.state = SiloState('ubuntu/landing-123', do_load=False)
        self.state._configfile = os_path_join_safe(
            self.data_dir, 'landing-123', 'config')
        self.state.load_config()

    @patch('cupstream2distro.silomanager.SiloState._configfile',
           os_path_join_safe(
               DirectoryAwareTestCase.data_dir, 'landing-123', 'config'))
    def test_silostate_safe_load_success(self):
        """Safe load returns something."""
        self.assertIsNotNone(SiloState.safe_load('foo'))

    @patch('cupstream2distro.silomanager.SiloState._configfile',
           'invalid_file_will_fail_to_load')
    def test_silostate_safe_load_failure(self):
        """Safe load returns None (and doesn't raise an exception)."""
        self.assertIsNone(SiloState.safe_load('foo'))

    @patch('cupstream2distro.silomanager.logging')
    @patch('cupstream2distro.silomanager.SiloState._configfile',
           os_path_join_safe(
               DirectoryAwareTestCase.data_dir, 'landing-123', 'config'))
    def test_silostate_is_project_unique(self, logMock):
        """Can determine whether a project is already assigned or not."""
        env.IGNORECONFLICTS = 'true'
        self.assertFalse(SiloState.is_project_unique(dest, series, 'qtmir'))
        logMock.warning.assert_called_once_with(
            'qtmir is already in ubuntu/landing-000')
        env.IGNORECONFLICTS = 'false'
        self.assertFalse(SiloState.is_project_unique(dest, series, 'qtmir'))
        logMock.error.assert_called_once_with(
            'qtmir is already in ubuntu/landing-000')
        self.assertTrue(SiloState.is_project_unique(dest, series, 'fubar'))
        series.self_link = 'not vivid'
        self.assertTrue(SiloState.is_project_unique(dest, series, 'qtmir'))

    @patch('cupstream2distro.silomanager.SiloState._configfile',
           os_path_join_safe(
               DirectoryAwareTestCase.data_dir, 'landing-123', 'config'))
    def test_silostate_is_project_unique_reconfigure(self):
        """Can determine whether a project is assigned while we reconfigure."""
        self.assertTrue(SiloState.is_project_unique(
            dest, series, 'qtmir', 'ubuntu/landing-000'))
        self.assertTrue(SiloState.is_project_unique(
            dest, series, 'fubar', 'ubuntu/landing-000'))
        series.self_link = 'not vivid'
        self.assertTrue(SiloState.is_project_unique(
            dest, series, 'qtmir', 'ubuntu/landing-000'))

    @patch('cupstream2distro.silomanager.SiloState.requestid', '1234')
    @patch('cupstream2distro.silomanager.SiloState.__init__')
    def test_silostate_find_by_requestid(self, initMock):
        """Find a silo given only the REQUEST_ID."""
        initMock.return_value = None
        silo_state = SiloState.find_by_requestid('1234')
        initMock.assert_called_once_with('ubuntu/landing-000')
        self.assertIsNotNone(silo_state)

    @patch('cupstream2distro.silomanager.SiloState.requestid')
    @patch('cupstream2distro.silomanager.SiloState.__init__')
    def test_silostate_find_by_requestid_not_found(self, initMock, idMock):
        """Find no silos given a non-assigned REQUEST_ID."""
        idMock.return_value = '1235'
        initMock.return_value = None
        silo_state = SiloState.find_by_requestid('1234')
        initMock.assert_called_with('ubuntu-rtm/landing-030')
        # 31 vivid silos, 31 RTM silos.
        self.assertEqual(initMock.call_count, 62)
        self.assertIsNone(silo_state)

    @patch('cupstream2distro.silomanager.os.makedirs')
    @patch('cupstream2distro.silomanager.os.path.isfile')
    def test_silostate_find_first_available(self, isfileMock, mkdirMock):
        """Find the first unassigned silo."""
        isfileMock.return_value = True
        self.assertIsNone(SiloState.find_first_available())
        expected = [
            '~/silos/ubuntu/landing-000/config',
            '~/silos/ubuntu/landing-001/config',
            '~/silos/ubuntu/landing-002/config',
            '~/silos/ubuntu/landing-003/config',
            '~/silos/ubuntu/landing-004/config',
            '~/silos/ubuntu/landing-005/config',
            '~/silos/ubuntu/landing-006/config',
            '~/silos/ubuntu/landing-007/config',
            '~/silos/ubuntu/landing-008/config',
            '~/silos/ubuntu/landing-009/config',
            '~/silos/ubuntu/landing-010/config',
            '~/silos/ubuntu/landing-011/config',
            '~/silos/ubuntu/landing-012/config',
            '~/silos/ubuntu/landing-013/config',
            '~/silos/ubuntu/landing-014/config',
            '~/silos/ubuntu/landing-015/config',
            '~/silos/ubuntu/landing-016/config',
            '~/silos/ubuntu/landing-017/config',
            '~/silos/ubuntu/landing-018/config',
            '~/silos/ubuntu/landing-019/config',
            '~/silos/ubuntu/landing-020/config',
            '~/silos/ubuntu/landing-021/config',
            '~/silos/ubuntu/landing-022/config',
            '~/silos/ubuntu/landing-023/config',
            '~/silos/ubuntu/landing-024/config',
            '~/silos/ubuntu/landing-025/config',
            '~/silos/ubuntu/landing-026/config',
            '~/silos/ubuntu/landing-027/config',
            '~/silos/ubuntu/landing-028/config',
            '~/silos/ubuntu/landing-029/config',
            '~/silos/ubuntu/landing-030/config',
        ]
        self.assertEqual(len(isfileMock.mock_calls), 31)
        for i, kall in enumerate(isfileMock.mock_calls):
            self.assertEqual(kall, call(os.path.expanduser(expected[i])))
        self.assertEqual(mkdirMock.mock_calls, [])

    @patch('cupstream2distro.silomanager.os.makedirs')
    @patch('cupstream2distro.silomanager.os.path.isfile')
    def test_silostate_find_first_available_rtm(self, isfileMock, mkdirMock):
        """Find the first unassigned RTM silo."""
        isfileMock.return_value = True
        self.assertIsNone(
            SiloState.find_first_available(distribution='ubuntu-rtm'))
        expected = [
            '~/silos/ubuntu-rtm/landing-000/config',
            '~/silos/ubuntu-rtm/landing-001/config',
            '~/silos/ubuntu-rtm/landing-002/config',
            '~/silos/ubuntu-rtm/landing-003/config',
            '~/silos/ubuntu-rtm/landing-004/config',
            '~/silos/ubuntu-rtm/landing-005/config',
            '~/silos/ubuntu-rtm/landing-006/config',
            '~/silos/ubuntu-rtm/landing-007/config',
            '~/silos/ubuntu-rtm/landing-008/config',
            '~/silos/ubuntu-rtm/landing-009/config',
            '~/silos/ubuntu-rtm/landing-010/config',
            '~/silos/ubuntu-rtm/landing-011/config',
            '~/silos/ubuntu-rtm/landing-012/config',
            '~/silos/ubuntu-rtm/landing-013/config',
            '~/silos/ubuntu-rtm/landing-014/config',
            '~/silos/ubuntu-rtm/landing-015/config',
            '~/silos/ubuntu-rtm/landing-016/config',
            '~/silos/ubuntu-rtm/landing-017/config',
            '~/silos/ubuntu-rtm/landing-018/config',
            '~/silos/ubuntu-rtm/landing-019/config',
            '~/silos/ubuntu-rtm/landing-020/config',
            '~/silos/ubuntu-rtm/landing-021/config',
            '~/silos/ubuntu-rtm/landing-022/config',
            '~/silos/ubuntu-rtm/landing-023/config',
            '~/silos/ubuntu-rtm/landing-024/config',
            '~/silos/ubuntu-rtm/landing-025/config',
            '~/silos/ubuntu-rtm/landing-026/config',
            '~/silos/ubuntu-rtm/landing-027/config',
            '~/silos/ubuntu-rtm/landing-028/config',
            '~/silos/ubuntu-rtm/landing-029/config',
            '~/silos/ubuntu-rtm/landing-030/config',
        ]
        self.assertEqual(len(isfileMock.mock_calls), 31)
        for i, kall in enumerate(isfileMock.mock_calls):
            self.assertEqual(kall, call(os.path.expanduser(expected[i])))

    @patch('cupstream2distro.silomanager.os.makedirs')
    @patch('cupstream2distro.silomanager.os.path.isfile')
    def test_silostate_find_first_available_found(self, isfileMock, mkdirMock):
        """Find first unassgined silo with one available."""
        isfileMock.return_value = False
        self.assertEqual(
            'ubuntu-rtm/landing-000',
            SiloState.find_first_available(distribution='ubuntu-rtm'))
        expected = [
            '~/silos/ubuntu-rtm/landing-000/config',
        ]
        self.assertEqual(len(isfileMock.mock_calls), 1)
        for i, kall in enumerate(isfileMock.mock_calls):
            self.assertEqual(kall, call(os.path.expanduser(expected[i])))
        mkdirMock.assert_called_once_with(
            os.path.expanduser('~/silos/ubuntu-rtm/landing-000'))

    def test_silostate_new_blank(self):
        """Create a new SiloState instance for a fresh silo assignment."""
        silo_state = SiloState.new_blank(
            siloname='ubuntu/landing-789',
            requestid='zippy',
            dest='mars',
            series='parallel',
            landers=['curiosity', 'spirit', 'opportunity'])
        self.assertEqual(silo_state.siloname, 'ubuntu/landing-789')
        self.assertEqual(
            silo_state.configfile,
            os.path.expanduser('~/silos/ubuntu/landing-789/config'))
        self.assertEqual(silo_state.requestid, 'zippy')
        self.assertEqual(silo_state.dest, 'mars')
        self.assertEqual(silo_state.series, 'parallel')
        self.assertEqual(
            silo_state.landers, ['curiosity', 'spirit', 'opportunity'])
        self.assertEqual(
            silo_state._config['global']['status'],
            dict(message='', ping=True, state=-1, url=''))

    def test_silostate_load_config(self):
        """SiloState can read it's config from disk."""
        self.state.load_config()
        self.assertEqual(
            list(self.state._config),
            ['siloname', 'global', 'sources', 'landers', 'mps', 'requestid'])
        self.assertEqual(
            self.state._config['landers'],
            ['mzanetti', 'greyback', 'kgunn'])

    def test_silostate_load_config_failure(self):
        """Raise an exception when the file cannot be read."""
        self.state._configfile = 'does_not_exist'
        self.assertRaises(IOError, self.state.load_config)

    @patch('cupstream2distro.silomanager.os.makedirs')
    @patch('cupstream2distro.silomanager.shutil')
    def test_silostate_save_config(self, shutilMock, mkdirMock):
        """Save config to disk."""
        tmp = os_path_join_safe(self.tempdir, 'config')
        self.state._configfile = tmp
        self.state.save_config()
        newState = SiloState('phony-siloname', do_load=False)
        newState._configfile = tmp
        newState.load_config()
        self.assertEqual(self.state._config, newState._config)
        mkdirMock.assert_called_once_with(
            os.path.expanduser('~/status/ubuntu'))
        shutilMock.copy2.assert_called_once_with(
            tmp,
            os.path.expanduser('~/status/ubuntu/landing-123'))

    @patch('cupstream2distro.silomanager.os.unlink')
    def test_silostate_remove_rsync_file(self, unlinkMock):
        """Remove the rsync file from the ~/status/ directory."""
        self.state.remove_rsync_file()
        unlinkMock.assert_called_once_with(
            os.path.expanduser('~/status/ubuntu/landing-123'))

    def test_silostate_rsyncfile(self):
        """Identify the location of the rsync file."""
        self.assertEqual(
            self.state.rsyncfile,
            os.path.expanduser('~/status/ubuntu/landing-123'))

    def test_silostate_config_step(self):
        """Read the config step."""
        self.assertEqual(self.state.config_step, 1)

    def test_silostate_set_config_step(self):
        """Set the config step."""
        self.state.config_step = 2
        self.assertEqual(self.state._config['global']['step'], 2)

    def test_silostate_set_config_status(self):
        """Set the config status."""
        silomanager.env.BUILD_URL = 'http://example.com/jenkins/'
        self.state.set_config_status(-1, 'Everything is broken!')
        self.assertEqual(
            self.state._config['global']['status'],
            dict(
                message='Everything is broken!',
                ping=True,
                state=-1,
                url='http://example.com/jenkins/console'))

    def test_silostate_set_package_version_list(self):
        """Set the package version list."""
        versions = defaultdict(set)
        versions['foo'].add('1.0')
        versions['foo'].add('1.0~utopic')
        versions['bar'].add('5.0')
        versions['bar'].add('1.0')
        self.state.set_package_version_list(versions)
        self.assertEqual(
            self.state._config['global']['pkgversionlist'],
            {'foo': ['1.0', '1.0~utopic'], 'bar': ['1.0', '5.0']})

    def test_silostate_all_projects(self):
        """List all projects including MPs, source uploads, and syncs."""
        self.state._config['sources'] = ['foobar']
        self.assertEqual(
            self.state.all_projects,
            ['foobar', 'platform-api', 'qtmir', 'qtmir-gles', 'qtubuntu',
                'unity-api', 'unity-notifications', 'unity8'])

    def test_silostate_booleans(self):
        """Boolean status properties have correct values."""
        self.assertFalse(self.state.is_merging)
        self.assertFalse(self.state.is_publishing)
        self.assertFalse(self.state.is_published)
        self.assertFalse(self.state.is_empty)
        self.assertFalse(self.state.is_cleaning)
        self.assertTrue(self.state.is_built_checked)

    def test_silostate_status_setters(self):
        """Set various silo statuses."""
        setMock = self.state.set_config_status = Mock()
        self.state.set_build_successful()
        setMock.assert_called_once_with(3, 'Packages built')
        self.state.set_cleaning('Foo')
        setMock.assert_called_with(11, 'Foo')
        self.state.set_merging()
        setMock.assert_called_with(9, 'Merging')
        self.state.set_merge_failed('Bar')
        setMock.assert_called_with(10, 'Bar')
        self.state.set_migrating('Grill')
        setMock.assert_called_with(8, 'Grill')
        self.state.set_building()
        setMock.assert_called_with(2, 'Building')
        self.state.set_build_failed('It dun broke')
        setMock.assert_called_with(4, 'It dun broke')
        self.state.set_preparing()
        setMock.assert_called_with(2, 'Preparing packages')
        self.state.set_publishing()
        setMock.assert_called_with(6, 'Publishing')
        self.state.set_publish_failed('Doh!')
        setMock.assert_called_with(7, "Can't publish: Doh!")

    def test_silostate_step_setters(self):
        """Set various silo steps."""
        self.state.set_empty()
        self.assertEqual(self.state._config['global']['step'], 0)
        self.state.set_built_checked()
        self.assertEqual(self.state._config['global']['step'], 1)
        self.state.set_published()
        self.assertEqual(self.state._config['global']['step'], 2)
        self.state.set_done()
        self.assertEqual(self.state._config['global']['step'], 3)

    def test_silostate_getters(self):
        """Read various state properties."""
        self.assertEqual(
            self.state.dest,
            'https://api.launchpad.net/devel/ubuntu/+archive/primary')
        self.assertEqual(
            self.state.mps,
            {'qtmir-gles':
                ['https://api.launchpad.net/devel/~mir-team/qtmir/gles-sync/'
                 '+merge/235275'],
             'qtubuntu':
                ['https://api.launchpad.net/devel/~gerboland/qtubuntu/'
                 'exposeOrientation/+merge/232252'],
             'unity-api':
                ['https://api.launchpad.net/devel/~dandrader/unity-api/splash/'
                 '+merge/233288'],
             'qtmir':
                ['https://api.launchpad.net/devel/~gerboland/qtmir/'
                 'exposeOrientation/+merge/232485',
                 'https://api.launchpad.net/devel/~dandrader/qtmir/'
                 'parse_splash_from_desktop_file/+merge/233287'],
             'unity-notifications':
                ['https://api.launchpad.net/devel/~macslow/'
                 'unity-notifications/dont-filter-newline/+merge/235447'],
             'unity8':
                ['https://api.launchpad.net/devel/~dandrader/unity8/'
                 'real_header_splashscreen/+merge/233286',
                 'https://api.launchpad.net/devel/~aacid/unity8/'
                 'need_schemas_qmluitests/+merge/235252',
                 'https://api.launchpad.net/devel/~mterry/unity8/'
                 'lc-all-fix2/+merge/235303',
                 'https://api.launchpad.net/devel/~mzanetti/unity8/'
                 'fix-greeter-left-edge/+merge/235508',
                 'https://api.launchpad.net/devel/~mzanetti/unity8/'
                 'fix-indicators-ap/+merge/235511',
                 'https://api.launchpad.net/devel/~aacid/unity8/'
                 'fix_gsv_test/+merge/235580'],
             'platform-api':
                ['https://api.launchpad.net/devel/~gerboland/platform-api/'
                 'exposeOrientation/+merge/232250']})
        self.assertEqual(
            self.state.ppa,
            'https://api.launchpad.net/devel/~ci-train-ppa-service/+archive/'
            'ubuntu/landing-123')
        self.assertEqual(
            self.state.series,
            'https://api.launchpad.net/devel/ubuntu/vivid')
        self.assertEqual(self.state.state, 3)
        self.assertEqual(self.state.landers, ['mzanetti', 'greyback', 'kgunn'])
        self.assertEqual(self.state.requestid, '1411123362781')
        self.assertEqual(self.state.siloname, 'ubuntu/landing-123')
        self.assertIsNone(self.state.source_sync_from_silo)
        self.assertIsNone(self.state.source_sync_all_contents)
        self.assertIsNone(self.state.source_archive)
        self.assertIsNone(self.state.source_series)
        self.assertEqual(self.state.sources, [])

    def test_silostate_property_setters(self):
        """Set various state properties."""
        self.state.config_step = 9000
        self.assertEqual(self.state._config['global']['step'], 9000)
        self.state.dest = 'far, far away'
        self.assertEqual(self.state._config['global']['dest'], 'far, far away')
        self.state.landers = list('foo')
        self.assertEqual(self.state._config['landers'], ['f', 'o', 'o'])
        self.state.mps = 'flibbyjib'
        self.assertEqual(self.state._config['mps'], 'flibbyjib')
        self.state.ppa = 'my ppa'
        self.assertEqual(self.state._config['global']['ppa'], 'my ppa')
        self.state.requestid = 'bazinga'
        self.assertEqual(self.state._config['requestid'], 'bazinga')
        self.state.series = 'superserial'
        self.assertEqual(self.state._config['global']['series'], 'superserial')
        self.state.source_sync_from_silo = True
        self.assertTrue(self.state._config['source_sync_from_silo'])
        self.state.source_sync_all_contents = True
        self.assertTrue(self.state._config['source_sync_all_contents'])
        self.state.source_archive = 'ubuntu'
        self.assertEqual(self.state._config['source_archive'], 'ubuntu')
        self.state.source_series = 'vivid'
        self.assertEqual(self.state._config['source_series'], 'vivid')
        self.state.state = 'snicker'
        self.state.sources = ['foo', 'bar']
        self.assertEqual(self.state._config['sources'], ['foo', 'bar'])
        self.assertEqual(
            self.state._config['global']['status']['state'], 'snicker')

    def test_silostate_set_ready(self):
        """Set ready to build status."""
        statusMock = self.state.set_config_status = Mock()
        self.state.set_ready()
        statusMock.assert_called_once_with(1, 'Silo ready to build')
        self.assertEqual(self.state.config_step, 0)

    def test_silostate_set_reconfigure_failed(self):
        """Set reconfigure failed status."""
        statusMock = self.state.set_config_status = Mock()
        self.state.set_reconfigure_failed('You did a bad thing.')
        statusMock.assert_called_once_with(
            5, 'Reconfigure failed: You did a bad thing.')
        self.assertEqual(self.state.config_step, -1)

    def test_silostate_append_mp(self):
        """Add an MP to the list of that project's MPs."""
        self.state._config['mps'] = {}
        self.state.append_mp('foobar', 'example.com')
        self.assertEqual(self.state._config['mps']['foobar'], ['example.com'])
        self.state.append_mp('foobar', 'example.com/zip')
        self.assertEqual(
            self.state._config['mps']['foobar'],
            ['example.com', 'example.com/zip'])
        self.state.append_mp('fringle', 'example.com/futz')
        self.assertEqual(
            self.state._config['mps'],
            {'foobar': ['example.com', 'example.com/zip'],
             'fringle': ['example.com/futz']})

    def test_silostate_remove_source(self):
        """Remove a source package from the silo configuration."""
        self.assertEqual(
            sorted(self.state.mps.keys()),
            ['platform-api', 'qtmir', 'qtmir-gles', 'qtubuntu', 'unity-api',
                'unity-notifications', 'unity8'])
        self.state.remove_source('unity-notifications')
        self.state.remove_source('unity-notifications')
        self.state.remove_source('platform-api')
        self.assertEqual(
            sorted(self.state.mps.keys()),
            ['qtmir', 'qtmir-gles', 'qtubuntu', 'unity-api', 'unity8'])
        self.state.sources = list('asdfg')
        self.state.remove_source('s')
        self.state.remove_source('f')
        self.state.remove_source('f')
        self.assertEqual(self.state.sources, ['a', 'd', 'g'])

    @patch('cupstream2distro.silomanager.shutil', Mock())
    @patch('cupstream2distro.silomanager.os.makedirs', Mock())
    def test_get_mtime(self):
        """Get the modified time of the config file."""
        tmp = os_path_join_safe(self.tempdir, 'config')
        first = self.state.get_mtime()
        self.assertGreater(first, 0)
        self.state._configfile = tmp
        self.state.save_config()
        second = self.state.get_mtime()
        self.assertGreater(second, first)
        time.sleep(0.01)
        self.assertEqual(second, self.state.get_mtime())

    def test_additional_targets(self):
        """Get and set the additional targets list."""
        self.assertEqual(self.state.additional_targets, None)
        self.state.additional_targets = list('abc')
        self.assertEqual(self.state.additional_targets, ['a', 'b', 'c'])
        self.assertEqual(
            self.state._config['additional_targets'], ['a', 'b', 'c'])

    def test_signal_handler(self):
        """Our signal handler is installed globally."""
        self.assertTrue(
            silomanager.cleanup_on_kill is signal.getsignal(signal.SIGTERM))

    def test_exception_handler(self):
        """Our exception handler is the global default."""
        self.assertTrue(
            silomanager.cleanup_on_exception is sys.excepthook)

    @patch('cupstream2distro.silomanager.SiloState')
    def test_save_error_message(self, silostateMock):
        """In the event of an unhandled exception we can save it to disk."""
        env.SILONAME = 'my/fake/silo'
        silomanager.save_error_message('Whoa, errors galore!')
        silostateMock.assert_called_once_with('my/fake/silo')
        silostateMock().set_config_status.assert_called_once_with(
            -1, 'Whoa, errors galore!')
        silostateMock().save_config.assert_called_once_with()

    @patch('cupstream2distro.silomanager.SiloState')
    def test_save_error_message_no_silo(self, silostateMock):
        """Errors not associated with silos are lost."""
        env.SILONAME = ''  # Equivalent to undefined
        silomanager.save_error_message('Error without a silo!')
        self.assertEqual(silostateMock.call_count, 0)

    @patch('cupstream2distro.silomanager.SiloState')
    def test_cleanup_on_kill(self, silostateMock):
        """Save the state when getting killed."""
        env.SILONAME = 'xyzzy/landing-ten-billion'
        silomanager.cleanup_on_kill(15, None)
        silostateMock.assert_called_once_with('xyzzy/landing-ten-billion')
        silostateMock().set_config_status.assert_called_once_with(
            -1, 'Caught signal 15, aborting.')
        silostateMock().save_config.assert_called_once_with()

    @patch('cupstream2distro.silomanager.SiloState')
    @patch('cupstream2distro.silomanager.default_excepthook')
    def test_cleanup_on_exception(self, exceptMock, silostateMock):
        """Save the state when there's an unhandled exception."""
        env.SILONAME = 'ubuntu/landing-negative-one'
        silomanager.cleanup_on_exception(None, Exception('It broke'), None)
        silostateMock.assert_called_once_with('ubuntu/landing-negative-one')
        silostateMock().set_config_status.assert_called_once_with(
            -1, 'Uncaught exception: Exception: It broke')
        silostateMock().save_config.assert_called_once_with()
