import os

from six.moves import range as gen_range
from mock import Mock, patch

from cupstream2distro.errors import UnexpectedAbsolutePathError
from cupstream2distro.utils import (
    SILO_DIR,
    env,
    log_value_of,
    os_path_join_safe,
    reorder_branches,
    find_all,
    run_script,
    singleton,
)

from tests.unit import DirectoryAwareTestCase


class Merge:
    """Mimic an LP Merge object."""
    prerequisite_branch = None

    def __init__(self, source_branch, prereq_branch=None):
        self.source_branch = Branch(source_branch)
        if prereq_branch:
            self.prerequisite_branch = Branch(prereq_branch)

    def __repr__(self):
        pre = self.prerequisite_branch
        pre = ', {}'.format(pre) if pre else ''
        return '{}({}{})'.format(
            self.__class__.__name__, self.source_branch, pre)


class Branch:
    """Mimic an LP Branch object."""
    def __init__(self, bzr_identity):
        self.bzr_identity = bzr_identity

    def __repr__(self):
        return self.bzr_identity


class UtilsTestCase(DirectoryAwareTestCase):
    @patch('cupstream2distro.utils.logging')
    def test_env_init(self, logMock):
        """Log environment variables when debugging enabled."""
        env.DEBUG = 'true'
        env.__init__()
        logMock.debug.assert_called_once_with(str(env))
        env.DEBUG = 'false'

    def test_env_setters(self):
        """Setting attributes on env sets real environment variables."""
        rand = os.path.basename(self.tempdir)
        setattr(env, rand, 'hello')
        self.assertEqual(os.getenv(rand), 'hello')
        setattr(env, 'hello', rand)
        self.assertEqual(os.getenv('hello'), rand)

    def test_env_getters(self):
        """All real environment variables are accessible as attrs on env."""
        for var in os.environ:
            self.assertEqual(os.environ[var], getattr(env, var))

    def test_os_path_join_safe(self):
        """Our os_path_join_safe gives identical results as os.path.join()."""
        home = os.path.expanduser('~')
        cases = (
            ('',),
            (home,),
            (home, 'foo'),
            [home] + list('abc'),
            [os.sep] + [str(i) for i in gen_range(10)],
        )
        for case in cases:
            self.assertEqual(os_path_join_safe(*case), os.path.join(*case))

    def test_os_path_join_safe_exception(self):
        """But raise an exception when there are too many abs paths."""
        home = os.path.expanduser('~')
        cases = (
            (home, home),
            (home, os.sep),
            (home, 'foo', home),
        )
        for case in cases:
            with self.assertRaises(UnexpectedAbsolutePathError):
                os_path_join_safe(*case)

    def test_silo_dir(self):
        """SILO_DIR() function can find the correct silo dir dynamically."""
        env.SILONAME = ''
        self.assertEqual(SILO_DIR(), os.path.expanduser('~/silos/'))
        env.SILONAME = 'ubuntu/landing-123'
        self.assertEqual(
            SILO_DIR(), os.path.expanduser('~/silos/ubuntu/landing-123'))
        env.SILONAME = 'fubar'
        self.assertEqual(SILO_DIR(), os.path.expanduser('~/silos/fubar'))

    def test_find_all(self):
        """We can recursively find files in the silo."""
        results = list(find_all('utils.py', self.root_dir))
        self.assertIn(
            os_path_join_safe(self.root_dir, 'tests', 'unit', 'test_utils.py'),
            results)
        self.assertIn(
            os_path_join_safe(self.root_dir, 'cupstream2distro', 'utils.py'),
            results)

    def test_find_all_ignore(self):
        """We can recursively find files in the silo."""
        results = list(find_all('utils.py', self.root_dir, 'tests/unit/'))
        self.assertNotIn(
            os_path_join_safe(self.root_dir, 'tests', 'unit', 'test_utils.py'),
            results)
        self.assertIn(
            os_path_join_safe(self.root_dir, 'cupstream2distro', 'utils.py'),
            results)

    @patch('cupstream2distro.utils.sys')
    def test_run_script_as_import(self, sysMock):
        """The run_script() func is a NOP when name != __main__."""
        mainMock = Mock()
        sysMock.argv = []
        run_script('some_module', 'docdocdoc', mainMock)
        self.assertEqual(mainMock.mock_calls, [])
        self.assertEqual(sysMock.exit.mock_calls, [])
        self.assertEqual(sysMock.stdout.write.mock_calls, [])

    @patch('cupstream2distro.utils.sys')
    def test_run_script_as_script(self, sysMock):
        """Invoke main() when run as a script."""
        mainMock = Mock()
        sysMock.argv = []
        run_script('__main__', 'docdocdoc', mainMock)
        mainMock.assert_called_once_with()
        sysMock.exit.assert_called_once_with(mainMock())
        self.assertEqual(sysMock.stdout.write.mock_calls, [])

    @patch('cupstream2distro.utils.sys')
    def test_run_script_with_help(self, sysMock):
        """Print help when the user asks for help."""
        mainMock = Mock()
        for h in ('-h', '--help'):
            sysMock.argv = [h]
            run_script('__main__', h*5, mainMock)
            self.assertEqual(mainMock.mock_calls, [])
            self.assertEqual(sysMock.exit.mock_calls, [])
            sysMock.stdout.write.assert_called_with(h*5)

    def order_shuffler(self, expected, shuffles):
        """Shuffle and re-sort values, then compare against expectations."""
        for shuf in shuffles:
            merges = [expected[i] for i in shuf]
            result = reorder_branches(merges)
            log_value_of.shuf()
            log_value_of.merges()
            log_value_of.expected()
            log_value_of.result()
            self.assertEqual(len(result), len(merges))
            for i, merge in enumerate(expected):
                self.assertIs(result[i], merge)

    def test_reorder_branches(self):
        """One chain with a few isolated nodes is ordered correctly."""
        expected = [
            Merge('0'),
            Merge('1'),
            Merge('2', '1'),
            Merge('3', '2'),
            Merge('4'),
        ]
        shuffles = [
            (0, 1, 2, 3, 4),
            (0, 3, 1, 4, 2),
            (0, 3, 2, 1, 4),
        ]
        self.order_shuffler(expected, shuffles)

    def test_reorder_branches_recursive(self):
        """One big chain always gets the right order."""
        expected = [
            Merge('v'),
            Merge('t', 'v'),
            Merge('g', 't'),
            Merge('l', 'g'),
            Merge('z', 'l'),
            Merge('a', 'z'),
        ]
        shuffles = [
            (0, 1, 2, 3, 4, 5),
            (5, 4, 3, 2, 1, 0),
            (0, 5, 2, 3, 4, 1),
            (3, 1, 4, 2, 5, 0),
            (3, 0, 1, 5, 4, 2),
            (2, 0, 1, 5, 4, 3),
        ]
        self.order_shuffler(expected, shuffles)

    def test_reorder_branches_nonlinear(self):
        """Nodes with multiple parents don't cause duplicate output."""
        expected = [
            Merge('alpha'),
            Merge('beta', 'alpha'),
            Merge('zeta', 'beta'),
            Merge('eta', 'beta'),
            Merge('delta', 'alpha'),
            Merge('gamma', 'delta'),
            Merge('epsilon', 'delta'),
        ]
        shuffles = [
            (0, 1, 2, 3, 4, 5, 6),
            (5, 6, 1, 4, 2, 3, 0),
            (5, 1, 6, 0, 2, 4, 3),
            (1, 0, 5, 2, 6, 3, 4),
        ]
        self.order_shuffler(expected, shuffles)

    def test_reorder_branches_multiple_graphs(self):
        """Two disconnected chains can be ordered consistently."""
        expected = [
            Merge('alpha'),
            Merge('beta', 'alpha'),
            Merge('gamma'),
            Merge('delta', 'gamma'),
        ]
        shuffles = [
            (0, 1, 2, 3),
            (1, 0, 3, 2),
        ]
        self.order_shuffler(expected, shuffles)

    def test_reorder_branches_missing_prereqs(self):
        """Merges defining absent prereqs should be counted as leafs."""
        expected = [
            Merge('alpha', 'missing'),
        ]
        shuffles = [
            (0,),
        ]
        self.order_shuffler(expected, shuffles)

    def test_reorder_branches_no_change(self):
        """Only reorder branches when they have pre-requisites."""
        merges = [Merge('1'), Merge('2'), Merge('3')]
        result = reorder_branches(merges)
        self.assertEqual(len(result), len(merges))
        for i, merge in enumerate(merges):
            self.assertIs(result[i], merge)

    def test_singleton(self):
        """There Can Be Only One."""
        @singleton
        class Highlander:
            pass
        self.assertIs(Highlander(), Highlander())
        self.assertIs(Highlander(), Highlander)
        self.assertEqual(id(Highlander()), id(Highlander()))
        self.assertEqual(id(Highlander), id(Highlander()))
